//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//    2 Jun 06  Brian Frank  Ported from Java to Fan
//

**
** CompilerLog manages logging compiler messages.  The default
** writes everything to standard output.
**
class CompilerLog
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  **
  ** Construct for specified output stream.
  **
  new make(OutStream out := Sys.out)
  {
    this.out = out
  }

//////////////////////////////////////////////////////////////////////////
// Methods
//////////////////////////////////////////////////////////////////////////

  **
  ** Is level configured to log verbosely
  **
  Bool isVerbose()
  {
    return level <= LogLevel.verbose
  }

  **
  ** Indent the output.
  **
  Void indent()
  {
    indentation++
  }

  **
  ** Unindent the output.
  **
  Void unindent()
  {
    indentation--
    if (indentation < 0) indentation = 0
  }

  **
  ** Convenience for log(LogLevel.error, msg, aux, err)
  **
  Void error(Str msg, Str aux := null, Err err := null)
  {
    log(LogLevel.error, msg, aux, err)
  }

  **
  ** Convenience for log(LogLevel.warning, msg, aux, err)
  **
  Void warning(Str msg, Str aux := null, Err err := null)
  {
    log(LogLevel.warning, msg, aux, err)
  }

  **
  ** Convenience for log(LogLevel.message, msg, aux, err)
  **
  Void message(Str msg, Str aux := null, Err err := null)
  {
    log(LogLevel.message, msg, aux, err)
  }

  **
  ** Convenience for log(LogLevel.verbose, msg, aux, err)
  **
  Void verbose(Str msg, Str aux := null, Err err := null)
  {
    log(LogLevel.verbose, msg, aux, err)
  }

  **
  ** Generate a log entry.  The log entry is only generated if the
  ** specified level is greater than or equal to the configured level
  ** field.  The msg indicates the message to report.  Aux is used
  ** to separate extra information, by convention it is typically the
  ** target of msg such as the file name.  Err is used to report the
  ** associated exception.
  **
  virtual Void log(LogLevel level, Str msg, Str aux := null, Err err := null)
  {
    if (level < this.level) return
    if (level >= LogLevel.warning) print(level.toStr.capitalize).print(": ")
    n := indentation*2
    print(Str.spaces(n))
    print(msg); n += msg.size
    if (aux != null)
    {
      print("  [").print(aux).print("]")
    }
    printLine

    if (err != null)
    {
      if (isVerbose)
        err.trace(out)
      else
        print(Str.spaces(indentation*2+4)).printLine(err)
    }
  }

//////////////////////////////////////////////////////////////////////////
// CompilerErr
//////////////////////////////////////////////////////////////////////////

  **
  ** Log a CompilerErr
  **
  virtual Void compilerErr(CompilerErr err)
  {
    if (LogLevel.error < this.level) return
    if (err.location != null)
    {
      printLine(err.location.toLocationStr + ": " + err.message)
    }
    else
    {
      printLine(err.message)
    }
    if (isVerbose) err.trace(out)
  }

//////////////////////////////////////////////////////////////////////////
// IO
//////////////////////////////////////////////////////////////////////////

  **
  ** Print a string without trailing newline.
  **
  CompilerLog print(Obj s)
  {
    out.print(s)
    return this
  }

  **
  ** Print a line.
  **
  CompilerLog printLine(Obj s := "")
  {
    out.printLine(s).flush
    return this
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  ** Max severity of log entries to report
  LogLevel level := LogLevel.message

  ** Current level of indentation
  Int indentation := 0

  ** Sink for all output
  OutStream out := null

}