//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   26 Dec 05  Brian Frank  Creation
//    5 Jun 06  Brian Frank  Ported from Java to Fan
//

**
** ResolveImports maps every Import node in each CompilationUnit to a pod
** and ensures that it exists and that no imports are duplicated.  Then we
** create a map for all the types which are imported into the CompilationUnit
** so that the Parser can quickly distinguish between a type identifier and
** other identifiers.  The results of this step populate Import.pod and
** CompilationUnit.importedTypes.
**
class ResolveImports : CompilerStep
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  **
  ** Constructor takes the associated Compiler
  **
  new make(Compiler compiler)
    : super(compiler)
  {
    resolved[compiler.pod.name] = compiler.pod
  }

//////////////////////////////////////////////////////////////////////////
// Methods
//////////////////////////////////////////////////////////////////////////

  **
  ** Run the step
  **
  override Void run()
  {
    log.verbose("ResolveImports")

    // process each unit for Import.pod
    units.each |CompilationUnit unit|
    {
      resolveImports(unit)
    }
    bombIfErr

    // process each unit for CompilationUnit.importedTypes
    units.each |CompilationUnit unit|
    {
      resolveImportedTypes(unit)
    }
    bombIfErr
  }

  **
  ** Resolve all the imports in the specified unit
  ** and ensure there are no duplicates.
  **
  private Void resolveImports(CompilationUnit unit)
  {
    // map to keep track of duplicate imports
    // within this compilation unit
    dups := Str:Using[:]

    // process each import statement (remember the
    // first one is an implicit import of sys)
    unit.usings.each |Using u|
    {
      podName := u.podName

      // check for duplicate imports
      key := podName
      if (u.typeName != null) key += "::$u.typeName"
      if (dups.containsKey(key))
      {
        err("Duplicate using '$key'", u.location)
        return
      }
      dups[key] = u

      // if already resolved, then just use it
      u.resolvedPod = resolved[podName]

      // resolve the import and cache in resolved map
      if (u.resolvedPod == null)
      {
        pod := Pod.find(podName, false)
        if (pod == null)
        {
          err("Pod not found '$podName'", u.location)
          return
        }
        resolved[podName] = u.resolvedPod = ns.importPod(pod)
      }

      // if type specified, then resolve type
      if (u.typeName != null)
      {
        u.resolvedType = u.resolvedPod.findType(u.typeName)
        if (u.resolvedType== null)
        {
          err("Type not found in pod '$podName::$u.typeName'", u.location)
          return
        }
      }
    }
  }

  **
  ** Create a unified map of type names to CType[] for all the
  ** imports in the specified unit (this includes types within
  ** the pod being compilied itself).  For example if foo::Thing
  ** and bar::Thing are imported, then importedTypes would contain
  **   "Thing" : [foo::Thing, bar::Thing]
  **
  private Void resolveImportedTypes(CompilationUnit unit)
  {
    // name -> CType[]
    types := Str:CType[][:]

    // add types for my own pod
    addAll(types, compiler.pod.types)

    // add pod level imports first
    unit.usings.each |Using u|
    {
      if (u.typeName == null)
        addAll(types, u.resolvedPod.types)
    }

    // add type specific imports last (these
    // override any pod level imports)
    unit.usings.each |Using u|
    {
      if (u.typeName != null)
      {
        if (u.asName == null)
        {
          types[u.typeName] = [u.resolvedType]
        }
        else
        {
          remove(types, u.resolvedType)
          types[u.asName] = [u.resolvedType]
        }
      }
    }

    /*
    // dump
    echo("--- types for $unit")
    ids := types.keys.sort
    ids.each |Str id| { echo("$id = ${types[id]}") }
    */

    // save away on unit
    unit.importedTypes = types
  }

  private Void addAll(Str:CType[] types, CType[] toAdd)
  {
    toAdd.each |CType t|
    {
      list := types[t.name]
      if (list == null) types[t.name] = list = CType[,]
      list.add(t)
    }
  }

  private Void remove(Str:CType[] types, CType t)
  {
    list := types[t.name]
    if (list != null)
    {
      for (i:=0; i<list.size; ++i)
        if (list[i].qname == t.qname) { list.removeAt(i); break }
    }
  }

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  **
  ** Resolve a fully qualified type name into its CType representation.
  ** This may be a TypeDef within the compilation units or could be
  ** an imported type.  If the type name cannot be resolved then we
  ** log an error and return null.
  **
  static CType resolveQualified(CompilerSupport cs, Str podName, Str typeName, Location location)
  {
    // first check pod being compiled
    if (podName == cs.compiler.pod.name)
    {
      t := cs.compiler.pod.findType(typeName)
      if (t == null)
      {
        cs.err("Type '$typeName' not found within pod being compiled", location)
        return null
      }
      return t
    }

    // otherwise we need to try to resolve pod
    pod := Pod.find(podName, false)
    if (pod == null)
    {
      cs.err("Pod not found '$podName'", location);
      return null
    }

    // now try to lookup type
    t := pod.findType(typeName, false)
    if (t == null)
    {
      cs.err("Type '$typeName' not found in pod '$podName'", location);
      return null
    }

    return cs.ns.importType(t)
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  Str:CPod resolved := Str:CPod[:]  // reuse ImportedPods across units

}
