//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   05 Apr 06  Andy Frank  Creation
//

using web

**
** FandocType.
**
class FandocType : FandocWeblet
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  new make()
  {
    pod   = (Pod)req.stash["pod"]
    ftype = (Type)req.stash["type"]

    sorter := |Slot a, Slot b -> Int| { return a.name <=> b.name }
    filter := |Slot s -> Bool|
    {
      return !s.isSynthetic && (ftype == Obj.type || s.parent != Obj.type)
    }

    methods = ftype.methods.rw.sort(sorter).findAll(filter)
    fields  = ftype.fields.rw.sort(sorter).findAll(filter)
  }

//////////////////////////////////////////////////////////////////////////
// Weblet
//////////////////////////////////////////////////////////////////////////

  override Void get()
  {
    doHeader("$ftype - Fandoc")

    out.div("class='header'")
    out.table("width='100%' cellspacing='0' cellpadding='0'")
    out.tr

    out.td
    doBreadcrumb(pod, ftype)
    out.tdEnd

    out.td("style='text-align: right;'")
    doSearch()
    out.tdEnd

    out.trEnd
    out.tableEnd
    out.divEnd

    out.div("class='content'")
    typeInfo()
    fieldTable()
    methodTable()
    out.divEnd

    doFooter
  }

//////////////////////////////////////////////////////////////////////////
// Type
//////////////////////////////////////////////////////////////////////////

  **
  ** Type info.
  **
  private Void typeInfo()
  {
    out.div("class='typeInfo'")

    // Tabs
    out.div("class='tabs'")
    out.a(`#`, "onclick='showAllSlots(); return false;' style='padding-right: 25px;'")
    out.w("Show All Slots").aEnd.nl
    out.w("Fandoc ").span("class='sep'").w("|").spanEnd.w(" ")
    out.a(typeSrcUri(ftype)).w("Source").aEnd
    out.divEnd

    // Type name
    out.h1(ftype.toStr)

    // Fully specified type name
    out.p
    if (ftype.isAbstract && ftype.isClass) out.w("abstract ")
    if (ftype.isClass) out.w("class ")
    else if (ftype.isMixin) out.w("mixin ")
    else if (ftype.isEnum) out.w("enum ")
    out.w(ftype.name)
    out.pEnd

    // Type comment
    out.p("class='doc'")
    doc := ftype.doc
    if (doc != null)
      out.w(safe(firstSentance(doc)))
    out.pEnd

    // Inheritance
    chain := [ftype]
    base  := ftype.base
    while (base != null)
    {
      chain.insert(0, base)
      base = base.base
    }
    out.h2("Inheritance")
    out.p
    chain.each |Type t, Int i|
    {
      if (i > 0)
        out.w(" &#8592; ")
      if (i < chain.size-1)
        writeTypeLink(t, null, true)
      else
        out.w(t.qname)
    }
    out.pEnd

    // Mixins
    mixins := ftype.inheritance.findAll |Type t->Bool| { return t.isMixin }
    first  := true
    if (!mixins.isEmpty && (mixins.size > 1 || mixins.first != ftype))
    {
      out.h2("Mixins")
      out.p
      mixins.each |Type t|
      {
        if (t == ftype) return // don't include self for mixins
        if (first) first = false
        else out.w(", ")
        writeTypeLink(t, null, true)
      }
      out.pEnd
    }

    // Subtypes (skip for sys::Obj and sys::Test)
    if (ftype != Obj.type && ftype != Test.type)
    {
      sub := findAllSubTypes(ftype)
      if (!sub.isEmpty)
      {
        out.h2("Sub Types")
        out.p
        sub.each |Type t, Int i|
        {
          if (i > 0) out.w(", ")
          writeTypeLink(t, null, true)
        }
        out.pEnd
      }
    }

    out.divEnd  // typeSummary
  }

//////////////////////////////////////////////////////////////////////////
// Slot summary
//////////////////////////////////////////////////////////////////////////

  **
  ** Field summary table.
  **
  private Void fieldTable()
  {
    col1 := |Obj obj, WebOutStream out|
    {
      f := (Field)obj
      writeScope(f)
      writeTypeLink(f.fieldType)
    }
    col2 := |Obj obj, WebOutStream out|
    {
      f := (Field)obj
      out.w("<b>").w(f.name).w("</b>")
    }
    slotTable("Fields", fields, col1, col2)
  }

  **
  ** Method summary table.
  **
  private Void methodTable()
  {
    col1 := |Obj obj, WebOutStream out|
    {
      m := (Method)obj
      if (m.isCtor) out.w("new")
      else
      {
        writeScope(m)
        writeTypeLink(m.returnType)
      }
    }
    col2 := |Obj obj, WebOutStream out|
    {
      m := (Method)obj
      out.w("<b>").w(m.name).w("</b>")
      out.w("(")
      p := m.params
      for (j:=0; j<p.size; j++)
      {
        if (j > 0) out.w(", ")
        writeTypeLink(p[j].paramType)
        out.w(" ${p[j].name}")
        if (p[j].hasDefault)
        out.w(" := def")
      }
      out.w(")")
    }
    slotTable("Methods", methods, col1, col2)
  }

  **
  ** Slot table.
  **
  Void slotTable(Str title, Slot[] rows,
    |Slot s, WebOutStream out| col1, |Slot s, WebOutStream out| col2)
  {
    // Never display table if no rows
    if (rows.size == 0) return

    out.div("class='slotTable'")
    out.h1(title)
    out.table("width='100%' cellspacing='0' cellpadding='0'")
    flip := true
    rows.each |Slot s|
    {
      shw := s.isPublic || s.isProtected
      cls := shw ? "rowOdd" : "internal"

      if (shw) out.tr
      else out.tr("class='hidden'")

      out.td("class='$cls col1'")
      col1.call2(s, out)
      out.tdEnd

      out.td("class='$cls' width='100%'")
      col2.call2(s, out)
      out.tdEnd

      out.td("class='$cls col4'")
      if (ftype != s.parent)
        writeTypeLink(s.parent)
      else
        out.w("&nbsp;") // b/c IE sucks
      out.tdEnd

      out.trEnd

      Str doc := s.doc
      Str docsum := (doc == null) ? "" : firstSentance(doc)
      more := doc != null && docsum.size < (doc.size-1)

      if (shw) out.tr("class='docsum'")
      else out.tr("class='docsum hidden'")

      out.td("class='rowEven'").w("&nbsp;").tdEnd
      out.td("class='rowEven' colspan='2'")
      if (doc == null)
        out.w("&nbsp;")
      else
      {
        out.div
        out.w(docsum).w(" ")
        if (more)
          out.a(`#`, "onclick='showFulldoc(event); return false;'").w("More").aEnd
        out.divEnd
        if (more)
        {
          out.div("class='fulldoc'")
          fantext(doc)
          out.divEnd
        }
      }
      out.tdEnd

      out.trEnd
    }
    out.tableEnd
    out.divEnd
  }

//////////////////////////////////////////////////////////////////////////
// Support Methods
//////////////////////////////////////////////////////////////////////////

  **
  ** Write a type link out in the form <a href='pod/type'>type.name</a>.
  **
  private Void writeTypeLink(Type t, Str attrs := null, Bool qname := false)
  {
    sig := t.signature

    if (!qname)
    {
      // Check for method sig
      i := sig.index("|")
      if (i != null)
      {
        j := sig.index("::")
        while (j != null)
        {
          k := sig.indexr(">", j)
          if (k == null) k = sig.indexr(",", j)
          if (k == null) k = sig.indexr("|", j)

          sig = sig[0..k] + sig[j+2..-1]
          j = sig.index("::")
        }
      }
      else
      {
        // Simply sig down to just type name
        sig = sig[sig.index("::")+2..-1]

        // Check if need to clean up more for map types
        i = sig.index("::")
        if (i != null)
          sig = sig[0..sig.index(":")] + sig[i+2..-2]
      }
    }

    out.a(typeUri(t), attrs).w(sig).aEnd
  }

  **
  ** Write out a method signature for this method.
  **
  private Void writeMethodSig(Method m)
  {
    if (m.isCtor) out.w("new")
    else
    {
      writeScope(m)
      writeTypeLink(m.returnType)
    }

    out.w(" $m.name(")
    p := m.params

    for (j:=0; j<p.size; j++)
    {
      if (j > 0) out.w(", ")
      writeTypeLink(p[j].paramType)
      out.w(" ${p[j].name}")
      if (p[j].hasDefault)
      out.w(" := def")
    }

    out.w(")")
  }

  **
  ** Write slot scope.
  **
  private Void writeScope(Slot s)
  {
    if (s.isStatic)    out.w("static ")
    if (s.isProtected) out.w("protected ")
    if (s.isPrivate)   out.w("private ")
    if (s.isInternal)  out.w("internal ")
  }

  **
  ** Return all the types that implement this type.
  **
  private Type[] findAllSubTypes(Type t)
  {
    acc := Type[,]
    lib := Sys.homeDir + `lib/fan/`
    lib.list.each |File f|
    {
      if (f.ext == "pod")
        findAllSubTypesInPod(acc, t, Pod.find(f.basename))
    }
    return acc
  }

  **
  ** Return all the types that implement this type in this pod.
  **
  private Void findAllSubTypesInPod(Type[] acc, Type t, Pod pod)
  {
    pod.types.each |Type x|
    {
      if (x.fits(t) && !x.isSynthetic && x.isPublic && t != x) acc.add(x)
    }
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  Pod pod
  Type ftype
  Method[] methods
  Field[] fields

}
