//
// Copyright (c) 2006, John Sublett
// Licensed under the Academic Free License version 3.0
//
// History:
//   29 Jun 06  John Sublett  Creation
//

**
** Node - This is a node in the dom for the HTML document.
**
enum NodeType
{
  doc,
  paragraph,
  text,
  header,
  lineBreak,
  codeBlock,
  inlineCode,
  hyperlink,
  orderedList,
  unorderedList,
  listItem
}

**
** The base class for all nodes in the tree.
**
abstract class Node
{
  new make()
  {
  }

  **
  ** Get the type for this node.
  **
  abstract NodeType nodeType()

  **
  ** Add a child node.
  **
  Void addChild(Node child)
  {
    if (children == null) children = Node[,]
    if (children.isEmpty || !children[-1].coalesce(child))
      children.add(child)
  }

  **
  ** Coalesce the specified node into this one.
  ** Returns true if the specified node
  ** was coalesced into this one, false
  ** otherwise.
  **
  virtual internal Bool coalesce(Node next)
  {
    return false
  }

  **
  ** Take the children from the specified parent.
  **
  Void takeChildren(Node parent)
  {
    children = parent.children
    parent.children = null
  }

  **
  ** Get a summary string for this node.
  **
  virtual Str summary() { return null }

  **
  ** Dump this node as a string.  This is NOT the method used
  ** to write the HTML.
  **
  override Str toStr()
  {
    return "[" + type().toStr() + "]:" + summary
  }

  **
  ** Write the HTML for this node recursively.  This
  ** node is responsible for writing itself and all
  ** of its children.
  **
  abstract Void write(OutStream out)

  **
  ** Dump the node to the specified output.  This is tree
  ** structure, NOT HTML.
  **
  virtual Void dump(OutStream out, Int indent)
  {
    typeName := type.name
    out.writeChars(Str.spaces(indent))
    out.writeChar('<');
    out.writeChars(typeName)
    summaryTxt := summary
    if (summary != null)
    {
      out.writeChars(": ")
      out.writeChars(summary)
    }

    if ((children != null) && (children.size != 0))
    {
      out.writeChar('>');
      children.each |Node child|
      {
        out.writeChar('\n')
        child.dump(out, indent + 2)
      }
      out.writeChars(">\n");
      out.writeChars(Str.spaces(indent))
      out.writeChars("</");
      out.writeChars(typeName)
      out.writeChar('>');
    }
    else
      out.writeChars("/>")
  }

  internal Node[] children;
}

//////////////////////////////////////////////////////////////////////////
// DocNode
//////////////////////////////////////////////////////////////////////////

**
** The root of the document tree.  Parsed
** fantext will have exactly one Doc.
**
class Doc : Node
{
  override NodeType nodeType() { return NodeType.doc }

  override Void write(OutStream out)
  {
    children.each |Node child| { child.write(out) }
  }
}

//////////////////////////////////////////////////////////////////////////
// ParagraphNode
//////////////////////////////////////////////////////////////////////////

**
** A paragraph.
**
class Paragraph : Node
{
  override NodeType nodeType() { return NodeType.paragraph }

  override Void write(OutStream out)
  {
    out.writeChars("<p>")
    children.each |Node child| { child.write(out) }
    out.writeChars("</p>")
  }
}

//////////////////////////////////////////////////////////////////////////
// CodeBlock
//////////////////////////////////////////////////////////////////////////

**
** A block of code that is not inlined in text.
**
class CodeBlock : Node
{
  new make(Str code)
  {
    this.code = code
  }

  override NodeType nodeType() { return NodeType.codeBlock }

  override Str summary() { return "len=" + code.size }

  override Void write(OutStream out)
  {
    out.writeChars("<pre><code>")
    out.writeChars(code)
    out.writeChars("</code></pre>")
  }

  Str code
}

//////////////////////////////////////////////////////////////////////////
// InlineCode
//////////////////////////////////////////////////////////////////////////

**
** Code that is inlined with other text.
**
class InlineCode : Node
{
  new make(Str code)
  {
    this.code = code
  }

  override NodeType nodeType() { return NodeType.inlineCode }

  override Str summary() { return "'" + code + "'" }

  override Void write(OutStream out)
  {
    out.writeChars("<code>")
    out.writeChars(code)
    out.writeChars("</code>")
  }

  Str code
}

//////////////////////////////////////////////////////////////////////////
// Header
//////////////////////////////////////////////////////////////////////////

**
** A header.
**
class Header : Node
{
  new make(Int level)
  {
    this.level = level
  }

  override NodeType nodeType() { return NodeType.header }

  override Str summary() { return level.toStr }

  override Void write(OutStream out)
  {
    out.writeChars("<h").writeChars(level.toStr).writeChar('>')
    children.each |Node child| { child.write(out) }
    out.writeChars("</h").writeChars(level.toStr).writeChar('>')
  }

  Int level
}
//////////////////////////////////////////////////////////////////////////
// Text
//////////////////////////////////////////////////////////////////////////

**
** A string of normal text.
**
class Text : Node
{
  new make(Str text) { this.text = text }

  override NodeType nodeType() { return NodeType.text }

  override Str summary() { return "'" + text + "'" }

  override Void write(OutStream out)
  {
    out.writeChars(text)
  }

  override Bool coalesce(Node next)
  {
    if (next.nodeType !== NodeType.text)
      return false
    else
    {
      text += ((Text)next).text
      return true
    }
  }

  Str text
}

//////////////////////////////////////////////////////////////////////////
// Break
//////////////////////////////////////////////////////////////////////////

**
** A line break.
**
class LineBreak : Node
{
  override NodeType nodeType() { return NodeType.lineBreak }

  override Void write(OutStream out)
  {
    out.writeChars("<br/>")
  }
}

//////////////////////////////////////////////////////////////////////////
// Hyperlink
//////////////////////////////////////////////////////////////////////////

**
** A hyperlink.
**
class Hyperlink : Node
{
  new make(Str text, Str url, Str title := null)
  {
    this.text = text
    this.url = url
    this.title = title
  }

  override NodeType nodeType() { return NodeType.hyperlink }

  override Str summary()
  {
    if (title == null)
      return text + "->" + url
    else
      return text + "->" + url + " (" + title + ")"
  }

  override Void write(OutStream out)
  {
    out.writeChars("<a href=\"").writeChars(url).writeChar('"')
    if (title != null) out.writeChars(" title=").writeChars(title).writeChar('"')
    out.writeChar('>').writeChars(text).writeChars("</a>")
  }

  Str text
  Str url
  Str title
}

//////////////////////////////////////////////////////////////////////////
// List
//////////////////////////////////////////////////////////////////////////

**
** A list with visible ordinal values for its items, typically numbered.
**
class OrderedList : Node
{
  override NodeType nodeType() { return NodeType.orderedList }

  override Void write(OutStream out)
  {
    out.writeChars("<ol>")
    children.each |Node child| { child.write(out) }
    out.writeChars("</ol>")
  }
}

**
** A list with no visible ordinal values for its items, typically bullets.
**
class UnorderedList : Node
{
  override NodeType nodeType() { return NodeType.unorderedList }

  override Void write(OutStream out)
  {
    out.writeChars("<ul>")
    children.each |Node child| { child.write(out) }
    out.writeChars("</ul>")
  }
}

//////////////////////////////////////////////////////////////////////////
// ListItem
//////////////////////////////////////////////////////////////////////////

**
** A single item in a list.
**
class ListItem : Node
{
  override NodeType nodeType() { return NodeType.listItem }

  override Void write(OutStream out)
  {
    out.writeChars("<li>")
    children.each |Node child| { child.write(out) }
    out.writeChars("</li>")
  }
}
