//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   17 Mar 06  Brian Frank  Creation
//
package fan.sys;

import java.lang.Thread;
import java.util.HashMap;
import java.util.Iterator;

/**
 * Map is a hashmap of key value pairs.
 */
public final class Map
  extends FanObj
{

//////////////////////////////////////////////////////////////////////////
// Constructors
//////////////////////////////////////////////////////////////////////////

  public static Map make(Type type)
  {
    return new Map((MapType)type, new HashMap());
  }

  public Map(Type k, Type v)
  {
    this(new MapType(k, v), new HashMap());
  }

  public Map(MapType type)
  {
    this(type, new HashMap());
  }

  public Map(MapType type, HashMap map)
  {
    if (type == null || map == null) { Thread.dumpStack(); throw new NullErr().val; }
    this.type = type;
    this.map  = map;
  }

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

  public final Type type()
  {
    return type;
  }

//////////////////////////////////////////////////////////////////////////
// Methods
//////////////////////////////////////////////////////////////////////////

  public final Bool isEmpty()
  {
    return map.size() == 0 ? Bool.True : Bool.False;
  }

  public final Int size()
  {
    return Int.pos(map.size());
  }

  public final Obj get(Obj key)
  {
    return (Obj)map.get(key);
  }

  public final Obj get(Obj key, Obj def)
  {
    Obj val = (Obj)map.get(key);
    if (val != null) return val;
    return def;
  }

  public final Bool containsKey(Obj key)
  {
    return Bool.make(map.containsKey(key));
  }

  public final List keys()
  {
    return new List(type.k, map.keySet());
  }

  public final List values()
  {
    return new List(type.v, map.values());
  }

  public final Map set(Obj key, Obj value)
  {
    modify();
    if (key != null && !key.isImmutable().val)
      throw NotImmutableErr.make("key is not immutable: " + key.type()).val;
    map.put(key, value);
    return this;
  }

  public final Map add(Obj key, Obj value)
  {
    modify();
    if (key != null && !key.isImmutable().val)
      throw NotImmutableErr.make("key is not immutable: " + key.type()).val;
    Object old = map.put(key, value);
    if (old != null)
    {
      map.put(key, old);
      throw ArgErr.make("Key already mapped: " + key).val;
    }
    return this;
  }

  public final Map setAll(Map m)
  {
    modify();
    Iterator it = m.map.keySet().iterator();
    while (it.hasNext())
    {
      Object key = it.next();
      Object val = m.map.get(key);
      map.put(key, val);
    }
    return this;
  }

  public final Map addAll(Map m)
  {
    modify();
    Iterator it = m.map.keySet().iterator();
    while (it.hasNext())
    {
      Obj key = (Obj)it.next();
      Obj val = (Obj)m.map.get(key);
      add(key, val);
    }
    return this;
  }

  public final Obj remove(Obj key)
  {
    modify();
    return (Obj)map.remove(key);
  }

  public final Map dup()
  {
    Map dup = new Map(type);
    dup.map = (HashMap)this.map.clone();
    return dup;
  }

  public final void clear()
  {
    modify();
    map.clear();
  }

  public final Bool equals(Obj that)
  {
    if (that instanceof Map)
    {
      return Bool.make(type.equals(that.type()).val &&
                       map.equals(((Map)that).map));
    }
    return Bool.False;
  }

  public final Int hash()
  {
    return Int.make(map.hashCode());
  }

  public final Str toStr()
  {
    if (map.size() == 0) return Str.make("[:]");
    StringBuilder s = new StringBuilder(32+map.size()*32);
    s.append("[");
    boolean first = true;
    Iterator it = map.keySet().iterator();
    while (it.hasNext())
    {
      Object key = it.next();
      Object val = map.get(key);
      if (!first) s.append(", ");
      else first = false;
      s.append(key).append(':').append(val);
    }
    s.append("]");
    return Str.make(s.toString());
  }

//////////////////////////////////////////////////////////////////////////
// Iterators
//////////////////////////////////////////////////////////////////////////

  public final void each(Method m)
  {
    Iterator it = map.keySet().iterator();
    while (it.hasNext())
    {
      Obj key = (Obj)it.next();
      Obj val = (Obj)map.get(key);
      m.call2(val, key);
    }
  }

  public final Obj find(Method m)
  {
    Iterator it = map.keySet().iterator();
    while (it.hasNext())
    {
      Obj key = (Obj)it.next();
      Obj val = (Obj)map.get(key);
      if (m.call2(val, key) == Bool.True)
        return val;
    }
    return null;
  }

  public final Map findAll(Method m)
  {
    Map acc = new Map(type);
    Iterator it = map.keySet().iterator();
    while (it.hasNext())
    {
      Obj key = (Obj)it.next();
      Obj val = (Obj)map.get(key);
      if (m.call2(val, key) == Bool.True)
        acc.set(key, val);
    }
    return acc;
  }

  public final Map exclude(Method m)
  {
    Map acc = new Map(type);
    Iterator it = map.keySet().iterator();
    while (it.hasNext())
    {
      Obj key = (Obj)it.next();
      Obj val = (Obj)map.get(key);
      if (m.call2(val, key) == Bool.False)
        acc.set(key, val);
    }
    return acc;
  }

  public final Obj reduce(Obj reduction, Method m)
  {
    Iterator it = map.keySet().iterator();
    while (it.hasNext())
    {
      Obj key = (Obj)it.next();
      Obj val = (Obj)map.get(key);
      reduction = m.call3(reduction, val, key);
    }
    return reduction;
  }

  public final Map map(Map acc, Method m)
  {
    Iterator it = map.keySet().iterator();
    while (it.hasNext())
    {
      Obj key = (Obj)it.next();
      Obj val = (Obj)map.get(key);
      acc.set(key, m.call2(val, key));
    }
    return acc;
  }

//////////////////////////////////////////////////////////////////////////
// Readonly
//////////////////////////////////////////////////////////////////////////

  public final Bool isRW()
  {
    return readonly ? Bool.False : Bool.True;
  }

  public final Bool isRO()
  {
    return readonly ? Bool.True : Bool.False;
  }

  public final Map rw()
  {
    if (!readonly) return this;

    Map rw = new Map(type);
    rw.map = (HashMap)map.clone();
    rw.readonly = false;
    rw.readonlyMap = this;
    return rw;
  }

  public final Map ro()
  {
    if (readonly) return this;
    if (readonlyMap == null)
    {
      Map ro = new Map(type);
      ro.map = map;
      ro.readonly = true;
      readonlyMap = ro;
    }
    return readonlyMap;
  }

  public final Bool isImmutable()
  {
    return Bool.make(immutable);
  }

  public final Map toImmutable()
  {
    if (immutable) return this;

    // make safe copy
    HashMap temp = new HashMap();
    Iterator it = map.keySet().iterator();
    while (it.hasNext())
    {
      Object key = it.next();

      Obj val = (Obj)map.get(key);
      if (val != null)
      {
        if (val instanceof List)
          val = ((List)val).toImmutable();
        else if (val instanceof Map)
          val = ((Map)val).toImmutable();
        else if (!val.isImmutable().val)
          throw NotImmutableErr.make("Item [" + key + "] not immutable " + val.type()).val;
      }

      temp.put(key, val);
    }

    // return new immutable map
    Map ro = new Map(type, temp);
    ro.readonly = true;
    ro.immutable = true;
    return ro;
  }

  private void modify()
  {
    // if readonly then throw readonly exception
    if (readonly)
      throw ReadonlyErr.make("Map is readonly").val;

    // if we have a cached readonlyMap, then detach
    // it so it remains immutable
    if (readonlyMap != null)
    {
      readonlyMap.map = (HashMap)map.clone();
      readonlyMap = null;
    }
  }


//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  private MapType type;
  private HashMap map;
  private boolean readonly;
  private boolean immutable;
  private Map readonlyMap;

}