//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   4 Jan 06  Brian Frank  Creation
//
package fan.sys;

import java.lang.reflect.*;
import fanx.fcode.*;

/**
 * Method is an invocable operation on a Type.
 */
public class Method
  extends Slot
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  /**
   * Constructor used by Type.reflect.
   */
  public Method(Type parent, Str name, int flags, Facets facets, Type returnType, Type inheritedReturnType, List params)
  {
    super(parent, name, flags, facets);
    this.returnType = returnType;
    this.inheritedReturnType = inheritedReturnType;
    this.params     = params;
    this.mask       = toMask(parent, returnType, params);
    this.generic    = null;
  }

  /**
   * Constructor used by GenericType and we are given the generic
   * method that is being parameterized.
   */
  public Method(Type parent, Str name, int flags, Facets facets, Type returnType, Type inheritedReturnType, List params, Method generic)
  {
    super(parent, name, flags, facets);
    this.returnType = returnType;
    this.inheritedReturnType = inheritedReturnType;
    this.params     = params;
    this.mask       = 0;
    this.generic    = generic;
  }

  /**
   * Compute if the method signature contains generic parameter types.
   */
  private static int toMask(Type parent, Type returnType, List params)
  {
    // we only use generics in Sys
    if (parent.pod() != Sys.SysPod) return 0;

    int p = returnType.isGenericParameter() ? 1 : 0;
    for (int i=0; i<params.sz(); ++i)
      p |= ((MethodParam)params.get(i)).paramType.isGenericParameter() ? 1 : 0;

    int mask = 0;
    if (p != 0) mask |= GENERIC;
    return mask;
  }

//////////////////////////////////////////////////////////////////////////
// Methods
//////////////////////////////////////////////////////////////////////////

  public Type type() { return Sys.MethodType; }

  public Type returnType() { return returnType; }

  public List params() { return params.ro(); }

  public Str signature()
  {
    StringBuilder s = new StringBuilder();
    s.append(returnType()).append(' ').append(name).append('(');
    for (int i=0; i<params.sz(); ++i)
    {
      if (i > 0) s.append(", ");
      MethodParam p = (MethodParam)params.get(i);
      s.append(p.paramType()).append(' ').append(p.name());
    }
    s.append(')');
    return Str.make(s.toString());
  }

  public Obj trap(Str name, List args)
  {
    // private undocumented access
    if (name.val.equals("inheritedReturnType"))
      return inheritedReturnType;
    else
      return super.trap(name, args);
  }

//////////////////////////////////////////////////////////////////////////
// Generics
//////////////////////////////////////////////////////////////////////////

  /**
   * Return if this method contains generic parameters in it's signature.
   */
  public boolean isGenericMethod()
  {
    return (mask & GENERIC) != 0;
  }

  /**
   * Return if this method is the parameterization of a generic method,
   * with all the generic parameters filled in with real types.
   */
  public boolean isGenericInstance()
  {
    return generic != null;
  }

  /**
   * If isGenericInstance is true, then return the generic method which
   * this method instantiates.  The generic method may be used to access
   * the actual signatures used in the Java code (via getRawType).  If
   * this method is not a generic instance, return null.
   */
  public Method getGenericMethod()
  {
    return generic;
  }

//////////////////////////////////////////////////////////////////////////
// Invoke
//////////////////////////////////////////////////////////////////////////

  public Obj call(List args)
  {
    int argsSize = args == null ? 0 : args.sz();

    boolean isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
    int p = checkArgs(argsSize, isStatic);
    Obj[] a = new Obj[p];

    if (isStatic)
    {
      if (args != null && a.length > 0) args.toArray(a, 0, a.length);
      return invoke(null, a);
    }
    else
    {
      Obj i = args.get(0);
      if (a.length > 0) args.toArray(a, 1, a.length);
      return invoke(i, a);
    }
  }

  public Obj callOn(Obj target, List args)
  {
    int argsSize = args == null ? 0 : args.sz();

    // from checkArgs perspective we pass true for isStatic
    // because we don't include target as part of arguments
    int p = checkArgs(argsSize, true);

    Obj[] a = new Obj[p];
    if (args != null && a.length > 0) args.toArray(a, 0, a.length);
    return invoke(target, a);
  }

  public Obj call0()
  {
    boolean isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
    checkArgs(0, isStatic);
    return invoke(null, noArgs);
  }

  public Obj call1(Obj a)
  {
    boolean isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
    int p = checkArgs(1, isStatic);
    Obj[] args = new Obj[p];
    if (isStatic)
    {
      switch (p)
      {
        case 1: args[0] = a;
      }
      return invoke(null, args);
    }
    else
    {
      return invoke(a, args);
    }
  }

  public Obj call2(Obj a, Obj b)
  {
    boolean isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
    int p = checkArgs(2, isStatic);
    Obj[] args = new Obj[p];
    if (isStatic)
    {
      switch (p)
      {
        case 2: args[1] = b;
        case 1: args[0] = a;
      }
      return invoke(null, args);
    }
    else
    {
      switch (p)
      {
        case 1: args[0] = b;
      }
      return invoke(a, args);
    }
  }

  public Obj call3(Obj a, Obj b, Obj c)
  {
    boolean isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
    int p = checkArgs(3, isStatic);
    Obj[] args = new Obj[p];
    if (isStatic)
    {
      switch (p)
      {
        case 3: args[2] = c;
        case 2: args[1] = b;
        case 1: args[0] = a;
      }
      return invoke(null, args);
    }
    else
    {
      switch (p)
      {
        case 2: args[1] = c;
        case 1: args[0] = b;
      }
      return invoke(a, args);
    }
  }

  public Obj call4(Obj a, Obj b, Obj c, Obj d)
  {
    boolean isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
    int p = checkArgs(4, isStatic);
    Obj[] args = new Obj[p];
    if (isStatic)
    {
      switch (p)
      {
        case 4: args[3] = d;
        case 3: args[2] = c;
        case 2: args[1] = b;
        case 1: args[0] = a;
      }
      return invoke(null, args);
    }
    else
    {
      switch (p)
      {
        case 3: args[2] = d;
        case 2: args[1] = c;
        case 1: args[0] = b;
      }
      return invoke(a, args);
    }
  }

  public Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e)
  {
    boolean isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
    int p = checkArgs(5, isStatic);
    Obj[] args = new Obj[p];
    if (isStatic)
    {
      switch (p)
      {
        case 5: args[4] = e;
        case 4: args[3] = d;
        case 3: args[2] = c;
        case 2: args[1] = b;
        case 1: args[0] = a;
      }
      return invoke(null, args);
    }
    else
    {
      switch (p)
      {
        case 4: args[3] = e;
        case 3: args[2] = d;
        case 2: args[1] = c;
        case 1: args[0] = b;
      }
      return invoke(a, args);
    }
  }

  public Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f)
  {
    boolean isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
    int p = checkArgs(6, isStatic);
    Obj[] args = new Obj[p];
    if (isStatic)
    {
      switch (p)
      {
        case 6: args[5] = f;
        case 5: args[4] = e;
        case 4: args[3] = d;
        case 3: args[2] = c;
        case 2: args[1] = b;
        case 1: args[0] = a;
      }
      return invoke(null, args);
    }
    else
    {
      switch (p)
      {
        case 5: args[4] = f;
        case 4: args[3] = e;
        case 3: args[2] = d;
        case 2: args[1] = c;
        case 1: args[0] = b;
      }
      return invoke(a, args);
    }
  }

  public Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g)
  {
    boolean isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
    int p = checkArgs(7, isStatic);
    Obj[] args = new Obj[p];
    if (isStatic)
    {
      switch (p)
      {
        case 7: args[6] = g;
        case 6: args[5] = f;
        case 5: args[4] = e;
        case 4: args[3] = d;
        case 3: args[2] = c;
        case 2: args[1] = b;
        case 1: args[0] = a;
      }
      return invoke(null, args);
    }
    else
    {
      switch (p)
      {
        case 6: args[5] = g;
        case 5: args[4] = f;
        case 4: args[3] = e;
        case 3: args[2] = d;
        case 2: args[1] = c;
        case 1: args[0] = b;
      }
      return invoke(a, args);
    }
  }

  public Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h)
  {
    boolean isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
    int p = checkArgs(8, isStatic);
    Obj[] args = new Obj[p];
    if (isStatic)
    {
      switch (p)
      {
        case 8: args[7] = h;
        case 7: args[6] = g;
        case 6: args[5] = f;
        case 5: args[4] = e;
        case 4: args[3] = d;
        case 3: args[2] = c;
        case 2: args[1] = b;
        case 1: args[0] = a;
      }
      return invoke(null, args);
    }
    else
    {
      switch (p)
      {
        case 7: args[6] = h;
        case 6: args[5] = g;
        case 5: args[4] = f;
        case 4: args[3] = e;
        case 3: args[2] = d;
        case 2: args[1] = c;
        case 1: args[0] = b;
      }
      return invoke(a, args);
    }
  }

  private int checkArgs(int args, boolean isStatic)
  {
    // ensure parent has finished emitting so that reflect is populated
    parent.finish();

    // compuate min/max parameters - reflect contains all the method versions
    // with full params at index zero, and full defaults at reflect.length-1
    int max = params.sz();
    int min = max-reflect.length+1;

    // do checking
    if (isStatic)
    {
      if (args < min) throw ArgErr.make("Too few arguments: " + args + " < " + min+".."+max).val;
    }
    else
    {
      if (args < min+1) throw ArgErr.make("Too few arguments: " + args + " < instance+" + min+".."+max).val;
    }

    // return size of arguments to pass to java method
    if (!isStatic) args--;
    if (args <= max) return args;
    return max;
  }

//////////////////////////////////////////////////////////////////////////
// Reflection
//////////////////////////////////////////////////////////////////////////

  Obj invoke(Object instance, Object[] args)
  {
    try
    {
      // zero index is full signature up to using max defaults
      int index = params.sz()-args.length;
      if (index < 0) index = 0;
      return (Obj)reflect[index].invoke(instance, (Object[])args);
    }
    catch (InvocationTargetException e)
    {
      if (e.getCause() instanceof Err.Val)
        throw (Err.Val)e.getCause();
      else
        throw Err.make(e.getCause()).val;
    }
    catch (Exception e)
    {
      if (reflect == null)
        throw Err.make("Method not mapped to java.lang.reflect correctly").val;

      //System.out.println("ERROR:      " + signature());
      //System.out.println("  instance: " + instance);
      //System.out.println("  args:     " + (args == null ? "null" : ""+args.length));
      //e.printStackTrace();

      throw Err.make("Cannot call '" + this + "': " + e).val;
    }
  }

//////////////////////////////////////////////////////////////////////////
// Indirect
//////////////////////////////////////////////////////////////////////////

  public static final int MaxIndirectParams = 8;  // max callX()

  /**
   * Constructor used by Indirect.
   */
  protected Method(MethodType methodType)
  {
    super(null, null, FConst.Internal, null);
    this.returnType = methodType.ret;
    this.inheritedReturnType = methodType.ret;
    this.params     = methodType.toMethodParams();
    this.mask       = 0;
    this.generic    = null;
  }

  /**
   * Indirect is the base class for the IndirectX classes, which are
   * used as the common base classes for closures and curried methods.
   * An Indirect method takes a MethodType for it's type, and
   * also extends Method with the call() implementation.
   */
  public static abstract class Indirect extends Method
  {
    protected Indirect(MethodType type) { super(type); this.type = type; }

    public Str  name()  { return Str.make(getClass().getName()); }
    public Type type()  { return type; }
    public Str  toStr() { return type.signature(); }
    public Err tooFewArgs(int given) { throw Err.make("Too few arguments: " + given + " < " + type.params.length).val; }

    MethodType type;
  }

  public static abstract class Indirect0 extends Indirect
  {
    protected Indirect0(MethodType type) { super(type); }
    public final Obj call(List args) { return call0(); }
    public abstract Obj call0();
    public final Obj call1(Obj a) { return call0(); }
    public final Obj call2(Obj a, Obj b) { return call0(); }
    public final Obj call3(Obj a, Obj b, Obj c) { return call0(); }
    public final Obj call4(Obj a, Obj b, Obj c, Obj d) { return call0(); }
    public final Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e) { return call0(); }
    public final Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { return call0(); }
    public final Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return call0(); }
    public final Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return call0(); }
  }

  public static abstract class Indirect1 extends Indirect
  {
    protected Indirect1(MethodType type) { super(type); }
    public final Obj call(List args) { return call1(args.get(0)); }
    public final Obj call0() { throw Err.make("Too few arguments: 0 < 1").val; }
    public abstract Obj call1(Obj a);
    public final Obj call2(Obj a, Obj b) { return call1(a); }
    public final Obj call3(Obj a, Obj b, Obj c) { return call1(a); }
    public final Obj call4(Obj a, Obj b, Obj c, Obj d) { return call1(a); }
    public final Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e) { return call1(a); }
    public final Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { return call1(a); }
    public final Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return call1(a); }
    public final Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return call1(a); }
  }

  public static abstract class Indirect2 extends Indirect
  {
    protected Indirect2(MethodType type) { super(type); }
    public final Obj call(List args) { return call2(args.get(0), args.get(1)); }
    public final Obj call0() { throw Err.make("Too few arguments: 0 < 2").val; }
    public final Obj call1(Obj a) { throw Err.make("Too few arguments: 1 < 2").val; }
    public abstract Obj call2(Obj a, Obj b);
    public final Obj call3(Obj a, Obj b, Obj c) { return call2(a, b); }
    public final Obj call4(Obj a, Obj b, Obj c, Obj d) { return call2(a, b); }
    public final Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e) { return call2(a, b); }
    public final Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { return call2(a, b); }
    public final Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return call2(a, b); }
    public final Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return call2(a, b); }
  }

  public static abstract class Indirect3 extends Indirect
  {
    protected Indirect3(MethodType type) { super(type); }
    public final Obj call(List args) { return call3(args.get(0), args.get(1), args.get(2)); }
    public final Obj call0() { throw Err.make("Too few arguments: 0 < 3").val; }
    public final Obj call1(Obj a) { throw Err.make("Too few arguments: 1 < 3").val; }
    public final Obj call2(Obj a, Obj b)  { throw Err.make("Too few arguments: 2 < 3").val; }
    public abstract Obj call3(Obj a, Obj b, Obj c);
    public final Obj call4(Obj a, Obj b, Obj c, Obj d) { return call3(a, b, c); }
    public final Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e) { return call3(a, b, c); }
    public final Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { return call3(a, b, c); }
    public final Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return call3(a, b, c); }
    public final Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return call3(a, b, c); }
  }

  public static abstract class Indirect4 extends Indirect
  {
    protected Indirect4(MethodType type) { super(type); }
    public final Obj call(List args) { return call4(args.get(0), args.get(1), args.get(2), args.get(3)); }
    public final Obj call0() { throw Err.make("Too few arguments: 0 < 4").val; }
    public final Obj call1(Obj a) { throw Err.make("Too few arguments: 1 < 4").val; }
    public final Obj call2(Obj a, Obj b)  { throw Err.make("Too few arguments: 2 < 4").val; }
    public final Obj call3(Obj a, Obj b, Obj c) { throw Err.make("Too few arguments: 3 < 4").val; }
    public abstract Obj call4(Obj a, Obj b, Obj c, Obj d);
    public final Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e) { return call4(a, b, c, d); }
    public final Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { return call4(a, b, c, d); }
    public final Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return call4(a, b, c, d); }
    public final Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return call4(a, b, c, d); }
  }

  public static abstract class Indirect5 extends Indirect
  {
    protected Indirect5(MethodType type) { super(type); }
    public final Obj call(List args) { return call5(args.get(0), args.get(1), args.get(2), args.get(3), args.get(4)); }
    public final Obj call0() { throw Err.make("Too few arguments: 0 < 5").val; }
    public final Obj call1(Obj a) { throw Err.make("Too few arguments: 1 < 5").val; }
    public final Obj call2(Obj a, Obj b)  { throw Err.make("Too few arguments: 2 < 5").val; }
    public final Obj call3(Obj a, Obj b, Obj c) { throw Err.make("Too few arguments: 3 < 5").val; }
    public final Obj call4(Obj a, Obj b, Obj c, Obj d) { throw Err.make("Too few arguments: 4 < 5").val; }
    public abstract Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e);
    public final Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { return call5(a, b, c, d, e); }
    public final Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return call5(a, b, c, d, e); }
    public final Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return call5(a, b, c, d, e); }
  }

  public static abstract class Indirect6 extends Indirect
  {
    protected Indirect6(MethodType type) { super(type); }
    public final Obj call(List args) { return call6(args.get(0), args.get(1), args.get(2), args.get(3), args.get(4), args.get(5)); }
    public final Obj call0() { throw Err.make("Too few arguments: 0 < 6").val; }
    public final Obj call1(Obj a) { throw Err.make("Too few arguments: 1 < 6").val; }
    public final Obj call2(Obj a, Obj b)  { throw Err.make("Too few arguments: 2 < 6").val; }
    public final Obj call3(Obj a, Obj b, Obj c) { throw Err.make("Too few arguments: 3 < 6").val; }
    public final Obj call4(Obj a, Obj b, Obj c, Obj d) { throw Err.make("Too few arguments: 4 < 6").val; }
    public final Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e) { throw Err.make("Too few arguments: 5 < 6").val; }
    public abstract Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f);
    public final Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return call6(a, b, c, d, e, f); }
    public final Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return call6(a, b, c, d, e, f); }
  }

  public static abstract class Indirect7 extends Indirect
  {
    protected Indirect7(MethodType type) { super(type); }
    public final Obj call(List args) { return call7(args.get(0), args.get(1), args.get(2), args.get(3), args.get(4), args.get(5), args.get(6)); }
    public final Obj call0() { throw Err.make("Too few arguments: 0 < 7").val; }
    public final Obj call1(Obj a) { throw Err.make("Too few arguments: 1 < 7").val; }
    public final Obj call2(Obj a, Obj b)  { throw Err.make("Too few arguments: 2 < 7").val; }
    public final Obj call3(Obj a, Obj b, Obj c) { throw Err.make("Too few arguments: 3 < 7").val; }
    public final Obj call4(Obj a, Obj b, Obj c, Obj d) { throw Err.make("Too few arguments: 4 < 7").val; }
    public final Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e) { throw Err.make("Too few arguments: 5 < 7").val; }
    public final Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { throw Err.make("Too few arguments: 6 < 7").val; }
    public abstract Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g);
    public final Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return call7(a, b, c, d, e, f, g); }
  }

  public static abstract class Indirect8 extends Indirect
  {
    protected Indirect8(MethodType type) { super(type); }
    public final Obj call(List args) { return call8(args.get(0), args.get(1), args.get(2), args.get(3), args.get(4), args.get(5), args.get(6), args.get(7)); }
    public final Obj call0() { throw Err.make("Too few arguments: 0 < 8").val; }
    public final Obj call1(Obj a) { throw Err.make("Too few arguments: 1 < 8").val; }
    public final Obj call2(Obj a, Obj b)  { throw Err.make("Too few arguments: 2 < 8").val; }
    public final Obj call3(Obj a, Obj b, Obj c) { throw Err.make("Too few arguments: 3 < 8").val; }
    public final Obj call4(Obj a, Obj b, Obj c, Obj d) { throw Err.make("Too few arguments: 4 < 8").val; }
    public final Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e) { throw Err.make("Too few arguments: 5 < 8").val; }
    public final Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { throw Err.make("Too few arguments: 6 < 8").val; }
    public final Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { throw Err.make("Too few arguments: 7 < 8").val; }
    public abstract Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h);
  }

  public static abstract class IndirectX extends Indirect
  {
    protected IndirectX(MethodType type) { super(type); }
    public abstract Obj call(List args);
    public final Obj call0() { throw tooFewArgs(0).val; }
    public final Obj call1(Obj a) { throw tooFewArgs(1).val; }
    public final Obj call2(Obj a, Obj b)  { throw tooFewArgs(2).val; }
    public final Obj call3(Obj a, Obj b, Obj c) { throw tooFewArgs(3).val; }
    public final Obj call4(Obj a, Obj b, Obj c, Obj d) { throw tooFewArgs(4).val; }
    public final Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e) { throw tooFewArgs(5).val; }
    public final Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { throw tooFewArgs(6).val; }
    public final Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { throw tooFewArgs(7).val; }
    public final Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { throw tooFewArgs(8).val; }
  }

//////////////////////////////////////////////////////////////////////////
// Compiler Support
//////////////////////////////////////////////////////////////////////////

  public MethodParam[] paramArr()
  {
    return (MethodParam[])params.toArray(new MethodParam[params.sz()]);
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  static final int GENERIC = 0x01;  // is this a generic method
  static final Object[] noArgs = new Object[0];

  final Type returnType;
  final Type inheritedReturnType; // for covariance
  final List params;
  final int mask;
  final Method generic;
  java.lang.reflect.Method[] reflect;

}