//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   4 Dec 05  Brian Frank  Creation
//
package fan.sys;

import java.util.HashMap;

/**
 * Thread models a thread of execution within a process.
 */
public class Thread
  extends FanObj
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public static Thread make() { return make(null, null); }
  public static Thread make(Str name) { return make(name, null); }
  public static Thread make(Str name, Method run)
  {
    Thread t = new Thread();
    make$(t, name, run);
    return t;
  }

  public static void make$(Thread t) { make$(t, null, null); }
  public static void make$(Thread t, Str name) { make$(t, name, null); }
  public static void make$(Thread t, Str name, Method run)
  {
    synchronized (topLock)
    {
      // check run method
      if (run != null && !run.isConst().val)
        throw NotImmutableErr.make("Run method not const: " + run).val;

      // auto generate name if null
      if (name == null)
        name = Str.make(t.type().pod().name() + "." + t.type().name() + "." + (autoNameCount++));

      // verify unique
      if (byName.get(name) != null)
        throw ArgErr.make("Duplicate thread name: " + name).val;

      // init and put into map
      t.name  = name;
      t.run   = run;
      t.state = NEW;
      byName.put(name, t);
    }
  }

  public Thread(String name)
  {
    this.name  = Str.make(name);
    this.state = NEW;
  }

  public Thread()
  {
  }

//////////////////////////////////////////////////////////////////////////
// Management
//////////////////////////////////////////////////////////////////////////

  public static Thread find(Str name) { return find(name, Bool.True); }
  public static Thread find(Str name, Bool checked)
  {
    synchronized (topLock)
    {
      Thread thread = (Thread)byName.get(name);
      if (thread != null) return thread;
      if (checked.val) throw Err.make("Unknown thread: " + name).val;
      return null;
    }
  }

  public static List list()
  {
    synchronized (topLock)
    {
      return new List(Sys.ThreadType, (Thread[])byName.values().toArray(new Thread[byName.size()])).ro();
    }
  }

  public static Thread current()
  {
    try
    {
      return ((Val)java.lang.Thread.currentThread()).thread;
    }
    catch (ClassCastException e)
    {
      throw Err.make("Current thread not a Fan thread: " + java.lang.Thread.currentThread()).val;
    }
  }

  public static Map locals() { return (Map)locals.get(); }
  private static final ThreadLocal locals = new ThreadLocal()
  {
    protected Object initialValue()
    {
      return new Map(Sys.StrType, Sys.ObjType);
    }
  };

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

  public final Bool equals(Obj obj)
  {
    return this == obj ? Bool.True : Bool.False;
  }

  public final int hashCode()
  {
    return name.hashCode();
  }

  public final Int hash()
  {
    return name.hash();
  }

  public Str toStr()
  {
    return name;
  }

  public Type type()
  {
    return Sys.ThreadType;
  }

  public final Str name()
  {
    return name;
  }

  public final void trace() { trace(Sys.out()); }
  public final void trace(OutStream out)
  {
    java.lang.Thread t = this.val;
    if (t == null) return;

    StackTraceElement[] trace = t.getStackTrace();
    for (int i=2; i<trace.length; ++i)
      out.printLine(Str.make(trace[i].toString()));
  }

//////////////////////////////////////////////////////////////////////////
// State
//////////////////////////////////////////////////////////////////////////

  public final synchronized Bool isNew()
  {
    return Bool.make(state == NEW);
  }

  public final synchronized Bool isRunning()
  {
    return Bool.make(state == RUNNING);
  }

  public final synchronized Bool isDead()
  {
    return Bool.make(state == DEAD);
  }

//////////////////////////////////////////////////////////////////////////
// Lifecycle
//////////////////////////////////////////////////////////////////////////

  public final synchronized Thread start()
  {
    if (state != NEW) throw Err.make("Invalid state for Thread.start").val;
    state = RUNNING;
    val = alloc();
    val.attach(this);
    return this;
  }

  public final synchronized void stop()
  {
    if (state == DEAD) return;
    state = DEAD;
    synchronized (topLock) { byName.remove(name); }
    stopMessages();
    notifyAll();
    if (val != null) val.interrupt();
  }

  public final Obj join() { return join(null); }
  public final synchronized Obj join(Duration duration)
  {
    try
    {
      if (state == NEW) throw Err.make("Thread not started yet").val;
      if (state == DEAD) return runResult;

      if (duration != null)
        wait(duration.millis());
      else
        while (state == RUNNING) wait();

      return runResult;
    }
    catch (InterruptedException e)
    {
      throw InterruptedErr.make(e).val;
    }
  }

  public static void sleep(Duration duration)
  {
    try
    {
      long ticks = duration.ticks;
      java.lang.Thread.sleep(ticks/1000000L, (int)(ticks%1000000L));
    }
    catch (InterruptedException e)
    {
      throw InterruptedErr.make(e).val;
    }
  }

//////////////////////////////////////////////////////////////////////////
// Thread Pool
//////////////////////////////////////////////////////////////////////////

  /**
   * Allocate a Java thread to use.  Reuse a lingering thread
   * in the pool if one is available, otherwise create a new one.
   */
  static Val alloc()
  {
    synchronized (topLock)
    {
      if (pool != null)
      {
        Val val = pool;
        pool = val.next;
        return val;
      }
      else
      {
        return new Val("Idle:" + (totalCount++));
      }
    }
  }

  /**
   * Add a lingering thread to the thread pool.
   */
  static void linger(Val val)
  {
    synchronized (topLock)
    {
      val.next = pool;
      pool = val;
    }
  }

  /**
   * Free a thread from the thread pool whose linger timeout has expired.
   */
  static void free(Val val)
  {
    synchronized (topLock)
    {
      if (pool == val)
      {
        pool = val.next;
      }
      else
      {
        for (Val p = pool; p != null; p = p.next)
          if (p.next == val) { p.next = p.next.next; break; }
      }
    }
  }

//////////////////////////////////////////////////////////////////////////
// Val
//////////////////////////////////////////////////////////////////////////

  /**
   * Val is the Java thread subclass attached to Fan threads.
   */
  static class Val extends java.lang.Thread
  {
    Val(String name)
    {
      super(name);
      this.unattachedName = name;
      setDaemon(false);
    }

    synchronized void attach(Thread thread)
    {
      this.thread = thread;
      thread.val = this;
      setName(thread.name.val);
      if (isAlive())
        notify();
      else
        start();
    }

    synchronized void detach()
    {
      thread.val = null;
      thread = null;
      setName(unattachedName);
    }

    public void run()
    {
      while (true)
      {
        // run attached thread
        Thread attached = this.thread;
        try
        {
          Obj result = attached.run();

          if (result != null && result.isImmutable().val)
            attached.runResult = result;
        }
        catch (Throwable e)
        {
          e.printStackTrace();
        }

        // attached thread is now dead
        detach();
        attached.stop();

        synchronized (this)
        {
          // add to thread pool
          linger(this);

          // linger for 5sec
          try { wait(5000); } catch (InterruptedException e) {}

          // if we have a thread to process, the do it again!
          if (thread != null) continue;

          // free from the pool and die
          free(this);
          return;
        }
      }
    }

    String unattachedName;  // default thread name
    Thread thread;          // Fan thread
    Val next;               // thread pool linked list
  }

//////////////////////////////////////////////////////////////////////////
// Run
//////////////////////////////////////////////////////////////////////////

  public Obj run()
    throws InterruptedException
  {
    if (run != null)
    {
      return run.call1(this);
    }
    else
    {
      System.out.println("WARNING: Thread.run not overridden");
      return null;
    }
  }

  public final void loop(Method received)
  {
    // ensure only called by myself
    if (current() != this)
      throw Err.make("Thread.current != this").val;

    // ensure received not null
    if (received == null)
      throw NullErr.make("received callback null").val;

    // main loop
    while (state == RUNNING)
    {
      try
      {
        dispatch(received, dequeue());
      }
      catch (Throwable e)
      {
        if (state == RUNNING) e.printStackTrace();
      }
    }
  }

  private void dispatch(Method received, Message msg)
  {
    try
    {
      Obj result = received.call1(msg.obj);

      if (result != null && !result.isImmutable().val)
        throw NotImmutableErr.make("received result").val;

      msg.finish(MSG_FINISH_OK, result);
    }
    catch (Err.Val e)
    {
      msg.finish(MSG_FINISH_ERR, e.err);
    }
    catch (Throwable e)
    {
      msg.finish(MSG_FINISH_ERR, Err.make(e));
    }
  }

//////////////////////////////////////////////////////////////////////////
// Messaging
//////////////////////////////////////////////////////////////////////////

  public final Obj sendSync(Obj obj)
  {
    if (!obj.isImmutable().val)
      throw NotImmutableErr.make("cannot send mutable msg").val;

    try
    {
      Message msg = new Message(MSG_SYNC, obj);
      enqueue(msg);
      return msg.waitUntilFinished();
    }
    catch(InterruptedException e)
    {
      throw InterruptedErr.make(e).val;
    }
  }

  public final void sendAsync(Obj obj)
  {
    if (!obj.isImmutable().val)
      throw NotImmutableErr.make("cannot send mutable msg").val;

    try
    {
      enqueue(new Message(MSG_ASYNC, obj));
    }
    catch(InterruptedException e)
    {
      throw InterruptedErr.make(e).val;
    }
  }

//////////////////////////////////////////////////////////////////////////
// Queue
//////////////////////////////////////////////////////////////////////////

  synchronized void stopMessages()
  {
    for (Message m = head; m != null; m = m.next)
      m.finish(MSG_STOPPED, null);
    head = tail = null;
    size = 0;
  }

  synchronized Message dequeue()
    throws InterruptedException
  {
    // wait until we have a message
    while (state == RUNNING && size == 0)
      wait();

    // read message of head of linked list
    Message m = head;
    head = m.next;
    if (head == null) tail = null;
    m.next = null;
    size--;

    // return message
    return m;
  }

  synchronized void enqueue(Message m)
    throws InterruptedException
  {
    // ensure new or running
    if (state > RUNNING)
      throw Err.make("thread not active").val;

    // flow control wait if at max
    if (size >= maxQueueSize) wait();

    // add to tail of linked list
    if (tail == null) { head = tail = m; m.next = null; }
    else { tail.next = m; tail = m; }
    size++;
    if (size > peek) peek = size;

    // notify get thread
    notify();
  }

//////////////////////////////////////////////////////////////////////////
// Message
//////////////////////////////////////////////////////////////////////////

  static class Message
  {
    Message(int state, Obj obj)
    {
      this.state = state;
      this.obj = obj;
    }

    synchronized Obj waitUntilFinished()
      throws InterruptedException
    {
      if (state == MSG_SYNC) wait();

      switch (state)
      {
        case MSG_SYNC:       throw InterruptedErr.make("sendSync timed out").val;
        case MSG_FINISH_OK:  return obj;
        case MSG_FINISH_ERR: throw ((Err)obj).val;
        case MSG_STOPPED:    throw InterruptedErr.make("thread stopped").val;
        default: throw new IllegalStateException(""+state);
      }
    }

    void finish(int newState, Obj obj)
    {
      if (this.state == MSG_ASYNC)
      {
        if (newState == MSG_FINISH_ERR)
          ((Err)obj).trace();
      }
      else
      {
        synchronized (this)
        {
          this.state = newState;
          this.obj = obj;
          notify();
        }
      }
    }

    int state;       // sync/async in, ok/err out
    Obj obj;         // message in, return/err out
    Message next;    // queue linked list
  }

  static final int MSG_ASYNC      = 0;
  static final int MSG_SYNC       = 1;
  static final int MSG_FINISH_OK  = 2;
  static final int MSG_FINISH_ERR = 3;
  static final int MSG_STOPPED    = 4;

//////////////////////////////////////////////////////////////////////////
// State Constants
//////////////////////////////////////////////////////////////////////////

  static final int NEW     = 0;
  static final int RUNNING = 1;
  static final int DEAD    = 2;

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  private static Object topLock = new Object();   // top level lock
  private static HashMap byName = new HashMap();  // String -> Thread
  private static int autoNameCount = 0;           // auto-generate unique name
  private static int totalCount = 0;              // Val unique name
  private static Val pool;                        // thread pool linked list
  private static int maxQueueSize = 1000;         // max messages to queue

  private Str name;            // thread name
  private int state;           // current state
  private Val val;             // Java thread if attached
  private Message head, tail;  // message queue linked list
  private int size, peek;      // message queue size
  private Method run;          // run method
  private Obj runResult;       // return of run method

}