//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.asm;

import java.io.*;
import java.util.*;
import fan.sys.*;
import fanx.fcode.*;
import fanx.util.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.ast.*;

/**
 * Assembler assembles all the ClassDefs into their fcode representation.
 */
public class Assembler
  extends CompilerSupport
  implements FConst
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public Assembler(Compiler compiler)
  {
    super(compiler);
  }

//////////////////////////////////////////////////////////////////////////
// Assemble
//////////////////////////////////////////////////////////////////////////

  public FPod assemblePod()
  {
    fpod = new FPod(compiler.pod.name().val, null);
    fpod.initForCompile();

    ClassDef[] defs = compiler.classDefs;
    fpod.types = new FType[defs.length];
    for (int i=0; i<defs.length; ++i)
      fpod.types[i] = assembleType(defs[i]);

    return fpod;
  }

  private FType assembleType(ClassDef def)
  {
    FType t = new FType(fpod);
    t.hollow     = false;
    t.flags      = def.flags;
    t.self       = typeRef(def.type);
    t.base       = (def.base == null) ? -1 : typeRef(def.base.type);
    t.mixins = new int[def.mixins.size()];
    for (int i=0; i<t.mixins.length; ++i)
      t.mixins[i] = typeRef(((TypeRef)def.mixins.get(i)).type);

    FieldDef[] fields = def.fields();
    t.fields = new FField[fields.length];
    for (int i=0; i<fields.length; ++i)
      t.fields[i] = assembleField(fields[i]);

    MethodDef[] methods = def.methods();
    t.methods = new FMethod[methods.length];
    for (int i=0; i<methods.length; ++i)
      t.methods[i] = assembleMethod(methods[i]);

    FacetAsm facets = new FacetAsm(fpod);
    if (compiler.pod.isScript)
      facets.sourceFile(""+def.location.file());
    else
      facets.sourceFile(Location.filename(def.location));
    facets.lineNumber(def.line());
    t.facets = facets.list();

    return t;
  }

  private FField assembleField(FieldDef def)
  {
    FField f = new FField();
    f.name  = name(def.name);
    f.flags = def.flags | FConst.Storage;
    f.type  = typeRef(def.type.type);

    FacetAsm facets = new FacetAsm(fpod);
    facets.lineNumber(def.line());
    f.facets = facets.list();

    return f;
  }

  private FMethod assembleMethod(MethodDef def)
  {
    FacetAsm facets = new FacetAsm(fpod);

    FMethod m = new FMethod();
    m.name         = name(def.name);
    m.flags        = def.flags;
    m.ret          = typeRef(def.ret.type);
    m.inheritedRet = m.ret;
    m.paramCount   = def.params.size();
    m.localCount   = def.vars.size() - def.params.size();
    m.vars         = new FMethodVar[def.vars.size()];

    for (int i=0; i<m.vars.length; ++i)
    {
      MethodVar v = (MethodVar)def.vars.get(i);
      FMethodVar f = m.vars[i] = new FMethodVar();
      f.name  = name(v.name);
      f.type  = typeRef(v.type);
      f.flags = v.flags;
      if (v.paramDef != null) f.def = code(v.paramDef.def);
    }

    m.code = code(def, facets);

    facets.lineNumber(def.line());
    m.facets = facets.list();

    return m;
  }

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  private int typeRef(Type type)
  {
    return fpod.typeRef(type);
  }

  private int name(String val)
  {
    return fpod.name(val);
  }

  private FBuf code(MethodDef def, FacetAsm facets)
  {
    Block block = def.code;
    if (block == null) return null;

    CodeAsm asm = new CodeAsm(compiler, fpod);
    if (def.ctorChain != null) asm.expr(def.ctorChain);
    asm.block(block);

    if (asm.errCount > 0) facets.add(ErrTable, 0, asm.errTable);
    if (asm.lineCount > 0) facets.add(LineNumbers, 0, asm.lines);

    return new FBuf(asm.code);
  }

  private FBuf code(Expr expr)
  {
    if (expr == null) return null;
    CodeAsm asm = new CodeAsm(compiler, fpod);
    asm.expr(expr);
    return new FBuf(asm.code);
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  FPod fpod;
}