//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.asm;

import java.util.*;
import fan.sys.*;
import fanx.fcode.*;
import fanx.util.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.ast.*;
import fanx.compiler.parser.*;
import fan.sys.*;
import fan.sys.Float;

/**
 * CodeAsm is used to assemble the fcode instructions of an Expr or Block.
 *   Expr  -> FCode
 *   Block -> FCode
 */
public class CodeAsm
  extends CompilerSupport
  implements FConst
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public CodeAsm(Compiler compiler, FPod pod)
  {
    super(compiler);
    this.pod = pod;
    this.code     = new Box();
    this.errTable = new Box(); errTable.u2(-1);
    this.lines    = new Box(); lines.u2(-1);
  }

//////////////////////////////////////////////////////////////////////////
// Statements
//////////////////////////////////////////////////////////////////////////

  public void block(Block block)
  {
    for (int i=0; i<block.statements.size(); ++i)
      stmt((Stmt)block.statements.get(i));
  }

  public void stmt(Stmt stmt)
  {
    switch (stmt.id())
    {
      case Stmt.NOP:           break;
      case Stmt.EXPR:          expr(((ExprStmt)stmt).expr); break;
      case Stmt.LOCAL_VAR_DEF: localVarDefStmt((LocalVarDefStmt)stmt); break;
      case Stmt.IF:            ifStmt((IfStmt)stmt); break;
      case Stmt.RETURN:        returnStmt((ReturnStmt)stmt); break;
      case Stmt.THROW:         throwStmt((ThrowStmt)stmt); break;
      case Stmt.FOR:           forStmt((ForStmt)stmt); break;
      case Stmt.WHILE:         whileStmt((WhileStmt)stmt); break;
      case Stmt.BREAK:         breakStmt((BreakStmt)stmt); break;
      case Stmt.CONTINUE:      continueStmt((ContinueStmt)stmt); break;
      case Stmt.SWITCH:        switchStmt((SwitchStmt)stmt); break;
      case Stmt.TRY:           tryStmt((TryStmt)stmt); break;
      default:                 throw new IllegalStateException(stmt.getClass().getName());
    }
  }

  private void ifStmt(IfStmt stmt)
  {
    int falseLabel, endLabel = -1;

    // TODO, often the condition is a boolean expression, that pushes
    // an explicit 1 or 0 onto the stack - we can optimze that to use
    // the result for the branch directly

    expr(stmt.condition);
    falseLabel = jump(JumpFalse);
    block(stmt.trueBlock);
    if (!stmt.trueBlock.lastIsMethodExit())
      endLabel = jump(Jump);
    mark(falseLabel);
    if (stmt.falseBlock != null)
      block(stmt.falseBlock);
    if (endLabel != -1) mark(endLabel);
  }

  private void returnStmt(ReturnStmt stmt)
  {
    if (stmt.expr != null)
    {
      expr(stmt.expr);
      op(ReturnObj);
    }
    else
    {
      op(ReturnVoid);
    }
  }

  private void throwStmt(ThrowStmt stmt)
  {
    expr(stmt.exception);
    op(Throw);
  }

  private void whileStmt(WhileStmt stmt)
  {
    // push myself onto the loop stack so that breaks
    // and continues can register for backpatching
    Loop loop = loopStack().push(stmt);

    // assemble the while loop code
    int continueLabel = mark();
    expr(stmt.condition);
    int breakJump = jump(JumpFalse);
    block(stmt.block);
    jump(Jump, continueLabel);
    int breakLabel = mark();
    mark(breakJump);

    // backpatch continues
    for (int i=0; i<loop.continues.size(); ++i)
      mark(((Integer)loop.continues.get(i)).intValue(), continueLabel);

    // backpatch breaks
    for (int i=0; i<loop.breaks.size(); ++i)
      mark(((Integer)loop.breaks.get(i)).intValue(), breakLabel);

    // TODO - the fcode will often contain Jumps to Jumps which can be optimized
  }

  private void forStmt(ForStmt stmt)
  {
    int breakJump = -1;

    // push myself onto the loop stack so that breaks
    // and continues can register for backpatching
    Loop loop = loopStack().push(stmt);

    // assemble init if available
    if (stmt.init != null) stmt(stmt.init);

    // assemble the for loop code
    int condLabel = mark();
    if (stmt.condition != null)
    {
      expr(stmt.condition);
      breakJump = jump(JumpFalse);
    }
    block(stmt.block);
    int updateLabel = mark();
    if (stmt.update != null) expr(stmt.update);
    jump(Jump, condLabel);
    int endLabel = mark();
    if (breakJump != -1) mark(breakJump, endLabel);

    // backpatch continues
    for (int i=0; i<loop.continues.size(); ++i)
      mark(((Integer)loop.continues.get(i)).intValue(), updateLabel);

    // backpatch breaks
    for (int i=0; i<loop.breaks.size(); ++i)
      mark(((Integer)loop.breaks.get(i)).intValue(), endLabel);

    // TODO - the fcode will often contain Jumps to Jumps which can be optimized
  }

  private void breakStmt(BreakStmt stmt)
  {
    // jump and register with Loop for backpatch
    int j = jump(Jump);
    loopStack().findLoop(stmt.loop).breaks.add(new Integer(j));
  }

  private void continueStmt(ContinueStmt stmt)
  {
    // jump and register with Loop for backpatch
    int j = jump(Jump);
    loopStack().findLoop(stmt.loop).continues.add(new Integer(j));
  }

  private void localVarDefStmt(LocalVarDefStmt stmt)
  {
    if (stmt.catchVar)
    {
      op(CatchErrStart, pod.typeRef(stmt.type.type));
      op(StoreVar, stmt.var.register);
    }
    else if (stmt.init != null)
    {
      expr(stmt.init);
      op(StoreVar, stmt.var.register);
    }
  }

//////////////////////////////////////////////////////////////////////////
// Switch
//////////////////////////////////////////////////////////////////////////

  private void switchStmt(SwitchStmt stmt)
  {
   // A table switch is a series of contiguous (or near contiguous)
   // cases which can be represented an offset into a jump table.
   if (stmt.computeTableRange())
      tableSwitchStmt(stmt);
    else
      throw err("non-table switches not supported yet", stmt);
  }

  private void tableSwitchStmt(SwitchStmt stmt)
  {
    int min = stmt.min;
    int max = stmt.max;
    boolean isEnum = stmt.condition.type.isEnum().val;

    // push condition onto the stack
    expr(stmt.condition);

    // if this is an enum get ordinal on the stack
    if (isEnum)
      op(CallVirtual, pod.methodRef(EnumOrdinal));

    // if min is not zero, then do a subtraction so that
    // our condition is a zero based index into the jump table
    if (min != 0)
    {
      op(LoadInt, pod.integer(Int.make(-min)));
      op(CallVirtual, pod.methodRef(IntPlus));
    }

    // now allocate our jump table
    int count = max-min+1;
    SwitchCase[] jumps = new SwitchCase[count];

    // walk thru each case, and map the jump offset to a block
    for (int i=0; i<stmt.cases.size(); ++i)
    {
      SwitchCase c = (SwitchCase)stmt.cases.get(i);
      for (int j=0; j<c.cases.size(); ++j)
      {
        Expr expr     = (Expr)c.cases.get(j);
        int literal   = (int)expr.asTableSwitchCase().val;
        int offset    = literal-min;
        jumps[offset] = c;
      }
    }

    // now write the bytecode
    op(Switch);
    code.u2(count);
    int jumpStart = code.len;
    code.skip(count*2);  // we'll backpatch the jump offsets last

    // default block goes first - it's the switch fall
    // thru, save offset to back patch jump
    int defaultStart = mark();
    int defaultEnd = switchBlock(stmt.defaultBlock);

    // now write each case block
    int[] caseEnds = new int[stmt.cases.size()];
    for (int i=0; i<stmt.cases.size(); ++i)
    {
      SwitchCase c = (SwitchCase)stmt.cases.get(i);
      c.startOffset = code.len;
      caseEnds[i] = switchBlock(c.block);
    }

    // backpatch the jump table
    int end = code.len;
    for (int i=0; i<jumps.length; ++i)
    {
      int backpatch = jumpStart + i*2;
      int jumpTo = jumps[i] == null ? defaultStart : jumps[i].startOffset;
      code.u2(backpatch, jumpTo);
    }

    // backpatch all the case blocks to jump here when done
    if (defaultEnd != -1) mark(defaultEnd);
    for (int i=0; i<caseEnds.length; ++i)
      if (caseEnds[i] != -1) mark(caseEnds[i]);
  }

  private int switchBlock(Block block)
  {
    if (block != null)
    {
      block(block);
      if (block.lastIsMethodExit()) return -1;
    }
    return jump(Jump);
  }

//////////////////////////////////////////////////////////////////////////
// Try
//////////////////////////////////////////////////////////////////////////

  private void tryStmt(TryStmt stmt)
  {
    if (stmt.finallyBlock != null)
      throw err("Finally not implemented yet (it's a tough one!", stmt);

    // assemble body of try block
    int start = code.len;
    block(stmt.block);
    int end = code.len;

    // jump over catch blocks
    int tryDone = -1;
    if (!stmt.block.lastIsMethodExit()) tryDone = jump(Jump);

    // assemble catch blocks
    int[] catchDones = new int[stmt.catches.size()];
    for (int i=0; i<stmt.catches.size(); ++i)
    {
      TryCatch c = (TryCatch)stmt.catches.get(i);
      catchDones[i] = tryCatch(c, start, end);
    }

    // mark next statement as jump destination for try block
    mark(tryDone);
    for (int i=0; i<catchDones.length; ++i) mark(catchDones[i]);
  }

  private int tryCatch(TryCatch c, int start, int end)
  {
    // assemble catch block
    int handler = code.len;
    if (c.errVariable == null) op(CatchAllStart);
    block(c.block);
    int done = -1;
    if (!c.block.lastIsMethodExit()) done = jump(Jump);

    // catch all is implicitly a catch for sys::Err
    Type type = c.errType != null ? c.errType.type : Sys.ErrType;

    // fill in err table
    errTable.u2(0, ++errCount);
    errTable.u2(start);
    errTable.u2(end);
    errTable.u2(handler);
    errTable.u2(pod.typeRef(type));

    // return location to back patch
    return done;
  }

//////////////////////////////////////////////////////////////////////////
// Expressions
//////////////////////////////////////////////////////////////////////////

  public void expr(Expr expr)
  {
    switch (expr.id())
    {
      case Expr.NULL_LITERAL:     nullLiteral(); break;
      case Expr.BOOL_LITERAL:     boolLiteral((LiteralExpr)expr); break;
      case Expr.INT_LITERAL:      intLiteral((LiteralExpr)expr); break;
      case Expr.FLOAT_LITERAL:    floatLiteral((LiteralExpr)expr); break;
      case Expr.STR_LITERAL:      strLiteral((LiteralExpr)expr); break;
      case Expr.DURATION_LITERAL: durationLiteral((LiteralExpr)expr); break;
      case Expr.URI_LITERAL:      uriLiteral((LiteralExpr)expr); break;
      case Expr.TYPE_LITERAL:     typeLiteral((LiteralExpr)expr); break;
      case Expr.RANGE_LITERAL:    rangeLiteral((RangeLiteralExpr)expr); break;
      case Expr.LIST_LITERAL:     listLiteral((ListLiteralExpr)expr); break;
      case Expr.MAP_LITERAL:      mapLiteral((MapLiteralExpr)expr); break;
      case Expr.BOOL_NOT:         not((UnaryExpr)expr); break;
      case Expr.NEGATE:           negate((UnaryExpr)expr); break;
      case Expr.POSITIVE:         expr(((UnaryExpr)expr).operand); break; // nop
      case Expr.PRE_INCREMENT:    increment((UnaryExpr)expr, ASSIGN_RET_PREFIX,  +1); break;
      case Expr.PRE_DECREMENT:    increment((UnaryExpr)expr, ASSIGN_RET_PREFIX,  -1); break;
      case Expr.POST_INCREMENT:   increment((UnaryExpr)expr, ASSIGN_RET_POSTFIX, +1); break;
      case Expr.POST_DECREMENT:   increment((UnaryExpr)expr, ASSIGN_RET_POSTFIX, -1); break;
      case Expr.ASSIGN:           assign((BinaryExpr)expr); break;
      case Expr.SAME:             same((BinaryExpr)expr); break;
      case Expr.NOT_SAME:         notSame((BinaryExpr)expr); break;
      case Expr.BOOL_OR:          or((BinaryExpr)expr); break;
      case Expr.BOOL_AND:         and((BinaryExpr)expr); break;
      case Expr.IS:               is((BinaryExpr)expr); break;
      case Expr.AS:               as((BinaryExpr)expr); break;
      case Expr.LOCAL_VAR:
      case Expr.THIS:
      case Expr.SUPER:            loadLocalVar((LocalVarExpr)expr); break;
      case Expr.CALL:
      case Expr.NEW:              call((CallExpr)expr); break;
      case Expr.SHORTCUT:         shortcut((ShortcutExpr)expr); break;
      case Expr.FIELD:            loadField((FieldExpr)expr); break;
      case Expr.TYPE_VAR:         /* ignore */ break;
      case Expr.CAST:             cast((CastExpr)expr); break;
      case Expr.CLOSURE:          expr(((ClosureExpr)expr).substitute); break;
      case Expr.TERNARY:          ternary((TernaryExpr)expr); break;
      default:                    throw new IllegalStateException(expr.getClass().getName());
    }
  }

//////////////////////////////////////////////////////////////////////////
// Literals
//////////////////////////////////////////////////////////////////////////

  private void nullLiteral()
  {
    op(LoadNull);
  }

  private void boolLiteral(LiteralExpr expr)
  {
    if (expr.asBool())
      op(LoadTrue);
    else
      op(LoadFalse);
  }

  private void intLiteral(LiteralExpr expr)
  {
    Int val = Int.make(expr.asInt());
    op(LoadInt, pod.integer(val));
  }

  private void floatLiteral(LiteralExpr expr)
  {
    Float val = Float.make(expr.asFloat());
    op(LoadFloat, pod.floats(val));
  }

  private void strLiteral(LiteralExpr expr)
  {
    Str val = Str.make(expr.asStr());
    op(LoadStr, pod.str(val));
  }

  private void durationLiteral(LiteralExpr expr)
  {
    Duration val = Duration.make(expr.asInt());
    op(LoadDuration, pod.duration(val));
  }

  private void uriLiteral(LiteralExpr expr)
  {
    Uri val = Uri.parse(Str.make(expr.asStr()));
    op(LoadUri, pod.uri(val));
  }

  private void typeLiteral(LiteralExpr expr)
  {
    Type val = expr.asType();
    op(LoadType, pod.typeRef(val));
  }

  private void rangeLiteral(RangeLiteralExpr expr)
  {
    expr(expr.start);
    expr(expr.end);
    if (expr.inclusive)
      op(CallNew, pod.methodRef(RangeMakeInclusive));
    else
      op(CallNew, pod.methodRef(RangeMakeExclusive));
  }

  private void listLiteral(ListLiteralExpr expr)
  {
    ListType type = (ListType)expr.type;
    Type ofType = type.v;

    op(LoadType, pod.typeRef(ofType));
    op(LoadInt,  pod.integer(Int.make(expr.items.size())));
    op(CallNew,  pod.methodRef(ListMake));

    int add = pod.methodRef(ListAdd);
    for (int i=0; i<expr.items.size(); ++i)
    {
      expr((Expr)expr.items.get(i));
      op(CallVirtual, add);
    }
  }

  private void mapLiteral(MapLiteralExpr expr)
  {
    op(LoadType, pod.typeRef(expr.type));
    op(CallNew,  pod.methodRef(MapMake));

    int set = pod.methodRef(MapSet);
    for (int i=0; i<expr.keys.size(); ++i)
    {
      expr((Expr)expr.keys.get(i));
      expr((Expr)expr.vals.get(i));
      op(CallVirtual, set);
    }
  }

//////////////////////////////////////////////////////////////////////////
// UnaryExpr
//////////////////////////////////////////////////////////////////////////

  private void negate(UnaryExpr expr)
  {
    expr(expr.operand);
    if (expr.type.isInt())
    {
      line(expr);
      op(CallVirtual, pod.methodRef(IntNegate));
    }
    else if (expr.type.isFloat())
    {
      line(expr);
      op(CallVirtual, pod.methodRef(FloatNegate));
    }
    else
    {
      throw new IllegalStateException(expr.toString());
    }
  }

  private void not(UnaryExpr expr)
  {
    expr(expr.operand);
    line(expr);
    op(CallVirtual, pod.methodRef(BoolNot));
  }

//////////////////////////////////////////////////////////////////////////
// BinaryExpr
//////////////////////////////////////////////////////////////////////////

  private void same(BinaryExpr expr)
  {
    if (expr.lhs.id() == Expr.NULL_LITERAL)
    {
      expr(expr.rhs);
      op(CompareNull);
    }
    else if (expr.rhs.id() == Expr.NULL_LITERAL)
    {
      expr(expr.lhs);
      op(CompareNull);
    }
    else
    {
      expr(expr.lhs);
      expr(expr.rhs);
      op(CompareSame);
    }
  }

  private void notSame(BinaryExpr expr)
  {
    if (expr.lhs.id() == Expr.NULL_LITERAL)
    {
      expr(expr.rhs);
      op(CompareNotNull);
    }
    else if (expr.rhs.id() == Expr.NULL_LITERAL)
    {
      expr(expr.lhs);
      op(CompareNotNull);
    }
    else
    {
      expr(expr.lhs);
      expr(expr.rhs);
      op(CompareNotSame);
    }
  }

  private void or(BinaryExpr expr)
  {
    // TODO: a||b||c... can be much more efficient by
    //   sharing push true and push false instructions

    expr(expr.lhs);
    int pushTrue  = jump(JumpTrue);
    expr(expr.rhs);
    int pushFalse = jump(JumpFalse);

    mark(pushTrue);
    op(LoadTrue);
    int end = jump(Jump);

    mark(pushFalse);
    op(LoadFalse);

    mark(end);
  }

  private void and(BinaryExpr expr)
  {
    // TODO: a&&b&&c... can be much more efficient by
    //   sharing push true and push false instructions

    expr(expr.lhs);
    int pushFalse1 = jump(JumpFalse);
    expr(expr.rhs);
    int pushFalse2 = jump(JumpFalse);

    op(LoadTrue);
    int end = jump(Jump);

    mark(pushFalse1);
    mark(pushFalse2);
    op(LoadFalse);

    mark(end);
  }

  private void is(BinaryExpr expr)
  {
    expr(expr.lhs);
    op(Is, pod.typeRef(expr.rhs.type));
  }

  private void as(BinaryExpr expr)
  {
    expr(expr.lhs);
    op(As, pod.typeRef(expr.rhs.type));
  }

//////////////////////////////////////////////////////////////////////////
// Assign
//////////////////////////////////////////////////////////////////////////

  /**
   * Simple assignment using =
   */
  private void assign(BinaryExpr expr)
  {
    switch (expr.lhs.id())
    {
      case Expr.LOCAL_VAR: storeLocalVar(expr); break;
      case Expr.FIELD:     storeField(expr); break;
      default: throw err("IllegalState: " + expr.lhs.getClass() + " - " + expr, expr);
    }
  }

  /**
   * Prefix and postfix increment/decrement operators: ++/--
   */
  private void increment(UnaryExpr expr, int retMode, int inc)
  {
    if (!expr.leave) retMode = ASSIGN_RET_NONE;
    new IncrementAsm(expr, retMode, inc).generate();
  }

  /**
   * IncrementAsm helper class
   */
  class IncrementAsm extends AssignAsm
  {
    IncrementAsm (UnaryExpr expr, int retMode, int inc)
    {
      super(expr.operand, retMode);
      this.expr = expr;
      this.inc = inc;
    }

    void compute()
    {
      // increment/decrement
      if (var.type.isInt())
      {
        line(expr);
        op(CallVirtual, inc > 0 ? pod.methodRef(IntIncrement) : pod.methodRef(IntDecrement));
      }
      else if (var.type.isFloat())
      {
        line(expr);
        op(LoadFloat, pod.floats(Float.make(1)));
        op(CallVirtual, inc > 0 ? pod.methodRef(FloatPlus) : pod.methodRef(FloatMinus));
      }
      else
      {
        throw new IllegalStateException(expr.toString());
      }
    }

    UnaryExpr expr;
    int inc;
  }

  /**
   * AssignAsm is a helper class used to generate boiler
   * plate code for assignment operations.
   */
  abstract class AssignAsm
  {
    AssignAsm(Expr var, int retMode)
    {
      this.var = var;
      this.retMode = retMode;
    }

    void generate()
    {
      // get the variable
      switch (var.id())
      {
        case Expr.LOCAL_VAR:
          loadLocalVar((LocalVarExpr)var);
          break;
        case Expr.FIELD:
          FieldExpr fexpr = (FieldExpr)var;
          loadField(fexpr, true);  // dup target on stack for upcoming set
          isInstanceField = !fexpr.field.isStatic().val;  // used to determine how to duplicate
          break;
        default: throw new IllegalStateException(var.getClass() + " : " + var);
      }

      // if postfix value, duplicate value before we preform computation
      if (retMode == ASSIGN_RET_POSTFIX) dup();

      // perform computation
      compute();

      // if prefix, duplicate after we've done increment/decrement
      if (retMode == ASSIGN_RET_PREFIX) dup();

      // save the variable back
      switch (var.id())
      {
        case Expr.LOCAL_VAR: storeLocalVar((LocalVarExpr)var); break;
        case Expr.FIELD:     storeField((FieldExpr)var); break;
        default: throw new IllegalStateException(var.getClass() + " : " + var);
      }
    }

    void dup()
    {
      op(isInstanceField ? DupDown : Dup);
    }

    abstract void compute();

    Expr var;
    boolean isInstanceField;
    int retMode;
  }

  static final int ASSIGN_RET_NONE    = 0;  // leave nothing on stack:   x += 2;
  static final int ASSIGN_RET_PREFIX  = 1;  // leave new value on stack: ++x; return x += 2;
  static final int ASSIGN_RET_POSTFIX = 2;  // leave old value on stack: x++;

//////////////////////////////////////////////////////////////////////////
// Local Var
//////////////////////////////////////////////////////////////////////////

  private void loadLocalVar(LocalVarExpr var)
  {
    op(LoadVar, var.register());
  }

  private void storeLocalVar(LocalVarExpr var)
  {
    op(StoreVar, var.register());
  }

  private void storeLocalVar(BinaryExpr assign)
  {
    expr(assign.rhs);
    if (assign.leave) op(Dup);
    storeLocalVar((LocalVarExpr)assign.lhs);
  }

//////////////////////////////////////////////////////////////////////////
// Field
//////////////////////////////////////////////////////////////////////////


  private void loadField(FieldExpr fexpr)
  {
    loadField(fexpr, false);
  }

  private void loadField(FieldExpr fexpr, boolean dupTarget)
  {
    if (fexpr.target != null)
    {
      expr(fexpr.target);
      if (dupTarget) op(Dup);
    }

    Field field = fexpr.field;

    if (fexpr.useAccessor)
    {
      Method getter = field.getter();

      int index = pod.methodRef(getter, 0);

      if (field.parent().isMixin().val)
      {
        if (getter.isStatic().val)
          op(CallMixinStatic, index);
        else
          op(CallMixinVirtual, index);
      }
      else
      {
        if (getter.isStatic().val)
          op(CallStatic, index);
        else
          op(CallVirtual, index);
      }
    }
    else
    {
      int index = pod.fieldRef(field);

      if (field.parent().isMixin().val)
      {
        if (field.isStatic().val)
          op(LoadMixinStatic, index);
        else
          op(LoadMixinInstance, index);
      }
      else
      {
        if (field.isStatic().val)
          op(LoadStatic, index);
        else
          op(LoadInstance, index);
      }
    }
  }

  private void storeField(BinaryExpr assign)
  {
    FieldExpr f = (FieldExpr)assign.lhs;
    boolean isInstanceField = !f.field.isStatic().val;  // used to determine how to duplicate

    if (f.target != null) expr(f.target);
    expr(assign.rhs);
    if (assign.leave)
    {
      if (isInstanceField)
        op(DupDown);
      else
        op(Dup);
    }
    storeField(f);
  }

  private void storeField(FieldExpr fexpr)
  {
    Field field = fexpr.field;
    if (fexpr.useAccessor)
    {
      Method setter = field.setter();

      int index = pod.methodRef(setter, 1);

      if (field.parent().isMixin().val)
      {
        if (setter.isStatic().val)
          op(CallMixinStatic, index);
        else
          op(CallMixinVirtual, index);
      }
      else
      {
        if (setter.isStatic().val)
          op(CallStatic, index);
        else
          op(CallVirtual, index);
      }
    }
    else
    {
      int index = pod.fieldRef(field);

      if (field.parent().isMixin().val)
      {
        if (field.isStatic().val)
          op(StoreMixinStatic, index);
        else
          op(StoreMixinInstance, index);
      }
      else
      {
        if (field.isStatic().val)
          op(StoreStatic, index);
        else
          op(StoreInstance, index);
      }
    }
  }


//////////////////////////////////////////////////////////////////////////
// Call
//////////////////////////////////////////////////////////////////////////

  private void call(CallExpr call)
  {
    if (call.isDynamic)
    {
      dynamicCall(call);
    }
    else
    {
      prepareCall(call);
      invokeCall(call, call.leave, call.id() == Expr.NEW);
    }
  }

  private void dynamicCall(CallExpr call)
  {
    // target
    expr(call.target);

    // name str literal
    op(LoadStr, pod.str(Str.make(call.name)));

    // args Obj[]
    op(LoadInt,  pod.integer(Int.make(call.args.size())));
    op(CallNew,  pod.methodRef(ListMakeObj));
    int add = pod.methodRef(ListAdd);
    for (int i=0; i<call.args.size(); ++i)
    {
      expr((Expr)call.args.get(i));
      op(CallVirtual, add);
    }

    // Obj.trap
    op(CallVirtual, pod.methodRef(ObjTrap));

    // pop return if no leave
    if (!call.leave) op(Pop);
  }

  private void prepareCall(CallExpr call)
  {
    if (call.target != null) expr(call.target);

    for (int i=0; i<call.args.size(); ++i)
      expr((Expr)call.args.get(i));
  }

  private void invokeCall(CallExpr call, boolean leave, boolean asNew)
  {
    Method method = call.method;
    int index = pod.methodRef(method, call.args.size());

    // write CallVirtual, CallNonVirtual, CallStatic, CallNew, or CallCtor;
    // note that if a constructor call has a target (this or super), then it
    // is a CallCtor instance call because we don't want to allocate
    // a new instance
    line(call);
    if (method.parent().isMixin().val)
    {
      if (method.isStatic().val)
        op(CallMixinStatic, index);
      else
        op(CallMixinVirtual, index);
    }
    else if (method.isStatic().val)
    {
      op(CallStatic, index);
    }
    else if (method.isCtor().val)
    {
      if (call.target == null || call.target.id() == Expr.TYPE_VAR)
        op(CallNew, index);
      else
        op(CallCtor, index);
    }
    else
    {
      // because CallNonVirtual maps to Java's invokespecial, we can't
      // use it for calls outside of the class (consider it like calling
      // protected method)
      int target = call.target.id();
      if (target == Expr.SUPER || (target == Expr.THIS && !method.isVirtual().val))
        op(CallNonVirtual, index);
      else
        op(CallVirtual, index);
    }

    // if we are leaving a value on the stack of a method which
    // has a parameterized return value, then we need to insert
    // a cast operation
    //   Int.toStr    -> non-generic - no cast
    //   Str[].toStr  -> return isn't parameterized - no cast
    //   Str[].get()  -> actual return is Obj, but we want Str - cast
    // TODO: this could be optimize a bit, because we are inserting
    //   an unnecessary cast for things like List.add().add(); however
    //   we need to make sure we cast for crazy things like a map
    //   of list values for Map.get()
    if (leave && method.isGenericInstance())
    {
      Type ret = method.getGenericMethod().returnType();
      if (ret.isGenericParameter())
        op(Cast, pod.typeRef(method.returnType()));
    }

    // if the method left a value on the stack, and we
    // aren't going to use it, then pop it off
    if (!leave)
    {
      // note we need to use the actual method signature (not parameterized)
      Method m = method;
      if (m.getGenericMethod() != null) m = m.getGenericMethod();

      if (!m.returnType().isVoid() || m.isCtor().val)
        op(Pop);
    }
  }

//////////////////////////////////////////////////////////////////////////
// Shortcut
//////////////////////////////////////////////////////////////////////////

  /**
   * Attempt to optimize a shortcut since most common ones can be
   * represented very efficiently directly in bytecode.
   */
  private void shortcut(ShortcutExpr expr)
  {
    // handle == null and != null
    if (expr.opToken == Token.EQ || expr.opToken == Token.NOT_EQ)
    {
      Expr lhs = expr.target;
      Expr rhs = (Expr)expr.args.get(0);
      if (lhs.id() == Expr.NULL_LITERAL)
      {
        expr(rhs);
        op(expr.opToken == Token.EQ ? CompareNull : CompareNotNull);
        return;
      }
      else if (rhs.id() == Expr.NULL_LITERAL)
      {
        expr(lhs);
        op(expr.opToken == Token.EQ ? CompareNull : CompareNotNull);
        return;
      }
    }

    // handle comparisions as special opcodes
    switch (expr.opToken)
    {
      case Token.EQ:      prepareCall(expr); op(CompareEQ); return;
      case Token.NOT_EQ:  prepareCall(expr); op(CompareNE); return;
      case Token.COMPARE: prepareCall(expr); op(Compare);   return;
      case Token.LT:      prepareCall(expr); op(CompareLT); return;
      case Token.LTEQ:    prepareCall(expr); op(CompareLE); return;
      case Token.GT:      prepareCall(expr); op(CompareGT); return;
      case Token.GTEQ:    prepareCall(expr); op(CompareGE); return;
    }

    // always check string concat first since it can
    // have string on either left or right hand side
    if (expr.opToken == Token.PLUS && expr.type.isStr())
    {
      addStr(expr, true);
      return;
    }

    // if assignment we need to wrap invokeShorcut with an
    // assignment  helper which sets up the stack for us
    if (expr.assign)
    {
      int retMode = expr.leave ? ASSIGN_RET_PREFIX : ASSIGN_RET_NONE;
      new ShortcutAssignAsm(expr, retMode).generate();
    }
    else
    {
      prepareCall(expr);
      invokeCall(expr, expr.leave, false);
    }
  }

  /**
   * ShortcutAssignAsm helper class
   */
  class ShortcutAssignAsm extends AssignAsm
  {
    ShortcutAssignAsm(ShortcutExpr call, int retMode)
    {
      super(call.target, retMode);
      this.call = call;
    }

    void compute()
    {
      expr((Expr)call.args.get(0)); // must be binary: target *= arg0
      invokeCall(call, true, false);
    }

    ShortcutExpr call;
  }

//////////////////////////////////////////////////////////////////////////
// Strings
//////////////////////////////////////////////////////////////////////////

  /**
   * Assemble code to build a string using sys::StrBuf.
   */
  private void addStr(ShortcutExpr expr, boolean topLevel)
  {
    if (topLevel)
    {
      line(expr);
      op(CallNew, pod.methodRef(StrBufMake));
    }

    Expr lhs = expr.target;
    Expr rhs = (Expr)expr.args.get(0);

    if (lhs instanceof ShortcutExpr && ((ShortcutExpr)lhs).opToken == Token.PLUS && lhs.type.isStr())
    {
      addStr((ShortcutExpr)lhs, false);
    }
    else
    {
      expr(lhs);
      line(lhs);
      op(CallVirtual, pod.methodRef(StrBufAdd));
    }

    expr(rhs);
    line(rhs);
    op(CallVirtual, pod.methodRef(StrBufAdd));

    if (topLevel) op(CallVirtual, pod.methodRef(StrBufToStr));
  }

//////////////////////////////////////////////////////////////////////////
// Cast
//////////////////////////////////////////////////////////////////////////

  private void cast(CastExpr cast)
  {
    // generate target expression
    expr(cast.target);
    line(cast);
    op(Cast, pod.typeRef(cast.toType.type));
  }

//////////////////////////////////////////////////////////////////////////
// Ternary
//////////////////////////////////////////////////////////////////////////

  private void ternary(TernaryExpr expr)
  {
    int falseLabel, endLabel = -1;

    expr(expr.condition);
    falseLabel = jump(JumpFalse);
    expr(expr.trueExpr);
    endLabel = jump(Jump);
    mark(falseLabel);
    expr(expr.falseExpr);
    mark(endLabel);
  }

//////////////////////////////////////////////////////////////////////////
// Code Buffer
//////////////////////////////////////////////////////////////////////////

  /**
   * Append a opcode with no arguments.
   */
  private void op(int opcode)
  {
    code.u1(opcode);
  }

  /**
   * Append a opcode with specified argument.
   */
  private void op(int opcode, int u2)
  {
    code.u1(opcode);
    code.u2(u2);
  }

//////////////////////////////////////////////////////////////////////////
// Backwards Jump
//////////////////////////////////////////////////////////////////////////

  /**
   * Get the current location as a mark to use for backwards jump.
   */
  private int mark()
  {
    return code.len;
  }

  /**
   * Used to specify a jump opcode, where we know the target offset (when
   * we are jumping backwards and have already marked jump location.
   */
  private void jump(int opcode, int offset)
  {
    code.u1(opcode);
    code.u2(offset);
  }

//////////////////////////////////////////////////////////////////////////
// Forward Jump
//////////////////////////////////////////////////////////////////////////

  /**
   * Add the specified jump opcode, leave two bytes for the jump
   * location, and return the offset to backpatch the jump
   * instruction using the mark() method.  This method  is used
   * for a forward jump where we don't know offset yet, and will
   * backpatch it later via mark(int).
   */
  private int jump(int opcode)
  {
    code.u1(opcode);
    int mark = code.len;
    code.u2(0xffff);
    return mark;
  }

  /**
   * Mark the current location as the target of the specified
   * jump location.  This method backpatches the offset.
   */
  private void mark(int jump)
  {
    if (jump < 0) return;
    code.u2(jump, code.len);
  }

  /**
   * Backpacth the location of the specified jump statement,
   * with the given offset location.
   */
  private void mark(int jump, int offset)
  {
    if (jump < 0) return;
    code.u2(jump, offset);
  }

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  /**
   * Map the opcode we are getting ready to add to the specified line number)
   */
  private void line(ILocation loc)
  {
    int line = loc.line();
    if (line <= 0) return;
    if (lastLine == line) return;
    lines.u2(0, ++lineCount);
    lines.u2(code.len);
    lines.u2(line);
    lastLine = line;
  }

//////////////////////////////////////////////////////////////////////////
// LoopStack
//////////////////////////////////////////////////////////////////////////

  /**
   * LoopStack is used to keep track of the breaks and continues
   * matched to their while/for loop so that we can backpatch the
   * jump offsets.
   */
  LoopStack loopStack()
  {
    if (loopStack == null) loopStack = new LoopStack();
    return loopStack;
  }

  static class LoopStack
  {
    void pop()
    {
      stack[--n] = null;
    }

    Loop push(Stmt stmt)
    {
      try
      {
        return stack[n++] = new Loop(stmt);
      }
      catch (Exception e)
      {
        throw new RuntimeException("Loop statement tree too deep");
      }
    }

    Loop findLoop(Stmt loopStmt)
    {
      for (int i=n-1; i>=0; --i)
        if (stack[i].stmt == loopStmt) return stack[i];
      throw new IllegalStateException();
    }

    Loop[] stack = new Loop[256];
    int n;
  }

  static class Loop
  {
    Loop(Stmt stmt) { this.stmt = stmt; }

    Stmt stmt;  //  WhileStmt or ForStmt
    ArrayList breaks    = new ArrayList(4);  // Integer marks
    ArrayList continues = new ArrayList(4);  // Integer marks
  }

//////////////////////////////////////////////////////////////////////////
// Signatures
//////////////////////////////////////////////////////////////////////////

  final static Method ObjTrap;
  final static Method BoolNot;
  final static Method IntNegate, IntIncrement, IntDecrement, IntPlus;
  final static Method FloatNegate, FloatPlus, FloatMinus;
  final static Method StrBufMake, StrBufAdd, StrBufToStr;
  final static Method ListMake, ListMakeObj, ListAdd;
  final static Method MapMake, MapSet;
  final static Method EnumOrdinal;
  final static Method RangeMakeInclusive, RangeMakeExclusive;
  static
  {
    ObjTrap      = Slot.findMethod("sys::Obj.trap",     true);
    BoolNot      = Slot.findMethod("sys::Bool.not",     true);
    IntNegate    = Slot.findMethod("sys::Int.negate",   true);
    IntIncrement = Slot.findMethod("sys::Int.increment",true);
    IntDecrement = Slot.findMethod("sys::Int.decrement",true);
    IntPlus      = Slot.findMethod("sys::Int.plus",     true);
    FloatNegate  = Slot.findMethod("sys::Float.negate",  true);
    FloatPlus    = Slot.findMethod("sys::Float.plus",    true);
    FloatMinus   = Slot.findMethod("sys::Float.minus",   true);
    StrBufMake   = Slot.findMethod("sys::StrBuf.make",  true);
    StrBufAdd    = Slot.findMethod("sys::StrBuf.add",   true);
    StrBufToStr  = Slot.findMethod("sys::StrBuf.toStr", true);
    ListMake     = Slot.findMethod("sys::List.make",    true);
    ListMakeObj  = Slot.findMethod("sys::List.makeObj", true);
    ListAdd      = Slot.findMethod("sys::List.add",     true);
    MapMake      = Slot.findMethod("sys::Map.make",     true);
    MapSet       = Slot.findMethod("sys::Map.set",      true);
    EnumOrdinal  = Slot.findMethod("sys::Enum.ordinal", true);
    RangeMakeInclusive = Slot.findMethod("sys::Range.makeInclusive", true);
    RangeMakeExclusive = Slot.findMethod("sys::Range.makeExclusive", true);
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public FPod pod;
  public Box code;
  public Box errTable;
  public int errCount;
  public Box lines;
  public int lineCount;
  private int lastLine = -1;
  private LoopStack loopStack;

}