//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.ast;

import java.util.*;
import fan.sys.*;
import fanx.compiler.*;
import fanx.fcode.*;

/**
 * ClassDef models a class definition
 */
public class ClassDef
  extends DefNode
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public ClassDef(ILocation location, CompilationUnit unit, String name, int flags)
  {
    super(location);
    this.unit    = unit;
    this.podName = unit.podName;
    this.name    = name;
    this.qname   = podName + "::" + name;
    this.flags   = flags;
    this.bases   = new ArrayList();
    this.mixins  = new ArrayList();
    this.slots   = new ArrayList();
    this.enums   = new ArrayList();
  }

//////////////////////////////////////////////////////////////////////////
// Methods
//////////////////////////////////////////////////////////////////////////

  /**
   * Return if this is a mixin (interface) class
   */
  public boolean isMixin()
  {
    return (flags & FConst.Mixin) != 0;
  }

  /**
   * Return if this is an enum class
   */
  public boolean isEnum()
  {
    return (flags & FConst.Enum) != 0;
  }

  /**
   * Return if this class is an anonymous class used
   * to implement a closure.
   */
  public boolean isClosure()
  {
    return closure != null;
  }

//////////////////////////////////////////////////////////////////////////
// Member Access
//////////////////////////////////////////////////////////////////////////

  /**
   * Return if this class has a slot definition for specified name.
   */
  public boolean hasSlot(String name)
  {
    return slot(name) != null;
  }

  /**
   * Return SlotDef for specified name or null.
   */
  public SlotDef slot(String name)
  {
    for (int i=0; i<slots.size(); ++i)
      if (((SlotDef)slots.get(i)).name.equalsIgnoreCase(name))
        return (SlotDef)slots.get(i);
    return null;
  }

  public FieldDef  field(String name)  { return (FieldDef)slot(name); }
  public MethodDef method(String name) { return (MethodDef)slot(name); }

  /**
   * Get the FieldDefs from the members list.
   */
  public FieldDef[] fields()
  {
    ArrayList acc = new ArrayList();
    for (int i=0; i<slots.size(); ++i)
      if (slots.get(i) instanceof FieldDef) acc.add(slots.get(i));
    return (FieldDef[])acc.toArray(new FieldDef[acc.size()]);
  }

  /**
   * Get the static FieldDefs from the members list.
   */
  public FieldDef[] staticFields()
  {
    ArrayList acc = new ArrayList();
    for (int i=0; i<slots.size(); ++i)
    {
      if (slots.get(i) instanceof FieldDef)
      {
        FieldDef def = (FieldDef)slots.get(i);
        if (def.isStatic()) acc.add(def);
      }
    }
    return (FieldDef[])acc.toArray(new FieldDef[acc.size()]);
  }

  /**
   * Get the instance FieldDefs from the members list.
   */
  public FieldDef[] instanceFields()
  {
    ArrayList acc = new ArrayList();
    for (int i=0; i<slots.size(); ++i)
    {
      if (slots.get(i) instanceof FieldDef)
      {
        FieldDef def = (FieldDef)slots.get(i);
        if (!def.isStatic()) acc.add(def);
      }
    }
    return (FieldDef[])acc.toArray(new FieldDef[acc.size()]);
  }

  /**
   * Get the MethodDefs from the members list.
   */
  public MethodDef[] methods()
  {
    ArrayList acc = new ArrayList();
    for (int i=0; i<slots.size(); ++i)
      if (slots.get(i) instanceof MethodDef) acc.add(slots.get(i));
    return (MethodDef[])acc.toArray(new MethodDef[acc.size()]);
  }

//////////////////////////////////////////////////////////////////////////
// Enum
//////////////////////////////////////////////////////////////////////////

  /**
   * Return EnumDef for specified name or null.
   */
  public EnumDef enumDef(String name)
  {
    for (int i=0; i<enums.size(); ++i)
    {
      EnumDef def = (EnumDef)enums.get(i);
      if (name.equals(def.name))
        return def;
    }
    return null;
  }

//////////////////////////////////////////////////////////////////////////
// Tree
//////////////////////////////////////////////////////////////////////////

  public Node walk(Visitor v)
  {
    v.enterClassDef(this);
    bases = walk(v, bases);
    //mixins = walk(v, mixins);
    slots = walk(v, slots);
    v.exitClassDef(this);
    return v.visit(this);
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    return "ClassDef " + qname;
  }

  public void tree(AstWriter out)
  {
    out.nl();
    out.w("ClassDef " + qname).nl();
    out.indent();

    if (enums.size() > 0)
    {
      out.w("enums:").nl();
      out.indent();
      for (int i=0; i<enums.size(); ++i)
        ((EnumDef)enums.get(i)).tree(out);
      out.unindent();
    }

    out.w("members:").nl();
    out.indent();
    for (int i=0; i<slots.size(); ++i)
      ((SlotDef)slots.get(i)).tree(out);
    out.unindent();

    out.unindent();
  }

  public void pretty(AstWriter out)
  {
    out.nl();
    if (isMixin())
      out.w("mixin " + qname).nl();
    else
      out.w("class " + qname).nl();

    if (base != null) out.w("  extends " + base).nl();
    if (mixins.size() > 0)
    {
      out.w("  mixins " + mixins.get(0));
      for (int i=1; i<mixins.size(); ++i) out.w(", " + mixins.get(i));
      out.nl();
    }

    out.w("{").nl();
    out.indent();

    for (int i=0; i<enums.size(); ++i)
      ((EnumDef)enums.get(i)).pretty(out);

    for (int i=0; i<slots.size(); ++i)
      ((SlotDef)slots.get(i)).pretty(out);

    out.unindent();
    out.w("}").nl();
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public CompilationUnit unit;   // parent unit
  public String podName;         // parent pod
  public String name;            // simple class name
  public String qname;           // podName::name
  public ShimType type;          // Type representation
  public ArrayList bases;        // TypeRef[]
  public TypeRef base;           // extends class
  public ArrayList mixins;       // TypeRef[]
  public ArrayList slots;        // SlotDef[]
  public ArrayList enums;        // EnumDef[] (only if isEnum)
  public ClosureExpr closure;    // if closure anonymous class
  public boolean synthetic;      // is this class auto-generated; TODO move to flags

}