//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   20 Jan 06  Brian Frank  Creation
//
package fanx.compiler.ast;

import java.util.*;
import fan.sys.*;
import fanx.compiler.*;
import fanx.compiler.ast.*;
import fanx.compiler.parser.*;
import fanx.fcode.*;

/**
 * ClosureExpr is an "inlined anonymous method" which closes over it's
 * lexical scope.  ClosureExpr is placed into the AST by the parser
 * with the actual method implementation.  But very early in the compiler
 * pipeline InitClosures remaps a ClosureExpr to an anonymous ClassDef
 * which extends Method.  The method implementation is moved to the
 * anonymous class's doCall() method.  However we leave ClosureExpr
 * in the AST in it's original location with a substitute expression.
 * The substitute expr just creates an instance of the anonymous class.
 * But by leaving the ClosureExpr in the tree, we can keep track of
 * the original lexical scope of the closure.
 */
public class ClosureExpr
  extends Expr
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public ClosureExpr(ILocation location)
  {
    super(location);
  }

//////////////////////////////////////////////////////////////////////////
// Scope
//////////////////////////////////////////////////////////////////////////

  public MethodVar resolveLocal(String name)
  {
    for (int i=0; i<enclosingLocals.size(); ++i)
    {
      MethodVar var = (MethodVar)enclosingLocals.get(i);
      if (var.name.equals(name)) return var;
    }
    return null;
  }

  /**
   * This method is called by ResolveVars to auto-generate the
   * implicit outer "this" field in the Closure's implementation
   * class:
   *   - add $this field to closure's implementation class
   *   - add $this param to closure's make constructor
   *   - set field from param in constructor
   *   - update substitute to make sure this is passed to ctor
   */
  public Field outerThisField()
  {
    if (outerThisField == null)
    {
      // define outer this as "$this"
      ShimType thisType = enclosingClass.type;
      ShimType implType = cls.type;
      ILocation loc = location;

      FieldDef def = new FieldDef(loc);
      cls.slots.add(def);
      def.name  = "$this";
      def.flags = FConst.Internal | FConst.Synthetic;
      def.type  = new TypeRef(loc, thisType);
      outerThisField = CompilerStep.mapField(implType, def);

      // add parameter to constructor
      MethodDef ctor = cls.method("$make");
      ctor.params.add(new MethodParamDef(loc, thisType, "$this"));
      MethodVar paramVar = new MethodVar(ctor.params.size(), thisType, "$this", FConst.Param, null);
      ctor.vars.add(paramVar);
      CompilerStep.mapMethod(implType, ctor);
      implType.restubSlots();

      // pass this to closure class constructor
      substitute.args.add(new ThisExpr(loc));
      substitute.method = ctor.method;

      // set field in constructor
      BinaryExpr assign = new BinaryExpr(loc, new Token(Token.ASSIGN),
        new FieldExpr(loc, outerThisField, new ThisExpr(location)),
        new LocalVarExpr(loc, paramVar));
      assign.leave = false;
      ctor.code.add(0, new ExprStmt(loc, assign));
    }
    return outerThisField;
  }

//////////////////////////////////////////////////////////////////////////
// Expr
//////////////////////////////////////////////////////////////////////////

  public int id()
  {
    return CLOSURE;
  }

//////////////////////////////////////////////////////////////////////////
// Tree
//////////////////////////////////////////////////////////////////////////

  protected void walkChildren(Visitor v)
  {
    signature = (MethodTypeRef)walk(v, signature);
    substitute = (CallExpr)walk(v, substitute );
    code = walk(v, code);
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  public String toString() { return "Closure " + name; }

  public void tree(AstWriter out)
  {
    out.w("Closure ").w(name).w(" ").w(leaveToString()).nl();
    out.indent();
    out.w("signature: "); signature.tree(out);

    if (substitute == null) out.w("substitute: null").nl();
    else { out.w("substitute: "); substitute.tree(out); }

    if (code == null) out.w("code: null").nl();
    else { out.w("code: "); code.tree(out); }

    out.unindent();
  }

  public void pretty(AstWriter out)
  {
    if (substitute != null)
    {
      substitute.pretty(out);
      return;
    }

    signature.pretty(out);
    out.nl();
    out.w("{").nl();
    out.indent();
    if (code != null) code.pretty(out, false);
    out.unindent();
    out.w("}");
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public MethodTypeRef signature;
  public CallExpr substitute;        // expr to create anonymous implementation class
  public Block code;                 // moved into a MethodDef in InitClosures
  public ClassDef enclosingClass;    // enclosing class
  public MethodDef enclosingMethod;  // enclosing method
  public ArrayList enclosingLocals;  // MethodVar[] available in my scope
  public String name;                // anonymous class name
  public ClassDef cls;               // anonymous class
  public Field outerThisField;       // auto-generated this$

}