//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.ast;

import java.util.*;
import fanx.compiler.*;
import fanx.compiler.parser.*;

/**
 * ShortcutExpr is used for operator expressions which are a shortcut
 * to a method call:
 *   a + b    -> a.plus(b)
 *   a - b    -> a.minus(b)
 *   a * b    -> a.star(b)
 *   a / b    -> a.slash(b)
 *   a % b    -> a.percent(b)
 *   a[b]     -> a.get(b)
 *   a[b] = c -> a.set(b, c)
 *   a[b]     -> a.slice(b) if b is Range
 *   a[b] = c -> a.splice(b, c) if b is Range
 *   a << b   -> a.lshift(b)
 *   a >> b   -> a.rshift(b)
 *   a & b    -> a.amp(b)
 *   a | b    -> a.pipe(b)
 *   a ^ b    -> a.caret(b)
 *   ~a       -> a.tilde()
 *   -a       -> a.negate()
 *   a == b   -> a.equals(b)
 *   a != b   -> ! a.equals(b)
 *   a <=>    -> a.compare(b)
 *   a > b    -> a.compare(b) > 0
 *   a >= b   -> a.compare(b) >= 0
 *   a < b    -> a.compare(b) < 0
 *   a <= b   -> a.compare(b) <= 0
 */
public class ShortcutExpr
  extends CallExpr
{

//////////////////////////////////////////////////////////////////////////
// Operations
//////////////////////////////////////////////////////////////////////////

  public static final int PLUS     = 0;
  public static final int MINUS    = 1;
  public static final int STAR     = 2;
  public static final int SLASH    = 3;
  public static final int PERCENT  = 4;
  public static final int LSHIFT   = 5;
  public static final int RSHIFT   = 6;
  public static final int AMP      = 7;
  public static final int PIPE     = 8;
  public static final int CARET    = 9;
  public static final int TILDE    = 10;
  public static final int NEGATE   = 11;
  public static final int EQUALS   = 12;
  public static final int COMPARE  = 13;
  public static final int GET      = 14;
  public static final int SET      = 15;
  public static final int SLICE    = 16;
  public static final int SPLICE   = 17;

  public static final String[] NAMES =
  {
    "plus",        // 0
    "minus",       // 1
    "star",        // 2
    "slash",       // 3
    "percent",     // 4
    "lshift",      // 5
    "rshift",      // 6
    "amp",         // 7
    "pipe",        // 8
    "caret",       // 9
    "tilde",       // 10
    "negate",      // 11
    "equals",      // 12
    "compare",     // 13
    "get",         // 14
    "set",         // 15
    "slice",       // 16
    "splice",      // 17
  };

//////////////////////////////////////////////////////////////////////////
// Token to Op
//////////////////////////////////////////////////////////////////////////

  /**
   * Map an operator token to it's special method name.
   * Rank is 1 for unary and 2 for binary.
   */
  public static int tokenToOp(Token token, int rank)
  {
    switch (token.type)
    {
      case Token.PLUS:           return PLUS;      // a + b
      case Token.MINUS:          return rank == 1 ? NEGATE : MINUS;  // -a; a - b
      case Token.STAR:           return STAR;      // a * b
      case Token.SLASH:          return SLASH;     // a / b
      case Token.PERCENT:        return PERCENT;   // a % b
      case Token.LSHIFT:         return LSHIFT;    // a << b
      case Token.RSHIFT:         return RSHIFT;    // a >> b
      case Token.AMP:            return AMP;       // a & b
      case Token.PIPE:           return PIPE;      // a | b
      case Token.CARET:          return CARET;     // a ^ b
      case Token.TILDE:          return TILDE;     // ~a
      case Token.EQ:             return EQUALS;    // a == b
      case Token.NOT_EQ:         return EQUALS;    // a != b
      case Token.COMPARE:        return COMPARE;   // a <=> b
      case Token.GT:             return COMPARE;   // a > b
      case Token.GTEQ:           return COMPARE;   // a >= b
      case Token.LT:             return COMPARE;   // a < b
      case Token.LTEQ:           return COMPARE;   // a <= b
      case Token.ASSIGN_PLUS:    return PLUS;      // a += b
      case Token.ASSIGN_MINUS:   return MINUS;     // a -= b
      case Token.ASSIGN_STAR:    return STAR;      // a *= b
      case Token.ASSIGN_SLASH:   return SLASH;     // a /= b
      case Token.ASSIGN_PERCENT: return PERCENT;   // a %= b
      case Token.ASSIGN_AMP:     return AMP;       // a &= b
      case Token.ASSIGN_PIPE:    return PIPE;      // a |= b
      case Token.ASSIGN_CARET:   return CARET;     // a ^= b
      case Token.ASSIGN_LSHIFT:  return LSHIFT;    // a <<= b
      case Token.ASSIGN_RSHIFT:  return RSHIFT;    // a >>= b
      default: throw new IllegalStateException(token.toString());
    }
  }

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public ShortcutExpr(ILocation location, int op, Token token, Expr a, Expr b)
  {
    super(location);

    this.op   = op;
    this.name = NAMES[op];

    if (token != null)
    {
      this.opToken = token.type;
      this.assign  = token.isAssign();
    }

    this.target = a;
    if (b != null) args.add(b);
    //if (c != null) args.add(c);
  }

  public ShortcutExpr(ILocation location, Token token, Expr a, Expr b)
  {
    this(location, tokenToOp(token, 2), token, a, b);
  }

  public ShortcutExpr(ILocation location, Token token, Expr a)
  {
    this(location, tokenToOp(token, 1), token, a, null);
  }

//////////////////////////////////////////////////////////////////////////
// Expr
//////////////////////////////////////////////////////////////////////////

  public int id()
  {
    return SHORTCUT;
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public int op;           // operation constant: PLUS, MINUS, etc
  public int opToken = -1; // token type: Token.ASSIGN_PLUS, Token.LT, etc or -1
  public boolean assign;   // is this a combo assignment like +=, -=, etc

}