//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   24 May 06  Brian Frank  Creation
//
package fanx.compiler.ast;

import java.util.ArrayList;
import fan.sys.*;
import fanx.compiler.*;

/**
 * SwitchStmt
 */
public class SwitchStmt
  extends Stmt
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public SwitchStmt(ILocation location)
  {
    super(location);
  }

//////////////////////////////////////////////////////////////////////////
// Stmt
//////////////////////////////////////////////////////////////////////////

  public int id()
  {
    return SWITCH;
  }

  public boolean isMethodExit()
  {
    // only if every case and the default block
    // end with a method exit
    for (int i=0; i<cases.size(); ++i)
    {
      SwitchCase c = (SwitchCase)cases.get(i);
      if (c.block == null) return false;
      if (!c.block.lastIsMethodExit()) return false;
    }

    if (defaultBlock != null)
      return defaultBlock.lastIsMethodExit();
    else
      return false;
  }

//////////////////////////////////////////////////////////////////////////
// Switch Stuff
//////////////////////////////////////////////////////////////////////////

  /**
   * Compute the range of this switch and store into min and max fields.
   * Return if the switch is a candidate for a table switch as a series of
   * contiguous (or near contiguous) cases which can be represented an
   * offset into a jump table.  Return false if the switch is not numeric
   * or too sparse to use as a table switch.
   */
  public boolean computeTableRange()
  {
    // we only compute ranges for Ints and Enums
    Type type = condition.type;
    boolean isInt = type.isInt();
    boolean isEnum = type.isEnum().val;
    if (!isInt && !isEnum)
      return false;

    // now we need to determine contiguous range
    min = Integer.MAX_VALUE;
    max = Integer.MIN_VALUE;
    int count = 0;
    for (int i=0; i<cases.size(); ++i)
    {
      SwitchCase c = (SwitchCase)cases.get(i);
      for (int j=0; j<c.cases.size(); ++j)
      {
        count++;
        Expr expr = (Expr)c.cases.get(j);
        Int literal = expr.asTableSwitchCase();
        if (literal == null) throw new IllegalStateException(""+expr);
        int val = (int)literal.val;
        if (val < min) min = val;
        if (val > max) max = val;
      }
    }

    // if no cases, then don't use tableswitch
    if (count == 0) return false;

    // enums and anything with less than 32 jumps if immediately
    // allowed, otherwise base the table on a percentage of cout
    int delta = max-min;
    if (isEnum) return true;
    if (delta < 32) return true;
    return count*32 > delta;
  }

//////////////////////////////////////////////////////////////////////////
// Tree
//////////////////////////////////////////////////////////////////////////

  protected void walkChildren(Visitor v)
  {
    condition = walk(v, condition);
    cases = walk(v, cases);
    defaultBlock = walk(v, defaultBlock);
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    return "switch (" + condition + ")";
  }

  public void tree(AstWriter out)
  {
    out.w("SwitchStmt").nl();
    out.indent();
    out.w("condition: "); condition.tree(out);
    for (int i=0; i<cases.size(); ++i)
      ((SwitchCase)cases.get(i)).tree(out);
    if (defaultBlock != null) { out.w("default: "); defaultBlock.tree(out); }
    out.unindent();
  }

  public void pretty(AstWriter out)
  {
    out.w("switch (");
    condition.pretty(out);
    out.w(")").nl();
    out.w("{").nl();
    out.indent();
    for (int i=0; i<cases.size(); ++i)
      ((SwitchCase)cases.get(i)).pretty(out);
    if (defaultBlock != null)
    {
      out.w("default:").nl();
      out.indent();
      defaultBlock.pretty(out, false);
      out.unindent();
    }
    out.unindent();
    out.w("}").nl();
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public Expr condition;                    // test expression
  public ArrayList cases = new ArrayList(); // SwitchCase[]
  public Block defaultBlock ;               // default block (or null)
  public int min, max;                      // see computeTableRange()

}