//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.resolver;

import java.util.*;
import fan.sys.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.ast.*;

/**
 * CallResolver handles the process of matching and call expression to the
 * best method to use.  It handles type checking and auto-conversions.
 */
public class CallResolver
  extends TypeChecker
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public CallResolver(CompilerSupport cs, ClassDef curClassDef, MethodDef curMethod, CallExpr call)
  {
    super(cs.compiler);
    this.loc         = call.location;
    this.call        = call;
    this.name        = call.name;
    this.args        = call.getArgs();
    this.curClassDef = curClassDef;
    this.curType     = curClassDef.type;
    this.curMethod   = curMethod;
    this.found       = call.method;
  }

//////////////////////////////////////////////////////////////////////////
// Resolve
//////////////////////////////////////////////////////////////////////////

  /**
   * Resolve the call, update call.method and call.type
   */
  public void resolve()
  {
try
{
    resolveBase();
    find();
    checkArgs();
    resolveExprType();
}
catch (RuntimeException e) // TODO hack
{
  System.out.println("Resolving " + call.location.format());
  throw e;
}
  }

//////////////////////////////////////////////////////////////////////////
// Resolve Base
//////////////////////////////////////////////////////////////////////////

  /**
   * Resolve the base type which defines the method we are calling.
   */
  private void resolveBase()
  {
    // if target unspecified, then assume a method on the current
    // class otherwise the method call must be on the target type
    if (call.target == null)
    {
      base = curType;
    }
    else
    {
      base = call.target.type;
    }

    if (base == null) throw new IllegalStateException();
  }

//////////////////////////////////////////////////////////////////////////
// Find
//////////////////////////////////////////////////////////////////////////

  /**
   * Find the method or field with the specified name.
   */
  private void find()
  {
    // if already "found", then skip this step
    if (found != null) return;

    // check if we are in a closure - they get implicit access
    // to the enclosing class (only if target was omitted)
    // TODO - this code is duplicated in ExprResolver
    Slot slot = null;
    if (call.target == null && curClassDef.isClosure())
    {
      ClosureExpr closure = curClassDef.closure;
      slot = closure.enclosingClass.type.slot(name, false);
      if (slot != null && !slot.isStatic().val)
        call.target = new FieldExpr(loc, closure.outerThisField(), new ThisExpr(loc));
    }

    // look it up in base type
    if (slot == null)
      slot = base.slot(name, false);

    // if still not found
    if (slot == null)
      throw err("Unknown method '" + errSig() + "'", loc);

    // if field
    if (slot instanceof Field)
    {
      Field field = (Field)slot;
      if (args.length == 0) slot = field.getter();
      else if (args.length == 1) slot = field.setter();
      else throw err("Expected method, not field '" + errSig() + "'", loc);
    }

    found = (Method)slot;
    call.method = found;
  }

  private String errSig()
  {
    String base;
    if (call.target == null && curClassDef.isClosure())
      base = curClassDef.closure.enclosingClass.type.qname().val;
    else
      base = this.base.qname().val;
    return base + "." + name;
  }

//////////////////////////////////////////////////////////////////////////
// Auto Convert
//////////////////////////////////////////////////////////////////////////

  /**
   * At this point we have selected the best match, and know that
   * TypeChecker has a conversion path.  Now we insert the auto convert
   * expressions into the call arguments.
   */
  private void checkArgs()
  {
    MethodParam[] params = found.paramArr();
    Expr[] oldArgs = args;
    ArrayList newArgs = new ArrayList();

    boolean err = false;
    if (params.length < args.length)
    {
      // if more args than params, always an err
      err = true;
    }
    else
    {
      for (int i=0; i<params.length; ++i)
      {
        if (i >= args.length)
        {
          // if param has a default value, then that is ok
          if (!params[i].hasDefault().val) { err = true; break; }
        }
        else
        {
          // ensure arg passed is convertable to param (cast
          // mixins to Obj if necessary)
          Expr newArg = convert(oldArgs[i], params[i].paramType());
          if (newArg == null) { err = true; break; }
          newArgs.add(newArg);
        }
      }
    }

    if (err)
    {
      StringBuilder msg = new StringBuilder("Invalid args ").append(name).append("(");
      for (int i=0; i<args.length; ++i)
      {
        if (i > 0) msg.append(", ");
        msg.append(args[i].type);
      }
      msg.append(") for ").append(found.name().val).append("(");
      for (int i=0; i<params.length; ++i)
      {
        if (i > 0) msg.append(", ");
        msg.append(params[i].paramType());
      }
      msg.append(")");
      throw err(msg.toString(), loc);
    }

    call.args = newArgs;
  }

//////////////////////////////////////////////////////////////////////////
// Resolve Expr Type
//////////////////////////////////////////////////////////////////////////

  /**
   * Compute the expression type the call itself (what get's left on the stack).
   */
  private void resolveExprType()
  {
    if (call.method.isCtor().val)
      call.type = call.method.parent();
    else
      call.type = call.method.returnType();
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  ILocation loc;           // constructor
  CallExpr call;           // constructor
  String name;             // constructor, checkConstructor()
  Expr[] args;             // constructor
  ClassDef curClassDef;    // constructor
  Type curType;            // constructor
  MethodDef curMethod;     // constructor
  Type base;               // resolveBase(), checkConstructor()
  boolean isNew;           // checkNew()
  Method found;            // find()

}