//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.resolver;

import fan.sys.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.ast.*;

/**
 * TypeChecker is used to check types of expressions including the ability
 * to auto-convert such as auto-casts, auto-boxing, auto-unboxing, etc.
 */
public class TypeChecker
  extends CompilerSupport
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public TypeChecker(Compiler compiler)
  {
    super(compiler);
  }

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  /**
   * Can the specified expression be expressed as the expected
   * type (with or without an auto-conversion).
   */
  public boolean is(Expr expr, Type expected)
  {
    // TODO - do we need convert?
    return convert(expr, expected) != null;
  }

  /**
   * Check that the specified expression matches the expected type.
   * If so then return the expression.  If it can match the expected
   * type via an automatic conversion, return conversion expression.
   * Otherwise return null.
   */
  public Expr convert(Expr expr, Type expected)
  {
    if (Sys.isStub) return expr;

    if (expr.id() == Expr.TYPE_VAR)
      throw err("Invalid use of type name as expression", expr);

    Type actual = expr.type;

    // void is never any type
    if (actual == Sys.VoidType) return null;

    // if actual is a closure, then use it's base to get it's
    // generic method signature
    if (actual.base() instanceof MethodType) actual = actual.base();
    if (expected.base() instanceof MethodType) expected = expected.base();

    // TODO: hack - if parameterized type has made it this far, then
    // something has most likely gone with higher level reflection?
    // this case shows up when using List, rather than Obj (should
    // we even allow that?)  We definitely want that for Method, so
    // probably we need to handle this case
    if (actual.isGenericParameter())   actual = actual.getRawType();
    if (expected.isGenericParameter()) expected = expected.getRawType();

    // check type reference equality first since
    // that checks most cases very efficiently
    if (expected == actual) return expr;

    // null can be used for any type (everything is boxed)
    if (expr.id() == Expr.NULL_LITERAL)
      return expr;

    // TODO: I changed from this approach to use java.lang.Object instead
    // if going from mixin/interface type to Obj - technically this
    // is allowed from a Fan perspective but Fan mixin/interfaces aren't
    // actually sys::Obj from the VM perspective; so we deal
    // with at the compiler level by inserting a cast operation
    // to make the VM happy
    /*
    if (expected == Sys.ObjType && actual.isMixin().val)
      return new CastExpr(Sys.ObjType, expr);
    */

    // check type compatibility
    if (actual.is(expected))
      return expr;

    // TODO hack b/c above fails in sys sometimes
    if (actual.qname().equals(expected.qname()).val)
      return expr;

    // handle auto upcast from Int to Float
    /*
    if (expr.type.isInt() && expected.isFloat())
      return new CastExpr(expected, expr);
    */

    // this is a type incompatibility error
    return null;
  }

}