//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   2 Dec 05  Brian Frank  Creation
//
package fanx.compiler.steps;

import java.io.*;
import java.util.*;
import fan.sys.*;
import fan.sys.List;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.ast.*;
import fanx.compiler.parser.*;

/**
 * CheckErrors walks the tree of statements and expressions looking
 * for errors the compiler can detect such as invalid type usage.  We
 * attempt to leave all the error reporting to this step, so that we
 * can batch report as many errors as possible.
 */
public class CheckErrors
  extends VisitorStep
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  public CheckErrors(Compiler compiler)
  {
    super(compiler);
  }

//////////////////////////////////////////////////////////////////////////
// Run
//////////////////////////////////////////////////////////////////////////

  public void run()
  {
    super.run();
    scanClasses();
    bombIfErr();
  }

  public Node visit(Node node)
  {
    if (node instanceof MethodDef) { checkMethod((MethodDef)node); return node; }
    if (node instanceof FieldDef)  { checkField((FieldDef)node); return node; }
    if (node instanceof Stmt)      { checkStmt((Stmt)node); return node; }
    if (node instanceof Expr)      { checkExpr((Expr)node); return node; }
    return node;
  }

//////////////////////////////////////////////////////////////////////////
// Scan Classes
//////////////////////////////////////////////////////////////////////////

  void scanClasses()
  {
    for (int i=0; i<compiler.classDefs.length; ++i)
      checkAbstract(compiler.classDefs[i]);
  }

  void checkAbstract(ClassDef cls)
  {
    if (cls.isAbstract()) return;

    MethodDef abstractMethod = null;

    // check for abstract methods cls declares itself
    for (int i=0; i<cls.slots.size(); ++i)
    {
      SlotDef slot = (SlotDef)cls.slots.get(i);
      if (slot instanceof MethodDef)
      {
        MethodDef method = (MethodDef)slot;
        if (abstractMethod == null && method.isAbstract())
          { abstractMethod = method; break; }
      }
    }

    if (abstractMethod != null)
      err("Class '" + cls.name + "' must be abstract since it contains abstract method '" + abstractMethod.name + "'", cls);

    // check for abstract methods inherited which aren't overridden
    List methods = cls.type.methods();
    for (int i=0; i<methods.sz(); ++i)
    {
      Method method = (Method)methods.get(i);
      if (method.isAbstract().val && method.parent() != cls.type)
        err("Class '" + cls.name + "' must be abstract since it inherits but doesn't override '" + method.qname() + "'", cls);
    }
  }

//////////////////////////////////////////////////////////////////////////
// FieldDef
//////////////////////////////////////////////////////////////////////////

  void checkField(FieldDef f)
  {
    ClassDef cls = curClassDef;

    if (f.isStatic() && !f.isConst())
      err("Static field '" + f.name + "' must be const", f);
  }

//////////////////////////////////////////////////////////////////////////
// MethodDef
//////////////////////////////////////////////////////////////////////////

  void checkMethod(MethodDef m)
  {
    ClassDef cls = curClassDef;

    // check ctors call super (or another this) ctor
    if (m.code != null && m.isCtor() && m.ctorChain == null)
    {
      if (!insertCtorSuper(cls, m))
        err("Must call a super class constructor in '" + cls.name + "." + m.name + "'", m);
    }
  }

  boolean insertCtorSuper(ClassDef cls, MethodDef m)
  {
    // never insert super call for synthetic classes
    if (cls.synthetic)
      return true;

    // skip mixins
    if (cls.isMixin())
      return true;

    // we can skip calling Obj.make - it is implied
    if (cls.type.base() == Sys.ObjType)
      return true;

    // implicitly insert a call to the super class constructor
    // if there is exactly one and it has no parameters
    Type base = cls.type.base();
    Method superCtor = null;
    List parentMethods = base.methods();
    for (int i=0; i<parentMethods.sz(); ++i)
    {
      Method pm = (Method)parentMethods.get(i);
      if (!pm.isCtor().val || pm.parent() != base) continue;
      if (superCtor != null) return false;  // more than one
      superCtor = pm;
    }

    // none found
    if (superCtor == null) return false;

    // check number of parameters
    if (superCtor.params().sz() != 0) return false;

    ILocation loc = m.location;
    m.ctorChain = new CallExpr(loc, new SuperExpr(loc), superCtor);
    m.ctorChain.isCtorChain = true;
    return true;
  }


//////////////////////////////////////////////////////////////////////////
// Statements
//////////////////////////////////////////////////////////////////////////

  void checkStmt(Stmt stmt)
  {
    switch (stmt.id())
    {
      case Stmt.IF:            checkIf((IfStmt)stmt); break;
      case Stmt.RETURN:        checkReturn((ReturnStmt)stmt); break;
      case Stmt.THROW:         checkThrow((ThrowStmt)stmt); break;
      case Stmt.FOR:           checkFor((ForStmt)stmt); break;
      case Stmt.WHILE:         checkWhile((WhileStmt)stmt); break;
      case Stmt.BREAK:         checkBreak((BreakStmt)stmt); break;
      case Stmt.CONTINUE:      checkContinue((ContinueStmt)stmt); break;
      case Stmt.LOCAL_VAR_DEF: localVarDef((LocalVarDefStmt)stmt); break;
      case Stmt.TRY:           checkTry((TryStmt)stmt); break;
    }
  }

  void checkIf(IfStmt stmt)
  {
    if (!stmt.condition.type.isBool())
      err("If condition must be Bool, not '" + stmt.condition.type + "'", stmt.condition);
  }

  void checkThrow(ThrowStmt stmt)
  {
    if (!is(stmt.exception, Sys.ErrType))
      err("Must throw Err, not '" + stmt.exception.type + "'", stmt.exception);
  }

  void checkFor(ForStmt stmt)
  {
    if (stmt.condition != null && !stmt.condition.type.isBool())
      err("For condition must be Bool, not '" + stmt.condition.type + "'", stmt.condition);
  }

  void checkWhile(WhileStmt stmt)
  {
    if (!stmt.condition.type.isBool())
      err("While condition must be Bool, not '" + stmt.condition.type + "'", stmt.condition);
  }

  void checkBreak(BreakStmt stmt)
  {
    if (stmt.loop == null)
      err("Break outside of loop (break is implicit in switch)", stmt);
  }

  void checkContinue(ContinueStmt stmt)
  {
    if (stmt.loop == null)
      err("Continue outside of loop", stmt);
  }

  void checkReturn(ReturnStmt stmt)
  {
    Type ret = curMethod.ret.type;
    if (stmt.expr == null)
    {
      if (!ret.isVoid())
        err("Must return a value from non-void method", stmt);
    }
    else
    {
      if (!is(stmt.expr, ret))
        err("Cannot return '" + stmt.expr.type + "' as '" + ret + "'", stmt);
    }
  }

  void localVarDef(LocalVarDefStmt stmt)
  {
    if (stmt.init != null)
    {
      if (!is(stmt.init, stmt.type.type))
        err("Type '" + stmt.init.type + "' is not assignable to '" + stmt.type.type + "'", stmt);
    }
  }

  void checkTry(TryStmt stmt)
  {
    boolean all = false;
    for (int i=0; i<stmt.catches.size(); ++i)
    {
      TryCatch c = (TryCatch)stmt.catches.get(i);
      if (c.errType == null)
      {
        if (all)
          err("Only one catch all allowed", c);
        all = true;
      }
      else
      {
        if (!c.errType.type.is(Sys.ErrType))
          err("Must catch Err, not '" + c.errType.type + "'", c);
      }
    }
  }

//////////////////////////////////////////////////////////////////////////
// Expr
//////////////////////////////////////////////////////////////////////////

  void checkExpr(Expr expr)
  {
    switch (expr.id())
    {
      case Expr.RANGE_LITERAL: checkRangeLiteral((RangeLiteralExpr)expr); break;
      case Expr.BOOL_NOT:      checkBool((UnaryExpr)expr); break;
      case Expr.ASSIGN:        checkAssign((BinaryExpr)expr); break;
      case Expr.BOOL_OR:       checkBool((BinaryExpr)expr); break;
      case Expr.BOOL_AND:      checkBool((BinaryExpr)expr); break;
      case Expr.IS:            checkIsAs((BinaryExpr)expr); break;
      case Expr.AS:            checkIsAs((BinaryExpr)expr); break;
      case Expr.CALL:          checkCall((CallExpr)expr); break;
      case Expr.FIELD:         checkField((FieldExpr)expr); break;
      case Expr.THIS:          checkThis((ThisExpr)expr); break;
      case Expr.TERNARY:       checkTernary((TernaryExpr)expr); break;
    }
  }

  void checkRangeLiteral(RangeLiteralExpr range)
  {
    if (!range.start.type.isInt() || !range.end.type.isInt())
      err("Range start..end types must be Int: '" + range.start.type + "'..'" + range.end.type + "'", range);
  }

  void checkBool(UnaryExpr expr)
  {
    Type operand = expr.operand.type;
    if (!operand.isBool())
      err("Cannot apply '" + Token.toString(expr.opToken) + "' operator to '" + operand + "'", expr);
  }

  void checkBool(BinaryExpr expr)
  {
    Type lhs = expr.lhs.type;
    Type rhs = expr.rhs.type;
    if (!lhs.isBool()|| !rhs.isBool())
      err("Cannot apply '" + Token.toString(expr.opToken) + "' operator to '" + lhs + "' and '" + rhs + "'", expr);
  }

  void checkIsAs(BinaryExpr expr)
  {
    if (expr.rhs.id() != Expr.TYPE_VAR)
      err("Right hand side of '" + Token.toString(expr.opToken) + "' operator must be a type name", expr.rhs);
  }

  void checkAssign(BinaryExpr expr)
  {
    if (!is(expr.rhs, expr.lhs.type))
      err("Type '" + expr.rhs.type + "' is not assignable to '" + expr.lhs.type + "'", expr);
  }

  void checkCall(CallExpr call)
  {
    // ensure we aren't calling constructors directly
    if (call.method.isCtor().val && !call.isCtorChain)
    {
      if (call.target != null && call.target.id() != Expr.TYPE_VAR)
        err("Cannot call constructor on existing instance", call);
    }

    // if instance method
    if (!call.method.isStatic().val && !call.method.isCtor().val)
    {
      // ensure we can't calling an instance method statically
      if (call.target == null || call.target.id() == Expr.TYPE_VAR)
        err("Cannot call instance method '" + call.method + "' in static context", call);
    }
  }

  void checkField(FieldExpr f)
  {
    if (!f.field.isStatic().val)
    {
      if (f.target == null || f.target.id() == Expr.TYPE_VAR)
        err("Cannot access instance field '" + f.field + "' in static context", f);
    }

    if (f.field.parent() != curType)
    {
      if (f.field.getter() != null) f.useAccessor = true;
    }
  }

  void checkThis(ThisExpr expr)
  {
    if (inStatic)
      err("Cannot access 'this' in static context", expr);
  }

  void checkTernary(TernaryExpr expr)
  {
    if (!expr.condition.type.isBool())
      err("Ternary condition must be Bool, not '" + expr.condition.type + "'", expr.condition);
  }

}