//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   16 Jan 06  Brian Frank  Creation
//
package fanx.compiler.steps;

import java.io.*;
import java.util.ArrayList;
import java.util.HashMap;
import fan.sys.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.ast.*;
import fanx.fcode.*;
import fanx.util.*;

/**
 * InitClosures:
 *   - Walk tree to find all ClosureExpr
 *   - Map ClosureExpr into an anonymous class
 *   - Create ClosureExpr substitute  with call to constructor
 *     of anonymous class
 *
 * See ClosureExpr for details how closures are implemented in
 * the compiler pipeline.
 */
public class InitClosures
  extends VisitorStep
{

  public InitClosures(Compiler compiler)
  {
    super(compiler);
  }

  public void run()
  {
    super.run();
    mapToClassDefs();
  }

  public Node visit(Node node)
  {
    if (node instanceof ClosureExpr) initClosure((ClosureExpr)node);
    return node;
  }

  void initClosure(ClosureExpr closure)
  {
    // set closure enclosing class, method, name (class$method$count)
    closure.enclosingClass  = curClassDef;
    closure.enclosingMethod = curMethod;
    closure.name = curClassDef.name + "$" + curMethod.name + "$" + curMethod.closureCount++;

    // the closure is typed by it's |...| signature
    closure.type = closure.signature.type;

    // add to my accumulator
    closures.add(closure);

    // replace ClosureExpr with ClosureCls.make
    ILocation loc = closure;
    closure.substitute = new CallExpr(loc, new TypeVarExpr(loc, closure.name), "$make");
  }

  void mapToClassDefs()
  {
    ClassDef[] cls = new ClassDef[closures.size()];
    for (int i=0; i<cls.length; ++i)
      cls[i] = mapToClassDef((ClosureExpr)closures.get(i));
    addClassDefs(cls);
  }

  ClassDef mapToClassDef(ClosureExpr closure)
  {
    // class Class$Method$Num
    // {
    //    Obj call(Obj a, Obj b, ...) { doCall(...) }
    //    R call(A a, B b, ...) { closure.code }
    // }
    //
    ILocation loc = closure.location;
    ClassDef cls = new ClassDef(loc, closure.enclosingClass.unit, closure.name, FConst.Internal | FConst.Synthetic);
    cls.synthetic = true;
    cls.base      = closure.signature;
    cls.closure   = closure;
    cls.slots.add( genMake(closure) );
    cls.slots.add( genCall(closure) );
    cls.slots.add( genDoCall(closure) );
    closure.cls  = cls;
    closure.code = null;
    return cls;
  }

  MethodDef genMake(ClosureExpr closure)
  {
    ILocation loc = closure.location;
    Block block = new Block(loc);
    block.statements.add(new ReturnStmt(loc));

    MethodDef m = new MethodDef(loc);
    m.flags = FConst.Public | FConst.Ctor | FConst.Synthetic;
    m.name = "$make";
    m.ret = new TypeRef(loc, Sys.VoidType);
    m.code = block;
    return m;
  }

  MethodDef genCall(ClosureExpr closure)
  {
    ILocation loc = closure.location;
    MethodTypeRef sig = closure.signature;
    int paramCount = sig.paramCount();

    // method def
    MethodDef m = new MethodDef(loc);
    m.name  = "call";
    m.flags = FConst.Override | FConst.Final | FConst.Synthetic;
    m.ret   = new TypeRef(loc, Sys.ObjType);
    m.code  = new Block(loc);

    // signature:
    //   call(List)                 // if > MaxIndirectParams
    //   call(Obj p0, Obj p1, ...)  // if <= MaxIndirectParams
    if (sig.paramCount() <= Method.MaxIndirectParams)
    {
      m.name += sig.paramCount();
      for (int i=0; i<sig.paramCount(); ++i)
      {
        m.params.add(new MethodParamDef(loc, Sys.ObjType, "p" + i));
        m.vars.add(new MethodVar(i+1, Sys.ObjType, "p"+i, FConst.Param, m.code));
      }
    }
    else
    {
      m.params.add(new MethodParamDef(loc, Sys.ObjType.toListOf(), "list"));
      m.vars.add(new MethodVar(1, Sys.ObjType.toListOf(), "list", FConst.Param, m.code));
    }

    // init the doCall() expr with arguments
    CallExpr c = new CallExpr(loc, new ThisExpr(loc), "doCall");
    for (int i=0; i<sig.paramCount(); ++i)
    {
      Expr arg;

      // p<i>
      if (sig.paramCount() <= Method.MaxIndirectParams)
      {
        arg = new LocalVarExpr(loc, (MethodVar)m.vars.get(i));
      }

      // list.get(<i>)
      else
      {
        CallExpr listGet = new CallExpr(loc, new LocalVarExpr(loc, (MethodVar)m.vars.get(0)), "get");
        listGet.args.add(new LiteralExpr(loc, Expr.INT_LITERAL, new Long(i)));
        arg = listGet;
      }

      // cast to closure param type
      arg = new CastExpr(sig.paramType(i), arg);

      // add to doCall() argument list
      c.args.add(arg);
    }

    // return:
    //   doCall; return null;  // if doCall() is void
    //   return doCall         // if doCall() non-void
    if (sig.ret.isVoid())
    {
      m.code.add(c);
      m.code.add(new ReturnStmt(new LiteralExpr(loc, Expr.NULL_LITERAL, null)));
    }
    else
    {
      m.code.add(new ReturnStmt(c));
    }

    return m;
  }

  MethodDef genDoCall(ClosureExpr closure)
  {
    ILocation loc = closure.location;
    MethodDef m = new MethodDef(loc);
    m.name  = "doCall";
    m.flags = FConst.Private | FConst.Synthetic;
    m.code  = closure.code;

    MethodTypeRef sig = closure.signature;
    m.ret = sig.ret;
    for (int i=0; i<sig.paramCount(); ++i)
    {
      TypeRef type = (TypeRef)sig.paramTypes.get(i);
      String name = (String)sig.paramNames.get(i);
      m.params.add(new MethodParamDef(type, name));
    }

    return m;
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  ArrayList closures = new ArrayList(64);
}