//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   16 Apr 06  Brian Frank  Creation
//
package fanx.compiler.steps;

import java.util.*;
import fan.sys.*;
import fanx.fcode.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.ast.*;
import fanx.compiler.parser.*;

/**
 * InitEnum is used to extract the EnumDefs of an enum into AST slots.
 */
public class InitEnum
  extends CompilerStep
{

  public InitEnum(Compiler compiler)
  {
    super(compiler);
  }

  public void run()
  {
    for (int i=0; i<compiler.classDefs.length; ++i)
      if (compiler.classDefs[i].isEnum())
        process(compiler.classDefs[i]);
  }

  void process(ClassDef cls)
  {
    makeCtor(cls);

    ArrayList toAdd = new ArrayList();

    for (int i=0; i<cls.enums.size(); ++i)
      toAdd.add(makeField(cls, (EnumDef)cls.enums.get(i)));

    toAdd.add(makeValues(cls));

    cls.slots.addAll(0, toAdd);
  }

  /**
   * Make constructor:  make(Int ordinal, Str name)
   */
  void makeCtor(ClassDef cls)
  {
    ILocation loc = cls.location;

    MethodDef m;

    // check if there is already a constructor called make
    if (cls.hasSlot("make"))
    {
      m = cls.method("make");
      // TODO: there is a ton of error checking we could do here:
      //   - is it ctor, must be private, can't have ordinal or name
    }
    else
    {
      m = new MethodDef(loc);
      m.name = "make";
      m.ret = new TypeRef(loc, Sys.VoidType);

      Block block = new Block(loc);
      block.statements.add(new ReturnStmt(loc));
      m.code = block;

      cls.slots.add(m);
    }

    // ensure private ctor
    m.flags = FConst.Ctor | FConst.Private;

    // Enum.make call
    CallExpr ctorChain = new CallExpr(loc, new SuperExpr(loc), "make");
    ctorChain.isCtorChain = true;
    ctorChain.args.add(new UnknownVarExpr(loc, "ordinal"));
    ctorChain.args.add(new UnknownVarExpr(loc, "name"));
    m.ctorChain = ctorChain;

    // insert ordinal, name params
    m.params.add(0, new MethodParamDef(loc, Sys.IntType, "ordinal"));
    m.params.add(1, new MethodParamDef(loc, Sys.StrType, "name"));
  }

  /**
   * Make enum value field:  public static final Foo name = make(ord, name)
   */
  FieldDef makeField(ClassDef cls, EnumDef def)
  {
    // ensure there isn't already a slot with same name
    if (cls.hasSlot(def.name))
      throw err("Enum value '" + def.name + "' conflicts with slot", def);

    ILocation loc = def.location;

    // initializer
    CallExpr init = new CallExpr(loc, null, "make");
    init.args.add(new LiteralExpr(loc, Expr.INT_LITERAL, new Long(def.ordinal)));
    init.args.add(new LiteralExpr(loc, Expr.STR_LITERAL, def.name));
    for (int i=0; i<def.ctorArgs.size(); ++i)
      init.args.add((Expr)def.ctorArgs.get(i));

    FieldDef f = new FieldDef(loc);
    f.flags = FConst.Public | FConst.Static | FConst.Const;
    f.name  = def.name;
    f.type  = new TypeRef(loc, cls.type);
    f.init  = init;
    return f;
  }

  /**
   * Make values field: List of Enum values
   */
  FieldDef makeValues(ClassDef cls)
  {
    // ensure there isn't already a slot with same name
    if (cls.hasSlot("values"))
      throw err("Enum 'values' conflicts with slot", cls.slot("values"));

    ILocation loc = cls.location;

    // initializer
    ListLiteralExpr list = new ListLiteralExpr(loc, new TypeRef(loc, cls.type));
    for (int i=0; i<cls.enums.size(); ++i)
    {
      // this is a bit hacked up in order to ensure we capture
      // right static field to add to the values list
      Expr typeTarget = new UnknownVarExpr(loc, cls.name);
      list.items.add(new UnknownVarExpr(loc, typeTarget, ((EnumDef)cls.enums.get(i)).name));
    }
    CallExpr init = new CallExpr(loc, list, "ro");

    // TODO - const field or accessor method?
    FieldDef f = new FieldDef(loc);
    f.flags = FConst.Public | FConst.Static | FConst.Const;
    f.name  = "values";
    f.type  = new ListTypeRef(loc, cls.type);
    f.init  = init;

    return f;
  }

}