//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   2 Dec 05  Brian Frank  Creation
//
package fanx.compiler.steps;

import java.io.*;
import java.util.ArrayList;
import java.util.HashMap;
import fan.sys.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.ast.*;
import fanx.fcode.*;
import fanx.util.*;

/**
 * At this point all the TypeRefs must be resolved to Types.  Now
 * we can map all the SlotDefs to Slots in the ShimType since
 * the signatures are resolved.  We perform duplicate slot name
 * checking too.
 *
 * foreach(ClassDef)
 *   base Type.slots     -> ShimType.slots
 *   FieldDef  -> Field  -> ShimType.slots
 *   MethodDef -> Method -> ShimType.slots
 */
public class InitShimSlots
  extends CompilerStep
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  public InitShimSlots(Compiler compiler)
  {
    super(compiler);
  }

//////////////////////////////////////////////////////////////////////////
// Run
//////////////////////////////////////////////////////////////////////////

  public void run()
  {
    while (!mapSlots());
  }

  private boolean mapSlots()
  {
    boolean done = true;
    for (int i=0; i<compiler.classDefs.length; ++i)
    {
      // get def and type
      ClassDef def = compiler.classDefs[i];
      ShimType type = def.type;

      // we can't process this type unless we have
      // processed all it's base types and mixins
      if (!isReadyToShim(def)) { done = false; continue; }

      // process inherited + declared slots
      mapSlots(def);
    }
    return done;
  }

  /**
   * Return that all the classes in the specified type's class
   * hiearchy are either reflected or already shimmed so that
   * we can process it.
   */
  private boolean isReadyToShim(ClassDef def)
  {
    if (!isReadyToUse(def.base)) return false;
    for (int i=0; i<def.mixins.size(); ++i)
      if (!isReadyToUse((TypeRef)def.mixins.get(i)))
        return false;
    return true;
  }

  /**
   * Return if the specified inherited type is ready to use
   * for shimming a type which uses it in it's class hierarchy.
   */
  private boolean isReadyToUse(TypeRef typeRef)
  {
    if (typeRef == null)
      return true;
    else if (typeRef.type instanceof ShimType)
      return ((ShimType)typeRef.type).init;
    else
      return true;
  }

//////////////////////////////////////////////////////////////////////////
// Map Slots
//////////////////////////////////////////////////////////////////////////

  void mapSlots(ClassDef def)
  {
    ShimType type = def.type;
    ArrayList jslots = new ArrayList(64);
    HashMap byName = new HashMap();
    HashMap byCase = new HashMap();
    HashMap nameToIndex = new HashMap(); // name -> Integer

    // map inherited slots
    merge(type.base, jslots, byName, byCase, nameToIndex);
    for (int i=0; i<def.mixins.size(); ++i)
      merge(((TypeRef)def.mixins.get(i)).type, jslots, byName, byCase, nameToIndex);

    // now process defs
    for (int i=0; i<def.slots.size(); ++i)
    {
      SlotDef slotDef = (SlotDef)def.slots.get(i);
      if (slotDef.isStaticInit()) continue;
      if (slotDef.isFieldAccessor()) { jslots.add(mapSlot(type, slotDef)); continue; }

      boolean override = checkDuplicate(def, slotDef, byName, byCase);

      if (override)
      {
        Integer index = (Integer)nameToIndex.get(slotDef.name);
        jslots.set(index.intValue(), mapSlot(type, slotDef));
      }
      else
      {
        jslots.add(mapSlot(type, slotDef));
      }
    }

    type.stubSlots((Slot[])jslots.toArray(new Slot[jslots.size()]));
  }

  private void merge(Type inheritedType, ArrayList slots, HashMap byName, HashMap byCase, HashMap nameToIndex)
  {
    if (inheritedType == null) return;
    List inheritedSlots = inheritedType.slots();
    for (int i=0; i<inheritedSlots.sz(); ++i)
      merge((Slot)inheritedSlots.get(i), slots, byName, byCase, nameToIndex);
  }

  private void merge(Slot slot, ArrayList slots, HashMap byName, HashMap byCase, HashMap nameToIndex)
  {
    if (slot.isCtor().val) return;
    String name = slot.name().val;
    Integer dup = (Integer)nameToIndex.get(name);
    if (dup != null)
    {
      Slot dupSlot = (Slot)slots.get(dup.intValue());
      // TODO - this hack needs to be thought thru, but for now
      // pick the non-abstract one
      if (slot.isAbstract().val && !dupSlot.isAbstract().val)
      {
      }
      else
      {
        slots.set(dup.intValue(), slot);
      }
    }
    else
    {
      slots.add(slot);
      nameToIndex.put(name, Integer.valueOf(slots.size()-1));
    }
    byName.put(name, slot);
    byCase.put(name.toLowerCase(), slot);
  }

  private boolean checkDuplicate(ClassDef def, SlotDef slotDef, HashMap byName, HashMap byCase)
  {
    String name  = slotDef.name;
    String lower = name.toLowerCase();
    ILocation loc = slotDef.location;
    boolean override = false;

    // handle duplicate by name
    Object dupName = byName.get(name);
    if (dupName != null)
    {
      // handle duplicate from base class
      if (dupName instanceof Slot)
      {
        Slot slot = (Slot)dupName;
        if (slotDef instanceof MethodDef && slot instanceof Method)
        {
          // don't bother checking closure, otherwise make sure
          // override follows all the right rules
          if (def.isClosure())
            override = true;
          else
            override = isOverrideOk((MethodDef)slotDef, (Method)slot);
        }
        else if (slotDef instanceof FieldDef && slot instanceof Method)
        {
          override = isFieldOverride((FieldDef)slotDef, (Method)slot);
        }
        else
        {
          throw err("Slot '" + name + "' already defined in '" + slot.parent() + "'", loc);
        }
      }
      // handle duplicate in my own class
      else
      {
        SlotDef dup = (SlotDef)dupName;
        throw err("Slot '" + name + "' already defined at " + dup.location.format(), loc);
      }
    }

    // handle duplicate by case
    Object dupCase = byCase.get(lower);
    if (dupCase != null && !override)
    {
      // handle duplicate from base class
      if (dupCase instanceof Slot)
      {
        Slot slot = (Slot)dupCase;
        throw err("Slot '" + name + "' differs only by case from '" + slot.qname() + "'", loc);
      }
      // handle duplicate in my own class
      else
      {
        SlotDef dup = (SlotDef)dupCase;
        throw err("Slot '" + name + "' differs only by case from " + dup.location.format(), loc);
      }
    }

    // check non-existant override
    if (slotDef.isOverride() && !override)
      throw err("Override of unknown virtual method '" + name + "'", loc);

    // add to duplicate maps
    byName.put(name,  slotDef);
    byCase.put(lower, slotDef);
    return override;
  }

  private boolean isOverrideOk(MethodDef def, Method base)
  {
    ILocation loc = def.location;

    // if bootstrap compile of sys assume ok
    if (compiler.isSys) return true;

    // if both base and my method are private it is ok
    if (base.isPrivate().val && (def.flags & FConst.Private) != 0)
      return true;

    // check base is virtual
    if (!base.isVirtual().val)
      throw err("Cannot override non-virtual method '" + base.qname() + "'", loc);

    // check override keyword was specified
    if (!def.isOverride())
      throw err("Must specify override keyword to override '" + base.qname() + "'", loc);

    // check that return types match
    if (!base.returnType().equals(def.ret.type).val && !base.returnType().isGenericParameter())
      throw err("Incompatible return type in override of '" + base.qname() + "' -> '" + base.returnType() + "' != '" + def.ret.type + "'", loc);

    // check that we have same parameter count
    if (base.params().sz() != def.params.size())
      throw err("Incompatible number of parameters in override of '" + base.qname() + "' -> " + base.params().sz() + " != " + def.params.size(), loc);

    // check that we have same parameter signatures
    for (int i=0; i<base.params().sz(); ++i)
    {
      MethodParam b    = (MethodParam)base.params().get(i);
      MethodParamDef d = (MethodParamDef)def.params.get(i);
      if (b.paramType() != d.type.type && !b.paramType().isGenericParameter())
        throw err("Incompatible " + StrUtil.toOrder(i) + " parameter in override of '" + base.qname() + "' -> '" + b.paramType() + "' != '" + d.type.type + "'", loc);
    }

    // looks good
    return true;
  }

  private boolean isFieldOverride(FieldDef field, Method method)
  {
    String name = field.name;
    if (method.params().sz() != 0 || !method.returnType().equals(field.type.type).val)
      throw err("Field '" + name + "' conflicts with method inherited from '" + method.parent() + "'", field);

    // ok override
    return true;
  }

}