//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   26 Dec 05  Brian Frank  Creation
//
package fanx.fcode;

import java.io.*;
import java.util.*;
import java.util.zip.*;
import fan.sys.*;
import fan.sys.Float;
import fanx.util.*;

/**
 * FPod is the read/write fcode representation of sys::Pod.  It's main job in
 * life is to manage all the pod-wide constant tables for names, literals,
 * type/slot references and type/slot definitions.
 */
public final class FPod
  implements FConst
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  public FPod(String podName, FStore store)
  {
    this.podName    = podName;
    this.store      = store;
    this.names      = new FTable.Names(this);
    this.typeRefs   = new FTable.TypeRefs(this);
    this.fieldRefs  = new FTable.FieldRefs(this);
    this.methodRefs = new FTable.MethodRefs(this);
    this.ints       = new FTable.Ints(this);
    this.floats     = new FTable.Floats(this);
    this.strs       = new FTable.Strs(this);
    this.durations  = new FTable.Durations(this);
    this.uris       = new FTable.Uris(this);
  }

//////////////////////////////////////////////////////////////////////////
// Lookup
//////////////////////////////////////////////////////////////////////////

  public FType type(String name)
  {
    for (int i=0; i<types.length; ++i)
      if (name(typeRef(types[i].self).typeName).equals(name))
        return types[i];
    throw UnknownTypeErr.make(name).val;
  }

//////////////////////////////////////////////////////////////////////////
// Compile
//////////////////////////////////////////////////////////////////////////

  /**
   * Initialize for compiler which requires bi-directional lookup.
   */
  public void initForCompile()
  {
    names.initForCompile();
    typeRefs.initForCompile();
    fieldRefs.initForCompile();
    methodRefs.initForCompile();
    ints.initForCompile();
    floats.initForCompile();
    strs.initForCompile();
    durations.initForCompile();
    uris.initForCompile();
  }

  /**
   * During runtime, once we have emitted the Pod class itself,
   * we no longer need to keep around the literal tables.
   */
  public void freeLiteralTables()
  {
    ints      = null;
    floats    = null;
    strs      = null;
    durations = null;
    uris      = null;
  }

//////////////////////////////////////////////////////////////////////////
// Convenience
//////////////////////////////////////////////////////////////////////////

  public final String name(int index)       { return (String)names.get(index);   }
  public final FTypeRef typeRef(int index)  { return (FTypeRef)typeRefs.get(index);  }
  public final FTuple fieldRef(int index)   { return (FTuple)fieldRefs.get(index);  }
  public final FTuple methodRef(int index)  { return (FTuple)methodRefs.get(index); }
  public final Int integer(int index)       { return (Int)ints.get(index);       }
  public final Float floats(int index)      { return (Float)floats.get(index);     }
  public final Str str(int index)           { return (Str)strs.get(index);       }
  public final Duration duration(int index) { return (Duration)durations.get(index); }
  public final Uri uri(int index)           { return (Uri)uris.get(index); }

  public final int name(String val)         { return names.put(val);  }
  public final int name(Str val)            { return names.put(val.val);  }
  public final int typeRef(FTypeRef val)    { return typeRefs.put(val); }
  public final int fieldRef(FTuple val)     { return fieldRefs.put(val); }
  public final int methodRef(FTuple val)    { return methodRefs.put(val); }
  public final int integer(Int val)         { return ints.put(val);   }
  public final int floats(Float val)        { return floats.put(val);  }
  public final int str(Str val)             { return strs.put(val);   }
  public final int duration(Duration val)   { return durations.put(val); }
  public final int uri(Uri val)             { return uris.put(val); }

//////////////////////////////////////////////////////////////////////////
// Compile Utils
//////////////////////////////////////////////////////////////////////////

  public final int typeRef(Type type)
  {
    int p = name(type.pod().name());
    int n = name(type.name());
    String sig = type.isGenericInstance() ? type.signature().val : "";
    return typeRef(new FTypeRef(p, n, sig));
  }

  public final int fieldRef(Field field)
  {
    int p = typeRef(field.parent());
    int n = name(field.name());
    int t = typeRef(field.fieldType());
    return fieldRef(new FTuple(p, n, t));
  }

  public final int methodRef(Method method)
  {
    return methodRef(method, method.params().sz());
  }

  public final int methodRef(Method method, int argCount)
  {
    try
    {
      // if this is a generic instantiation, we want to call
      // against the original generic method using it's raw
      // types, since that is how the system library will
      // implement the type
      if (method.isGenericInstance())
        method = method.getGenericMethod();

      int[] x = new int[3+argCount];
      x[0] = typeRef(method.parent());
      x[1] = name(method.name());
      x[2] = typeRef(method.returnType().getRawType());
      for (int i=0; i<argCount; ++i) // can be less than param count
      {
        MethodParam p = (MethodParam)method.params().get(i);
        x[i+3] = typeRef(p.paramType().getRawType());
      }
      return methodRef(new FTuple(x));
    }
    catch (RuntimeException e)
    {
      System.out.println("FPod.methodRef: " + method.parent() + " -> " + method.signature() + "  argCount=" + argCount);
      throw e;
    }
  }

//////////////////////////////////////////////////////////////////////////
// Java Emit Utils
//////////////////////////////////////////////////////////////////////////

  /**
   * Map a fcode method signature to a Java method emit signature.
   */
  public final String jcall(int index, int opcode)
  {
    if (jcalls == null) jcalls = new String[methodRefs.size()];
    String jcall = jcalls[index];
    if (jcall == null)
    {
      int[] m = methodRef(index).val;
      String parent = jname(m[0]);
      String name = name(m[1]);
      boolean onObj = parent.equals("fan/sys/Obj");

      // static methods on sys::Obj are really FanObj
      if (onObj && (opcode == CallStatic || opcode == CallNonVirtual))
      {
        parent = "fan/sys/FanObj";
      }

      StringBuilder s = new StringBuilder();
      s.append(parent);
      if (opcode == CallMixinStatic) s.append('$');
      s.append('.').append(name).append('(');
      for (int i=3; i<m.length; ++i)
        s.append('L').append(jname(m[i])).append(';');
      s.append(')');

      String ret = jname(m[2]);
      if (opcode == CallNew) s.append('L').append(parent).append(';'); // factory
      else if (ret.equals("fan/sys/Void")) s.append('V');
      else s.append('L').append(ret).append(';');

      jcall = s.toString();
      if (!onObj) jcalls[index] = jcall;
    }
    return jcall;
  }

  /**
   * Map a fcode field signature to a Java field emit signature.
   */
  public final String jfield(int index, boolean mixin)
  {
    if (jfields == null) jfields = new String[fieldRefs.size()];
    String jfield = jfields[index];
    if (jfield == null)
    {
      int[] f = fieldRef(index).val;
      StringBuilder s = new StringBuilder();
      s.append(jname(f[0]));
      if (mixin) s.append('$');
      s.append('.').append(name(f[1]))
       .append(':').append('L').append(jname(f[2])).append(';');
      jfield = jfields[index] = s.toString();
    }
    return jfield;
  }

  /**
   * Map a Java type name (:: replaced wit /) via typeRefs table.
   */
  public final String jname(int index)
  {
    if (jnames == null) jnames = new String[typeRefs.size()];
    String n = jnames[index];
    if (n == null)
    {
      FTypeRef ref = typeRef(index);
      n = jnames[index] = "fan/" + name(ref.podName) + '/' + name(ref.typeName);
    }
    return n;
  }

//////////////////////////////////////////////////////////////////////////
// Read
//////////////////////////////////////////////////////////////////////////

  /**
   * Read from a FStore which provides random access
   */
  public void read() throws IOException
  {
    names.read(store.read("names.def", true));
    typeRefs.read(store.read("typeRefs.def"));
    fieldRefs.read(store.read("fieldRefs.def"));
    methodRefs.read(store.read("methodRefs.def"));
    ints.read(store.read("ints.def"));
    floats.read(store.read("floats.def"));
    strs.read(store.read("strs.def"));
    durations.read(store.read("durations.def"));
    uris.read(store.read("uris.def"));

    // pod meta
    readPodMeta(store.read("pod.def", true));

    // type meta
    readTypeMeta(store.read("types.def", true));

    // full fcode always lazy loaded in Type.reflect()
  }

  /// debug forces full load of Types too
  public void readFully() throws IOException
  {
    read();
    for (int i=0; i<types.length; ++i)
      types[i].read();
  }

  /**
   * Read from an input stream (used for loading scripts from memory)
   */
  public void readFully(final ZipInputStream zip) throws IOException
  {
    FStore.Input in = new FStore.Input(zip)
    {
      public void close() throws IOException { zip.closeEntry(); }
    };

    ZipEntry entry;
    while ((entry = zip.getNextEntry()) != null)
    {
      String name = entry.getName();

      if (name.equals("names.def")) names.read(in);
      else if (name.equals("typeRefs.def")) typeRefs.read(in);
      else if (name.equals("fieldRefs.def")) fieldRefs.read(in);
      else if (name.equals("methodRefs.def")) methodRefs.read(in);
      else if (name.equals("ints.def")) ints.read(in);
      else if (name.equals("floats.def")) floats.read(in);
      else if (name.equals("strs.def")) strs.read(in);
      else if (name.equals("durations.def")) durations.read(in);
      else if (name.equals("uris.def")) uris.read(in);
      else if (name.equals("pod.def")) readPodMeta(in);
      else if (name.equals("types.def")) readTypeMeta(in);
      else if (name.endsWith(".fcode")) readType(name, in);
      else System.out.println("WARNING: unexpected file in pod: " + name);
    }
  }

  private void readPodMeta(FStore.Input in) throws IOException
  {
// TODO
    int first2 = in.u2();
    if (first2 != 0x0FC0)
    {
      System.out.println("FPod.readPodMeta: old version of " + podName);
      version = "0";
      depends = new Depend[0];
      facets  = new FFacet[0];
      return;
    }
    if (in.u2() != 0xDE05)
      throw new IOException("Invalid magic");

    int ver = in.u4();
    if (ver != FConst.FCodeVersion)
      throw new IOException("Invalid version 0x" + Integer.toHexString(ver));

    podName = in.utf();
    version = in.utf();
    depends = new Depend[in.u1()];
    for (int i=0; i<depends.length; ++i)
      depends[i] =Depend.parse(Str.make(in.utf()));

    facets = new FFacet[in.u2()];
    for (int i=0; i<facets.length; ++i)
      facets[i] = new FFacet().read(in);

    in.close();
  }

  private void readTypeMeta(FStore.Input in) throws IOException
  {
    types = new FType[in.u2()];
    for (int i=0; i<types.length; ++i)
    {
      types[i] = new FType(this).readMeta(in);
      types[i].hollow = true;
    }
    in.close();
  }

  private void readType(String name, FStore.Input in) throws IOException
  {
    if (types == null || types.length == 0)
      throw new IOException("types.def must be defined first");

    String typeName = name.substring(0, name.length()-".fcode".length());
    for (int i=0; i<types.length; ++i)
    {
      String n = name(typeRef(types[i].self).typeName);
      if (n.equals(typeName )) { types[i].read(in); return; }
    }

    throw new IOException("Unexpected fcode file: " + name);
  }

//////////////////////////////////////////////////////////////////////////
// Write
//////////////////////////////////////////////////////////////////////////

  public void write() throws IOException
  {
    if (names.size()      > 0) names.write(store.write("names.def"));
    if (typeRefs.size()   > 0) typeRefs.write(store.write("typeRefs.def"));
    if (fieldRefs.size()  > 0) fieldRefs.write(store.write("fieldRefs.def"));
    if (methodRefs.size() > 0) methodRefs.write(store.write("methodRefs.def"));
    if (ints.size()       > 0) ints.write(store.write("ints.def"));
    if (floats.size()     > 0) floats.write(store.write("floats.def"));
    if (strs.size()       > 0) strs.write(store.write("strs.def"));
    if (durations.size()  > 0) durations.write(store.write("durations.def"));
    if (uris.size()       > 0) uris.write(store.write("uris.def"));

    // write pod meta-data
    FStore.Output out = store.write("pod.def");
    out.u2(name(podName));
    out.close();

    // write type meta-data
    out = store.write("types.def");
    out.u2(types.length);
    for (int i=0; i<types.length; ++i)
      types[i].writeMeta(out);
    out.close();

    // write type full fcode
    for (int i=0; i<types.length; ++i)
      types[i].write();
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public String podName;     // pod's unique name
  public String version;     // pod version
  public Depend[] depends;   // pod dependencies
  public FFacet[] facets;    // pod facets
  public FStore store;       // store we using to read/write
  public FType[] types;      // pod's declared types
  public FTable names;       // identifier names: foo
  public FTable typeRefs;    // types refs:   [pod,type,variances*]
  public FTable fieldRefs;   // fields refs:  [parent,name,type]
  public FTable methodRefs;  // methods refs: [parent,name,ret,params*]
  public FTable ints;        // Int literals
  public FTable floats;      // Float literals
  public FTable strs;        // Str literals
  public FTable durations;   // Duration literals
  public FTable uris;        // Uri literals
  private String[] jnames;   // cached fan typeRef   -> java name
  private String[] jcalls;   // cached fan methodRef -> java method signatures
  private String[] jfields;  // cached fan fieldRef  -> java field signatures

}