//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   26 Dec 05  Brian Frank  Creation
//
package fanx.fcode;

import java.io.*;
import java.util.*;
import fan.sys.*;
import fanx.util.*;

/**
 * FPod is the read/write fcode representation of sys::Pod.
 */
public class FPrinter
  extends PrintWriter
  implements FConst
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  public FPrinter(FPod pod, OutputStream out)
  {
    super(out);
    this.pod = pod;
  }

  public FPrinter(FPod pod)
  {
    this(pod, System.out);
  }

//////////////////////////////////////////////////////////////////////////
// Dump
//////////////////////////////////////////////////////////////////////////

  public void all()
  {
    tables();
    types();
    flush();
  }

//////////////////////////////////////////////////////////////////////////
// Const Tables
//////////////////////////////////////////////////////////////////////////

  public void tables()
  {
    println("##### Tables #####");
    println("--- names ---");      pod.names.dump(pod, this);
    println("--- typeRefs ---");   pod.typeRefs.dump(pod, this);
    println("--- fieldRefs ---");  pod.fieldRefs.dump(pod, this);
    println("--- methodRefs ---"); pod.methodRefs.dump(pod, this);
    println("--- ints ---");       pod.ints.dump(pod, this);
    println("--- floats ---");     pod.floats.dump(pod, this);
    println("--- strs ---");       pod.strs.dump(pod, this);
    println("--- durations ---");  pod.durations.dump(pod, this);
    flush();
  }

//////////////////////////////////////////////////////////////////////////
// Types
//////////////////////////////////////////////////////////////////////////

  public void types()
  {
    println("##### Types #####");
    for (int i=0; i<pod.types.length; ++i)
      type(pod.types[i]);
    flush();
  }

  public void type(FType type)
  {
    println("--" + typeRef(type.self) + " extends " + typeRef(type.base) + "--");
    if (type.mixins.length > 0)
    {
      print("  mixin ");
      for (int i=0; i<type.mixins.length; ++i)
      {
        if (i > 0) print(", ");
        print(typeRef(type.mixins[i]));
      }
      println();
    }
    facets(type.facets);
    println();
    for (int i=0; i<type.fields.length; ++i)  field(type.fields[i]);
    for (int i=0; i<type.methods.length; ++i) method(type.methods[i]);
    flush();
  }

  public void slot(FSlot s)
  {
    if (s instanceof FField)
      field((FField)s);
    else
      method((FMethod)s);
    flush();
  }

  public void field(FField f)
  {
    println("  " + typeRef(f.type) + " " + name(f.name) + " [" + StrUtil.flagsToString(f.flags).trim() + "]");
    facets(f.facets);
    println();
  }

  public void method(FMethod m)
  {
    print("  " + typeRef(m.ret) + " " + name(m.name) + "(");
    FMethodVar[] params = m.params();
    for (int i=0; i<params.length; ++i)
    {
      FMethodVar p = params[i];
      if (i > 0) print(", ");
      print(typeRef(p.type) + " " + name(p.name));
    }
    println(") [" + StrUtil.flagsToString(m.flags).trim() + "]");
    for (int i=0; i<m.vars.length; ++i)
    {
      FMethodVar v = m.vars[i];
      String role = v.isParam() ?  "Param" : "Local";
      int reg = i + ((m.flags & FConst.Static) != 0 ? 0 : 1);
      println("    [" + role + " " + reg + "] " + name(v.name) + ": " + typeRef(v.type));
      if (v.def != null) code(v.def);
    }
    if (m.code != null)
    {
      println("    [Code]");
      code(m.code);
    }
    facets(m.facets);
    println();
  }

  public void code(FBuf code)
  {
    if (!showCode) return;
    flush();
    new FCodePrinter(pod, out).code(code);
  }


//////////////////////////////////////////////////////////////////////////
// Facets
//////////////////////////////////////////////////////////////////////////

  public void facets(FFacet[] facets)
  {
    if (facets == null) return;
    for (int i=0; i<facets.length; ++i)
      facet(facets[i]);
  }

  public void facet(FFacet facet)
  {
    String name = name(facet.name);
    if (name.equals(LineNumbers) && !showLines) return;
    println("    [" + name + "] len=" + facet.data.len);
    if (name.equals(SourceFile))  sourceFile(facet);
    if (name.equals(LineNumber))  lineNumber(facet);
    if (name.equals(ErrTable))    errTable(facet);
    if (name.equals(LineNumbers)) lineNumbers(facet);
  }

  public void sourceFile(FFacet facet)
  {
    println("      " + facet.data.utf());
  }

  public void lineNumber(FFacet facet)
  {
    println("      " + facet.data.u2());
  }

  public void errTable(FFacet facet)
  {
    byte[] buf = facet.data.buf;
    int len = facet.data.len;
    int count = buf[0] << 8 | buf[1];
    for (int i=2; i<len; i += 8)
    {
      int start   = buf[i+0] << 8 | buf[i+1];
      int end     = buf[i+2] << 8 | buf[i+3];
      int handler = buf[i+4] << 8 | buf[i+5];
      flush();
      String type = typeRef(buf[i+6] << 8 | buf[i+7]);
      println("      " + start + " to " + end + " -> " + handler + "  " + type);
    }
  }

  public void lineNumbers(FFacet facet)
  {
    byte[] buf = facet.data.buf;
    int len = facet.data.len;
    int count = buf[0] << 8 | buf[1];
    //println("      count=" + count);
    for (int i=2; i<len; i += 4)
    {
      int pc   = buf[i+0] << 8 | buf[i+1];
      int line = buf[i+2] << 8 | buf[i+3];
      println("      " + pc + ": " + line);
    }
  }

//////////////////////////////////////////////////////////////////////////
// Dump Utils
//////////////////////////////////////////////////////////////////////////

  private String typeRef(int index)
  {
    if (index == 65535) return "null";
    return pod.typeRefs.toString(index) + showIndex(index);
  }

  private String name(int index)
  {
    return pod.name(index) + showIndex(index);
  }

  private String showIndex(int index)
  {
    if (showIndex) return "[" + index + "]";
    return "";
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public final FPod pod;
  public boolean showIndex = false;
  public boolean showCode  = true;
  public boolean showLines = false;

}