//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   1 Jan 06  Brian Frank  Creation
//
package fanx.test;

import java.lang.reflect.*;
import fan.sys.*;
import fan.sys.Type;
import fanx.compiler.*;

/**
 * InheritTest
 */
public class InheritTest
  extends CompileTest
{

//////////////////////////////////////////////////////////////////////////
// Main
//////////////////////////////////////////////////////////////////////////

  public void doRun()
    throws Exception
  {
    verifyForward();
    verifyBackward();
    verifyErrors();
  }

//////////////////////////////////////////////////////////////////////////
// Forward
//////////////////////////////////////////////////////////////////////////

  String forward  =
    "class A : Test\n" +
    "{\n" +
    "  Str a() { return \"a\" }\n" +
    "  virtual Str f() { return \"af\" }\n" +
    "}\n" +

    "class B : A\n" +
    "{\n" +
    "  Str b()  { return \"b\" }\n" +
    "  Str bi() { return b() }\n" +
    "  Str ai() { return a() }\n" +
    "  override Str f()  { return \"bf\" }\n" +
    "}";

  public void verifyForward()
    throws Exception
  {
    Type[] types = compileToTypes(forward);
    Class clsA = types[0].emit();
    Class clsB = types[1].emit();

    verify(clsB.getSuperclass() == clsA);

    Object a = make(clsA);
    Object b = make(clsB);

    verify(invoke(a, "a"),  Str.make("a"));
    verify(invoke(b, "b"),  Str.make("b"));
    verify(invoke(b, "bi"), Str.make("b"));
    verify(invoke(b, "ai"), Str.make("a"));

    verify(invoke(a, "f"),  Str.make("af"));
    verify(invoke(b, "f"),  Str.make("bf"));

    verify(invoke(a, "setup"),    null);
    verify(invoke(a, "teardown"), null);
  }

//////////////////////////////////////////////////////////////////////////
// Backward
//////////////////////////////////////////////////////////////////////////

  String backward =
    "class B : A\n" +
    "{\n" +
    "  Str b() { return \"b\" }\n" +
    "  Str bi() { return b() }\n" +
    "  Str ai() { return a() }\n" +
    "}" +

    "class A\n" +
    "{\n" +
    "  Str a() { return \"a\" }\n" +
    "}\n";

  public void verifyBackward()
    throws Exception
  {
    Type[] types = compileToTypes(forward);
    Class clsA = types[0].emit();
    Class clsB = types[1].emit();

    verify(clsB.getSuperclass() == clsA);

    Object a = make(clsA);
    Object b = make(clsB);

    verify(invoke(a, "a"),  Str.make("a"));
    verify(invoke(b, "b"),  Str.make("b"));
    verify(invoke(b, "bi"), Str.make("b"));
    verify(invoke(b, "ai"), Str.make("a"));
  }

//////////////////////////////////////////////////////////////////////////
// Override Errors
//////////////////////////////////////////////////////////////////////////

  public void verifyErrors()
    throws Exception
  {
    // non-virtual, no-override

    // same class
    // method/method
    verifyErr("class A { Void f() {} Void f() {} }", "Slot 'f' already defined at Memory(1,11)");
    verifyErr("class A { Void f() {} Void F() {} }", "Slot 'F' differs only by case from Memory(1,11)");
    // method/field
    verifyErr("class A { Void f() {} Str f; }", "Slot 'f' already defined at Memory(1,11)");
    verifyErr("class A { Void f() {} Str F; }", "Slot 'F' differs only by case from Memory(1,11)");
    // field/method
    verifyErr("class A { Str f; Void f() {} }", "Slot 'f' already defined at Memory(1,11)");
    verifyErr("class A { Str f; Void F() {} }", "Slot 'F' differs only by case from Memory(1,11)");

    // different classes
    // method/method (only case, match is actually override error)
    verifyErr("Void f() {}", "Void F() {}", "Slot 'F' differs only by case from 'InheritErr::A.f'");
    // method/field
    verifyErr("Void f() {}", "Str f;", "Field 'f' conflicts with method inherited from 'InheritErr::A'");
    verifyErr("Void f() {}", "Str F;", "Slot 'F' differs only by case from 'InheritErr::A.f'");
    // field/method
    verifyErr("Str f;", "Void f() {}", "Slot 'f' already defined in 'InheritErr::A'");
    verifyErr("Str f;", "Void F() {}", "Slot 'F' differs only by case from 'InheritErr::A.f'");

    // method overrides - mismatched keywords
    verifyErr("Void f() {}", "Void f() {}", "Cannot override non-virtual method 'InheritErr::A.f'");
    verifyErr("Void f() {}", "override Void f() {}", "Cannot override non-virtual method 'InheritErr::A.f'");
    verifyErr("Void f() {}", "override Void g() {}", "Override of unknown virtual method 'g'");

    // method overrides wrong return
    verifyErr("virtual Void f() {}",      "override Str f() {}", "Incompatible return type in override of 'InheritErr::A.f' -> 'sys::Void' != 'sys::Str'");
    verifyErr("virtual Int f() {}",       "override Void f() {}", "Incompatible return type in override of 'InheritErr::A.f' -> 'sys::Int' != 'sys::Void'");
    // method wrong number parameters
    verifyErr("virtual Void f(Str x) {}",  "override Void f() {}", "Incompatible number of parameters in override of 'InheritErr::A.f' -> 1 != 0");
    verifyErr("virtual Void f() {}",       "override Void f(Str x) {}", "Incompatible number of parameters in override of 'InheritErr::A.f' -> 0 != 1");
    // method mismatched parameters
    verifyErr("virtual Void f(Obj x) {}",         "override Void f(Str x) {}", "Incompatible first parameter in override of 'InheritErr::A.f' -> 'sys::Obj' != 'sys::Str'");
    verifyErr("virtual Void f(Obj x, Str y) {}",  "override Void f(Obj x, Obj y) {}", "Incompatible second parameter in override of 'InheritErr::A.f' -> 'sys::Str' != 'sys::Obj'");
  }

  public void verifyErr(String af, String bf, String msg)
    throws Exception
  {
    String src =
    "class A\n" +
    "{\n" +
    "  " + af + "\n" +
    "}\n" +

    "class B : A\n" +
    "{\n" +
    "  " + bf + "\n" +
    "}\n";

    verifyErr(src, msg);
  }

  public void verifyErr(String src, String msg)
    throws Exception
  {
    CompileException ex = null;
    try
    {
      compiler.define("InheritErr", src, "Memory");
    }
    catch (CompileException e)
    {
      ex = e;
    }
    verify(ex != null);
    //System.out.println(ex.msg + " ?= " + msg);
    verify(ex.msg.equals(msg));
  }


}