//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   3 Sep 05  Brian Frank  Creation
//
package fanx.test;

import java.io.*;
import java.util.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.parser.*;

/**
 * TokenizerTest
 */
public class TokenizerTest
  extends Test
{

//////////////////////////////////////////////////////////////////////////
// Main
//////////////////////////////////////////////////////////////////////////

  public void run()
    throws Exception
  {
    // empty source
    verify("",         new Token[0]);
    verify(" ",        new Token[0]);
    verify("\n",       new Token[0]);
    verify(" \n \n",   new Token[0]);

    // single tokens
    verify("a",        id("a"));
    verify(" a",       id("a"));
    verify("a ",       id("a"));
    verify(" a ",      id("a"));
    verify("ab",       id("ab"));
    verify("a\n",      id("a"));
    verify("a77",      id("a77"));
    verify("x9y",      id("x9y"));
    verify("_",        id("_"));
    verify("_foo_",    id("_foo_"));
    verify("For",      id("For"));
    verify("java",     id("java"));
    verify("net",      id("net"));
    verify("jalias",   id("jalias"));
    verify("nalias",   id("nalias"));
    verify("jcode",    id("jcode"));
    verify("ncode",    id("ncode"));

    // keywords
    verify("abstract", new Token(Token.ABSTRACT));
    verify("as",       new Token(Token.AS));
    verify("for",      new Token(Token.FOR));
    verify("void",     new Token(Token.VOID));
    verify("while",    new Token(Token.WHILE));

    // numbers
    verify("3",            i(3));
    verify("3f",           r(3d));
    verify("3F",           r(3D));
    verify("3.0",          r(3.0));
    verify("3.0",          r(3.0));
    verify("73",           i(73));
    verify("73f",          r(73d));
    verify("73F",          r(73D));
    verify("73.0",         r(73.0));
    verify("73.0",         r(73.0));
    verify("123456",       i(123456));
    verify("123456f",      r(123456d));
    verify("123456F",      r(123456D));
    verify("123456.0",     r(123456.0));
    verify("123456.0",     r(123456.0));
    verify("07",           i(07));
    verify("7f",           r(7f));
    verify("07.0",         r(07f));
    verify(".2",           r(.2));
    verify("0.2",          r(0.2));
    verify("0.007",        r(0.007));
    verify(".12345",       r(.12345));
    verify("0.12345",      r(.12345));
    verify("12345.6789",   r(12345.6789));
    verify("3e6",          r(3e6));
    verify("3E6",          r(3E6));
    verify("3.0e6",        r(3.0e6));
    verify("3.0E6",        r(3.0E6));
    verify("3e-6",         r(3e-6));
    verify("3E-6",         r(3E-6));
    verify(".2e+6",        r(.2e+6));
    verify(".2E+6",        r(.2E+6));
    verify(".2e-03",       r(.2e-03));
    verify(".2E-03",       r(.2E-03));
    verify("1_234",        i(1234));
    verify("1_234_567",    i(1234567));
    verify("1.234_567F",   r(1.234567));
    verify("1.2e3_00",     r(1.2e300));
    verify("1_2.3_7e5_6",  r(12.37e56));
    verify("0x3",          i(0x3));
    verify("0x03",         i(0x03));
    verify("0x123",        i(0x123));
    verify("0xabcdef",     i(0xabcdef));
    verify("0xABCDEF",     i(0xABCDEF));
    verify("0x3aF7cE",     i(0x3aF7cE));
    verify("0x12345678",   i(0x12345678));
    verify("0xffffffff",   i(0xffffffffL));
    verify("0xfedcba98",   i(0xfedcba98L));
    verify("0xfedcba9812345678",     i(0xfedcba9812345678L));
    verify("0xFFFFFFFFFFFFFFFF",     i(0xFFFFFFFFFFFFFFFFL));
    verify("0xffff_ffff",            i(0xffffffffL));
    verify("0xFFFF_FFFF_FFFF_FFFF",  i(0xFFFFFFFFFFFFFFFFL));
    verify("2147483647",             i(2147483647L));
    //verify("-9223372036854775808",   i(-9223372036854775808L));
    verify("9223372036854775807",    i(9223372036854775807L));
    //verify("-9_223_372_036_854_775_808", i(-9223372036854775808L));
    verify("9_223_372_036_854_775_807", i(9223372036854775807L));
    verify("1.5E-45",                r(1.5E-45));
    verify("3.402E77",               r(3.402E77));
    verify("1.4E-100",               r(1.4E-100));
    verify("1.7976931348623157E38",  r(1.7976931348623157E38));
    verifyInvalid("3e");
    verifyInvalid("-3e");
    verifyInvalid("-.3e");
    verifyInvalid("+0.3e");
    verifyInvalid("0x");
    verifyInvalid("0xG");
    verifyInvalid("0x1FFFFFFFFFFFFFFFF");
    /* int bounds
    verifyInvalid("-2147483649");
    verifyInvalid("2147483648");
    verifyInvalid("-12345678900");
    verifyInvalid("+12345678900");
    */
    // these are not caught right now (20+ digits is caught)
    // verify("-9223372036854775808",  l(-9223372036854775808L));
    // verify("9223372036854775807",   l(9223372036854775807L));
    //verifyInvalid("-92233720368547758080");
    verifyInvalid("92233720368547758070");
    //verifyInvalid("-92233720368547758080");
    verifyInvalid("92233720368547758070");

    // durations
    verify("0ns",        duration(0));
    verify("5ns",        duration(5));
    verify("1ms",        duration(1000L*1000L));
    verify("1sec",       duration(1000L*1000L*1000L));
    verify("1min",       duration(60L*1000L*1000L*1000L));
    verify("1hr",        duration(60L*60L*1000L*1000L*1000L));
    verify("0.5ms",      duration(500L*1000L));
    verify("3.2ms",      duration(3200L*1000L));
    verify("0.001sec",   duration(1000L*1000L));
    verify("0.25min",    duration(15L*1000L*1000L*1000L));
    verify("24hr",       duration(24L*60L*60L*1000L*1000L*1000L));
    verify("876000hr",   duration(876000L*60L*60L*1000L*1000L*1000L));  // 100yr
    verify("1days",      duration(24L*60L*60L*1000L*1000L*1000L)); // 1day
    verify("0.5days",    duration(12L*60L*60L*1000L*1000L*1000L)); // 1/2yr
    verify("30days",     duration(30L*24L*60L*60L*1000L*1000L*1000L)); // 1day
    verify("36500days",  duration(876000L*60L*60L*1000L*1000L*1000L));  // 100yr

    // strings
    verify("\"\"",        s(""));
    verify("\"a\"",       s("a"));
    verify("\"ab\"",      s("ab"));
    verify("\"abc\"",     s("abc"));
    verify("\"a b\"",     s("a b"));
    verify("\"a\\nb\"",   s("a\nb"));
    verify("\"ab\\ncd\"", s("ab\ncd"));
    verify("\"\\b\"",     s("\b"));
    verify("\"\\t\"",     s("\t"));
    verify("\"\\n\"",     s("\n"));
    verify("\"\\f\"",     s("\f"));
    verify("\"\\r\"",     s("\r"));
    verify("\"\\\"\"",    s("\""));
    verify("\"''\"",      s("''"));
    verify("\"\\r\\n\"",  s("\r\n"));
    verify("\"\\u0001\"", s("\u0001"));
    verify("\"\\u0010\"", s("\u0010"));
    verify("\"\\u0100\"", s("\u0100"));
    verify("\"\\u1000\"", s("\u1000"));
    verify("\"\\uF000\"", s("\uF000"));
    verify("\"\\uFFFF\"", s("\uFFFF"));
    verify("\"\\uabcd\"", s("\uabcd"));
    verify("\"\\uABCD\"", s("\uABCD"));
    verify("\"a\nb\"",    s("a\nb"));      // with newline
    verify("\"a\nb\n c\"",s("a\nb\n c"));  // with newline
    verifyInvalid("\"");
    verifyInvalid("\"a");
    verifyInvalid("\"a\n");
    //verifyInvalid("\"\n\"");
    //verifyInvalid("\"a\n\"");
    verifyInvalid("\"\\u000g\"");
    verifyInvalid("\"\\u00g0\"");
    verifyInvalid("\"\\u0g00\"");
    verifyInvalid("\"\\ug000\"");

    // string interpolation
    verify("\"$x\"",  new Token[] { lparen(), s(""), plus(), id("x"), rparen() });
    verify("\"$x.toStr\"",  new Token[] { lparen(), s(""), plus(), id("x"), dot(), id("toStr"), rparen() });
    verify("\"foo$x.toStr bar\"",  new Token[] { lparen(), s("foo"), plus(), id("x"), dot(), id("toStr"), plus(), s(" bar"), rparen() });
    verify("\"foo${x.toStr}bar\"",  new Token[] { lparen(), s("foo"), plus(), lparen(), id("x"), dot(), id("toStr"), rparen(), plus(), s("bar"), rparen() });

    // uri literals
    verify("``",          uri(""));
    verify("`.`",         uri("."));
    verify("`'\"`",       uri("'\""));
    verify("`http://f/`", uri("http://f/"));

    // char literals
    verify("'a'",       c('a'));
    verify("'X'",       c('X'));
    verify("' '",       c(' '));
    verify("'\"'",      c('"'));
    verify("'\\n'",     c('\n'));
    verify("'\\r'",     c('\r'));
    verify("'\\''",     c('\''));
    verify("'\\uabcd'", c('\uabcd'));
    verifyInvalid("'a");
    verifyInvalid("'ab'");
    verifyInvalid("'\\q'");
    verifyInvalid("'\\ug000'");

    // comments
    verifyImpl("// foo bar",       new Token[0]);
    verifyImpl("/* foo bar */",    new Token[0]);
    verifyImpl("a// /* */",        new Token[] { id("a") } );
    verifyImpl("a// /* */more...", new Token[] { id("a") } );
    verifyImpl("a// /* */more\n",  new Token[] { id("a") } );
    verifyImpl("a// /* */more\nx", new Token[] { id("a"), id("x") } );
    verifyImpl("a/* foo bar */",   new Token[] { id("a") } );
    verifyImpl("a// blah blah\nb", new Token[] { id("a"), id("b") } );
    verifyImpl("a/* blah blah*/b", new Token[] { id("a"), id("b") } );
    verifyImpl("a/* 33 // 33 */b", new Token[] { id("a"), id("b") } );
    verifyImpl("a/* /*33*/ // 33 */b", new Token[] { id("a"), id("b") } );

    // operators
    verify(".",       new Token(Token.DOT));
    verify(";",       new Token(Token.SEMICOLON));
    verify(",",       new Token(Token.COMMA));
    verify("+",       new Token(Token.PLUS));
    verify("-",       new Token(Token.MINUS));
    verify("*",       new Token(Token.STAR));
    verify("/",       new Token(Token.SLASH));
    verify("%",       new Token(Token.PERCENT));
    verify("++",      new Token(Token.INCREMENT));
    verify("--",      new Token(Token.DECREMENT));
    verify("!",       new Token(Token.BANG));
    verify("?",       new Token(Token.QUESTION));
    verify("~",       new Token(Token.TILDE));
    verify("|",       new Token(Token.PIPE));
    verify("&",       new Token(Token.AMP));
    verify("^",       new Token(Token.CARET));
    verify("||",      new Token(Token.DOUBLE_PIPE));
    verify("&&",      new Token(Token.DOUBLE_AMP));
    verify("==",      new Token(Token.EQ));
    verify("!=",      new Token(Token.NOT_EQ));
    verify("===",     new Token(Token.SAME));
    verify("!==",     new Token(Token.NOT_SAME));
    verify("<",       new Token(Token.LT));
    verify("<=",      new Token(Token.LTEQ));
    verify(">",       new Token(Token.GT));
    verify(">=",      new Token(Token.GTEQ));
    verify("<<",      new Token(Token.LSHIFT));
    verify(">>",      new Token(Token.RSHIFT));
    verify("=",       new Token(Token.ASSIGN));
    verify("+=",      new Token(Token.ASSIGN_PLUS));
    verify("-=",      new Token(Token.ASSIGN_MINUS));
    verify("*=",      new Token(Token.ASSIGN_STAR));
    verify("/=",      new Token(Token.ASSIGN_SLASH));
    verify("%=",      new Token(Token.ASSIGN_PERCENT));
    verify("&=",      new Token(Token.ASSIGN_AMP));
    verify("|=",      new Token(Token.ASSIGN_PIPE));
    verify("^=",      new Token(Token.ASSIGN_CARET));
    verify("<<=",     new Token(Token.ASSIGN_LSHIFT));
    verify(">>=",     new Token(Token.ASSIGN_RSHIFT));
    verify("->",      new Token(Token.ARROW));
    verify("{",       new Token(Token.LBRACE));
    verify("}",       new Token(Token.RBRACE));
    verify("(",       new Token(Token.LPAREN));
    verify(")",       new Token(Token.RPAREN));
    verify("[",       new Token(Token.LBRACKET));
    verify("]",       new Token(Token.RBRACKET));
    verify(":",       new Token(Token.COLON));
    verify(":=",      new Token(Token.DEF_ASSIGN));
    verify("..",      new Token(Token.DOT_DOT));
    verify("...",     new Token(Token.DOT_DOT_DOT));
    verify("<=>",     new Token(Token.COMPARE));
    verifyInvalid("`");
    verifyInvalid("@");

    // double tokens
    verify("a b",       id("a"),  id("b"));
    verify("a\nb",      id("a"),  id("b"));
    verify("a3\n_f",    id("a3"), id("_f"));
    verify("a*b",       id("a"),  new Token(Token.STAR), id("b"));
    verify("for For",   new Token(Token.FOR),      id("For"));
    verify("1 2",       i(1), i(2));
    verify("70 * 12",   i(70), new Token(Token.STAR),  i(12));
    verify("70f / 3.5", r(70), new Token(Token.SLASH), r(3.5));
    verify("70*12",     i(70), new Token(Token.STAR),  i(12));
    verify("70f/3.5",   r(70), new Token(Token.SLASH), r(3.5));
    verify("7.toHex",   i(7),  new Token(Token.DOT),   id("toHex"));

    // position checking
    Token[] tok;
    tok = tokenize("\nfoo");
    verifyPos(tok[0], 2, 1);  verify(tok[0].equals(id("foo")));
    tok = tokenize("\n\n bar");
    verifyPos(tok[0], 3, 2);  verify(tok[0].equals(id("bar")));
    tok = tokenize("a\nb\nc");
    verifyPos(tok[0], 1, 1);  verify(tok[0].equals(id("a")));
    verifyPos(tok[1], 2, 1);  verify(tok[1].equals(id("b")));
    verifyPos(tok[2], 3, 1);  verify(tok[2].equals(id("c")));
    tok = tokenize(
    /*         123456789_123456789_1234 *
    /*  1 */  "foo bar\n" +
    /*  2 */  "here*\"there\"//junk\n" +
    /*  3 */  " 308 1.0 55 8f\n" +
    /*  4 */  "/* a b /*c*/ d \n" +
    /*  5 */  "*/ +=;\n" +
    /*  6 */  "  a\n" +
    /*  7 */  " /*/* /*f*/ */!** */  b\n" +
    /*  8 */  " !\n" +
    /*  9 */  "\n" +
    /* 10 */  "     bear\n" +
    /* 11 */  "}");
    verifyPos(tok[0],   1, 1);  // foo
    verifyPos(tok[1],   1, 5);  // bar
    verifyPos(tok[2],   2, 1);  // here
    verifyPos(tok[3],   2, 5);  // *
    verifyPos(tok[4],   2, 6);  // "there"
    verifyPos(tok[5],   3, 2);  // 308
    verifyPos(tok[6],   3, 6);  // 1.0
    verifyPos(tok[7],   3, 10); // 55
    verifyPos(tok[8],   3, 13); // 8r
    verifyPos(tok[9],   5, 4);  // +=
    verifyPos(tok[10],  5, 6);  // ;
    verifyPos(tok[11],  6, 3);  // a
    verifyPos(tok[12],  7, 23); // b
    verifyPos(tok[13],  8, 2);  // !
    verifyPos(tok[14], 10, 6);  // bear
    verifyPos(tok[15], 11, 1);  // }

    // comments
    verifyComment("** foo", "foo");
    verifyComment("*** foo", "foo");
    verifyComment("**  foo", "foo");
    verifyComment("**\n** foo\n **", "foo");
    verifyComment("**\n*** foo\n **", "foo");
    verifyComment("**\n**** foo\n++", "foo");
    verifyComment("** foo\n** bar", "foo\nbar");
    verifyComment("** foo\n ** bar", "foo\nbar");
    verifyComment("** foo\n **  bar", "foo\n bar");
    verifyComment("** foo\n **   bar", "foo\n  bar");
    verifyComment("**  foo\n ** bar", "foo\nbar");
    verifyComment(" **\n ** foo\n ** bar\n **\n public", "foo\nbar");
    verifyComment(" **\n ** foo\n ** bar\n **\n ** cool", "foo\nbar\n\ncool");
    verifyComment(" **\n ** foo\n ** bar\n **\n  **  cool", "foo\nbar\n\n cool");
  }

//////////////////////////////////////////////////////////////////////////
// Util
//////////////////////////////////////////////////////////////////////////

  public Token[] tokenize(String src)
  {
    Compiler c = new Compiler();
    c.includeDoc = true;
    c.log.silent = true;
    Tokenizer tokenizer = new Tokenizer(c, src.toCharArray(), "test");
    Token[] withEof = tokenizer.tokenize();

    // remove last eof
    Token[] result = new Token[withEof.length-1];
    System.arraycopy(withEof, 0, result, 0, result.length);
    return result;
  }

  public void verify(String src, Token want)
  {
    verify(src, new Token[] { want });
  }

  public void verify(String src, Token want0, Token want1)
  {
    verify(src, new Token[] { want0, want1 });
  }

  public void verify(String src, Token want0, Token want1, Token want2)
  {
    verify(src, new Token[] { want0, want1, want2 });
  }

  public void verify(String src, Token[] want)
  {
    // try exact
    verifyImpl(src, want);

    // try with trailing semi colon to ensure
    // tokenizer  left in correct state
    Token[] withSemi = new Token[want.length+1];
    System.arraycopy(want, 0, withSemi, 0, want.length);
    withSemi[want.length] = new Token(Token.SEMICOLON);
    verifyImpl(src + ";", withSemi);
  }

  public void verifyImpl(String src, Token[] want)
  {
    Token[] got = tokenize(src);

    /*
    System.out.println("-- Tokenize \"" + src.replace('\n', ' ') + "\" -> " + got.length);
    for (int i=0; i<got.length; ++i)
      System.out.println("got[" + i + "] =" + got[i]);
    */

    verify(got.length == want.length);
    for (int i=0; i<got.length; ++i)
    {
      if (verbose) System.out.println("  [" + i + "] " + got[i] + " ?= " + want[i]);
      verify(got[i].equals(want[i]));
    }
  }

  public void verifyInvalid(String src)
  {
    CompileException ex = null;
    try
    {
      tokenize(src);
    }
    catch (CompileException e)
    {
      //System.out.println(e);
      ex = e;
    }
    verify(ex != null);
  }

  public void verifyPos(Token t, int line, int col)
  {
    //System.out.println(" pos " + t.line + ":" + t.col + " ?= " + line + ":" + col + "  " + t);
    verify(t.line == line);
    verify(t.col  == col);
  }

  public void verifyComment(String src, String comment)
  {
    Token[] t = tokenize(src);
    verify(t.length >= 1);
    verify(t[0].type == Token.DOC_COMMENT);
    // System.out.println("\"" + comment.replace('\n', '_') + "\" ?= \"" + t[0].val.toString().replace('\n', '_') + "\"");
    verify(t[0].val.equals(comment));
  }

  public static void dump(File f)
    throws Exception
  {
    InputStream in = new FileInputStream(f);
    CharArrayWriter out = new CharArrayWriter();
    int last = -1;
    int c;
    while ((c = in.read()) > 0)
    {
      if (last == '\r' && c == '\n') continue;
      out.write((char)c);
      last = c;
    }
    char[] buf = out.toCharArray();
    for (int i=0; i<buf.length; ++i) if (buf[i] == '\r') buf[i] = '\n';
    Tokenizer t = new Tokenizer(null, buf, f.getName());
    Token[] toks = t.tokenize();
    for (int i=0; i<toks.length; ++i)
    {
      Token tok = toks[i];
      //System.out.println(tok.line + ":" + tok.col + "  " + tok);
    }
  }

  public Token i(long v)     { return new Token(Token.INT_LITERAL,  new Long(v));    }
  public Token r(double v)   { return new Token(Token.FLOAT_LITERAL, new Double(v));  }
  public Token c(char v)     { return new Token(Token.CHAR_LITERAL, new Character(v)); }
  public Token s(String v)   { return new Token(Token.STR_LITERAL,  v); }
  public Token duration(long v)  { return new Token(Token.DURATION_LITERAL,  new Long(v)); }
  public Token uri(String v) { return new Token(Token.URI_LITERAL,  v); }
  public Token id(String v)  { return new Token(Token.ID, v); }
  public Token lparen()      { return new Token(Token.LPAREN); }
  public Token rparen()      { return new Token(Token.RPAREN); }
  public Token plus()        { return new Token(Token.PLUS); }
  public Token dot()         { return new Token(Token.DOT); }

}