//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Dec 06  Andy Frank  Creation
//

#include <stdio.h>
#include <mscoree.h>
#include <Windows.h>
#include "launcher.h"
#include "utils.h"

#import <mscorlib.tlb> raw_interfaces_only
using namespace mscorlib;

//////////////////////////////////////////////////////////////////////////
// Globals
//////////////////////////////////////////////////////////////////////////

ICorRuntimeHost* clr  = NULL; // ptr to CLR host
_AppDomain* appDomain = NULL; // default AppDomain

//////////////////////////////////////////////////////////////////////////
// Error
//////////////////////////////////////////////////////////////////////////

int err(HRESULT hr, const char* msg1, LPWSTR msg2)
{
  printf("ERROR: ");
  if (msg1 != NULL) printf(msg1);
  if (msg2 != NULL) printf("%S", msg2);
  printf("\n");

  if (hr != NULL)
    printf("[0x%08x] %s\n", hr, _com_error(hr).ErrorMessage());

  if (clr != NULL)
  {
    clr->Stop();
    if (debug) printf("--   clr.stop\n");
  }

  return -1;
}

int err(HRESULT hr, const char* msg)
{
  return err(hr, msg, NULL);
}


//////////////////////////////////////////////////////////////////////////
// Load the CLR
//////////////////////////////////////////////////////////////////////////

int loadClr()
{
  LPWSTR ver     = NULL;    // version of CLR to load (use NULL for latest ver)
  LPWSTR flavor  = L"wks";  // build flavor ("wks" or "svr")
  DWORD flags    = STARTUP_LOADER_OPTIMIZATION_SINGLE_DOMAIN | STARTUP_CONCURRENT_GC;
  REFCLSID clsid = CLSID_CorRuntimeHost;
  REFIID riid    = IID_ICorRuntimeHost;

  // load CLR
  HRESULT hr = CorBindToRuntimeEx(ver, flavor, flags, clsid, riid, (void**)&clr);
  if (FAILED(hr)) return err(hr, "Could not load CLR");

  // start CLR
  hr = clr->Start();
  if (FAILED(hr)) return err("Could not start CLR");
  if (debug) printf("--   clr.start\n");

  // get version of the CLR just loaded
  wchar_t clrVer[50];
  DWORD sz = 50;
  hr = GetCORVersion(clrVer, sz, &sz);
  if (FAILED(hr)) return err(hr, "Could not get CLR version");
  if (debug) printf("--   clr.ver = %S\n", clrVer);

  // get default app domain
  IUnknown* unk;
  hr = clr->GetDefaultDomain(&unk);
  if (FAILED(hr)) return err(hr, "Could not get default app domain");
  hr = unk->QueryInterface(&appDomain);
  if (FAILED(hr)) return err(hr, "Could query for _AppDomain interface");

  return 0;
}

//////////////////////////////////////////////////////////////////////////
// Load sys.dll
//////////////////////////////////////////////////////////////////////////

int loadSys()
{
  if (debug) printf("--   runIt\n");

  // get sys path
  char sysPath[MAX_PATH];
  strcpy(sysPath, fanHome);
  strcat(sysPath, "\\lib\\net\\sys.dll");
  if (debug) printf("--   sysPath = %s\n", sysPath);

  // figure out main
  char temp[256];
  sprintf(temp, "Fanx.Tools.%s", FAN_MAIN);
  ULONG sz = strlen(temp) + 1;
  LPWSTR mainClassName = new WCHAR[sz];
  MultiByteToWideChar(CP_ACP, 0, temp, sz, mainClassName, sz);
  if (debug) printf("--   fan.main = %S\n", mainClassName);

  HRESULT hr;
  _Assembly* assembly;
  BSTR bstr;

  FILE *fp;
  long len;
  char *buf;

  fp = fopen(sysPath,"rb");
  if (fp == NULL) return err("Could not find sys.dll");

  // get file size
  fseek(fp, 0, SEEK_END);
  len = ftell(fp);

  // read into buf and close
  buf = new char[len]; //(char*)malloc(len);
  fseek(fp, 0, SEEK_SET);
  fread(buf, len, 1, fp);
  fclose(fp);

  // create SAFEARRAY
  SAFEARRAY *psa = SafeArrayCreateVector(VT_UI1, 0, len);
  unsigned char *pData = NULL;
  SafeArrayAccessData(psa, (void**)&pData);
  memcpy(pData, buf, len);
  SafeArrayUnaccessData(psa);

  // load sys assembly
  hr = appDomain->Load_3(psa, &assembly);
  SafeArrayDestroy(psa); // cleanup before error check
  delete buf;
  if (FAILED(hr)) return err(hr, "Could not load sys.dll assembly");

  // get correct type
  _Type* type;
  bstr = SysAllocString(mainClassName); //L"Fanx.Tools.Fan");
  hr = assembly->GetType_2(bstr, &type);
  SysFreeString(bstr);
  if (FAILED(hr)) return err(hr, "Could not load type ", mainClassName);

  // create instance of type
  VARIANT target;
  bstr = SysAllocString(mainClassName); //L"Fanx.Tools.Fan");
  hr = assembly->CreateInstance(bstr, &target);
  SysFreeString(bstr);
  if (FAILED(hr)) return err(hr, "Could not create instance of ", mainClassName);

  // invoke the Run() method
  VARIANT retVal;
  bstr = SysAllocString(L"Run");
  hr = type->InvokeMember_3(bstr, BindingFlags_InvokeMethod, NULL, target, NULL, &retVal);
  SysFreeString(bstr);
  if (FAILED(hr)) return err(hr, "Could not invoke Run() on ", mainClassName);

  delete[] mainClassName;
  return 0;
}

//////////////////////////////////////////////////////////////////////////
// Unload the CLR
//////////////////////////////////////////////////////////////////////////

int unloadClr()
{
  // stop CLR
  HRESULT hr = clr->Stop();
  if (FAILED(hr)) return err("Could not stop CLR");
  if (debug) printf("--   clr.stop\n");
  return 0;
}

//////////////////////////////////////////////////////////////////////////
// .NET Main
//////////////////////////////////////////////////////////////////////////

int runNet()
{
  if (loadClr())   return -1;
  if (loadSys())   return -1;
  if (unloadClr()) return -1;
  return 0;
}
