//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   24 Sep 06  Andy Frank  Creation
//

using System;
using System.IO;

namespace Fan.Sys
{
  /// <summary>
  /// Err.
  /// </summary>
  public class Err : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Java to Fan Mapping
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Map a java exception to it's Fan Err counter part.  Common runtime
    /// exceptions are mapped into explicit Fan types.  Otherwise we just
    /// wrap the exception with a generic Err.
    /// </summary>
    public static Err Make(Exception ex)
    {
      // NOTE: everything in this list must be synchronized
      // with the mapping used below for FCodeEmit.errTable()
      // and tested in TryTest
      if (ex is Val)                       return ((Val)ex).err;
      if (ex is NullReferenceException)    return new NullErr(ex);
      //if (ex is InvalidCastException)      return new CastErr(ex);
      if (ex is IndexOutOfRangeException)  return new IndexErr(ex);
      //if (ex is IllegalArgumentException)  return new ArgErr(ex);
      //if (ex is UnsupportedOperationException)  return new UnsupportedErr(ex);
      //if (ex is IOException)               return new IOErr(ex);
      return new Err(new Err.Val(), ex);
    }

    /*/
     * This method is used by FCodeEmit to generate extra entries in the
     * exception table - for example if fcode says to trap NullErr, then
     * we also need to trap java.lang.NullPointerException.  Basically this
     * is the inverse of the mapping done in make(Throwable).
     */
    /*
    public static String fanToJava(String jtype)
    {
      if (jtype.equals("fan/sys/NullErr"))  return "java/lang/NullPointerException";
      if (jtype.equals("fan/sys/CastErr"))  return "java/lang/ClassCastException";
      if (jtype.equals("fan/sys/IndexErr")) return "java/lang/IndexOutOfBoundsException";
      if (jtype.equals("fan/sys/ArgErr"))   return "java/lang/IllegalArgumentException";
      if (jtype.equals("fan/sys/UnsupportedErr")) return "java/lang/UnsupportedOperationException";
      if (jtype.equals("fan/sys/IOErr"))    return "java/io/IOException";
      return null;
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // C# Convenience
  //////////////////////////////////////////////////////////////////////////

    public static Err Make(string msg) { return Make(Str.Make(msg)); }
    public static Err Make(string msg, Exception e) { return Make(Str.Make(msg)); }

  //////////////////////////////////////////////////////////////////////////
  // Fan Constructors
  //////////////////////////////////////////////////////////////////////////

    public static Err Make() { return Make((Str)null, (Err)null); }
    public static Err Make(Str msg) { return Make(msg, null); }
    public static Err Make(Str msg, Err cause)
    {
      Err err = new Err(new Err.Val());
      Make_(err, msg, cause);
      return err;
    }

    public static void Make_(Err self) { Make_(self, null);  }
    public static void Make_(Err self, Str msg) { Make_(self, msg, null); }
    public static void Make_(Err self, Str msg, Err cause)
    {
      self.message = msg;
      self.cause   = cause;
    }

  //////////////////////////////////////////////////////////////////////////
  // C# Constructors
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// All subclasses must call this constructor with their
    /// typed Val exception which the real Java exception we use.
    /// </summary>
    public Err(Val val)
    {
      this.val = val;
      val.err = this;
    }

    /// <summary>
    /// This constructor is used by special subclasses which provide
    /// a transparent mapping between Java and Fan exception types.
    /// </summary>
    public Err(Val val, Exception actual)
    {
      this.val = val;
      val.err = this;
      this.actual = actual;
      this.message = Str.Make(actual.ToString());
    }

  //////////////////////////////////////////////////////////////////////////
  // Methods
  //////////////////////////////////////////////////////////////////////////

    public Str Message()
    {
      return message;
    }

    public Err Cause()
    {
      return cause;
    }

    public void Trace()
    {
      if (actual != null)
        Console.WriteLine(actual.ToString());
      else
        Console.WriteLine(val.ToString());
      if (cause != null) cause.Trace();   // TODO - how to handle?
    }

    public void Trace(Stream output)
    {
      StreamWriter writer = new StreamWriter(output);
      if (actual != null)
        Console.WriteLine(actual.ToString());
      else
        writer.WriteLine(val.ToString());
      writer.Flush();
      if (cause != null) cause.Trace();  // TODO - how to handle?
    }

    public override Type Type()
    {
      return Sys.ErrType;
    }

    public override Str ToStr()
    {
      if (message == null)
        return Type().Qname().ToStr();
      else
        return Str.Make(Type().Qname() + ": " + message);
    }

  //////////////////////////////////////////////////////////////////////////
  // Val
  //////////////////////////////////////////////////////////////////////////

    public class Val : Exception
    {
      public override string Message { get { return err.ToString(); }}
      public Err Err() { return err; }
      internal Err err;
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public readonly Val val;
    internal Str message;
    internal Err cause = null;
    internal Exception actual;

  }
}