//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   13 Sep 06  Andy Frank  Creation
//

using System;

namespace Fan.Sys
{
  ///
  /// Float is a 64-bit floating point value.
  ///
  public sealed class Float : Num
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    public static Float Parse(Str s) { return Parse(s, Bool.True); }
    public static Float Parse(Str s, Bool check)
    {
      try
      {
        string x = s.val;
        if (x == "NaN")  return nan;
        if (x == "+Inf") return posInf;
        if (x == "-Inf") return negInf;
        return Make(Double.Parse(x));
      }
      catch (FormatException)
      {
        if (!check.val) return null;
        throw ArgErr.Make("Invalid Float: '" + s + "'").val;
      }
    }

    public static Float Make(double val)
    {
      if (val == 0) return zero;
      return new Float(val);
    }

    public static Float MakeBits(Int bits)
    {
      return Make(BitConverter.Int64BitsToDouble(bits.val));
    }

    public static Float MakeBits32(Int bits)
    {
      return Make(BitConverter.ToSingle(BitConverter.GetBytes(bits.val), 0));
    }

    public static Float PosInf() { return posInf; }

    public static Float NegInf() { return negInf; }

    public static Float NaN() { return nan; }

    public static Float Pi() { return pi; }

    public static Float E() { return e; }

    private Float(double val) { this.val = val; }

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override bool Equals(Object obj)
    {
      return val == (obj as Float).val;
    }

    public override Bool Equals(Obj obj)
    {
      if (obj is Float)
      {
        double x = ((Float)obj).val;
        if (Double.IsNaN(val)) return Bool.Make(Double.IsNaN(x));
        return val == x ? Bool.True : Bool.False;
      }
      return Bool.False;
    }

    public Bool Approx(Float that) { return Approx(that, null); }
    public Bool Approx(Float that, Float tolerance)
    {
      // need this to check +inf, -inf, and nan
      if (Equals(that).val) return Bool.True;

      double t;
      if (tolerance == null)
        t = Math.Min( Math.Abs(val/1e6), Math.Abs(that.val/1e6) );
      else
        t = tolerance.val;
      return Math.Abs(val - that.val) <= t ? Bool.True : Bool.False;
    }

    public override Int Compare(Obj obj)
    {
      double that = ((Float)obj).val;
      if (Double.IsNaN(val))
      {
        return (Double.IsNaN(that)) ? Int.EQ : Int.LT;
      }
      else if (Double.IsNaN(that))
      {
        return Int.GT;
      }
      else
      {
        if (val < that) return Int.LT; return val == that ? Int.EQ : Int.GT;
      }
    }

    public override int GetHashCode()
    {
      long hash = BitConverter.DoubleToInt64Bits(val);
      //return (int)(hash ^ (hash >>> 32));
      return (int)(hash ^ (hash >> 32));
    }

    public override Int Hash()
    {
      return Bits();
    }

    public Int Bits()
    {
      return Int.Make(BitConverter.DoubleToInt64Bits(val));
    }

    public Int Bits32()
    {
      return Int.Make(BitConverter.ToInt32(BitConverter.GetBytes((float)val), 0));
    }

    public override Type Type()
    {
      return Sys.FloatType;
    }

  //////////////////////////////////////////////////////////////////////////
  // Operators
  //////////////////////////////////////////////////////////////////////////

    public Float Negate() { return Make(-val); }
    public Float Star      (Float x) { return Make(val * x.val); }
    public Float Slash     (Float x) { return Make(val / x.val); }
    public Float Percent   (Float x) { return Make(val % x.val); }
    public Float Plus      (Float x) { return Make(val + x.val); }
    public Float Minus     (Float x) { return Make(val - x.val); }
    public Float Increment ()        { return Make(val+1); }
    public Float Decrement ()        { return Make(val-1); }

  //////////////////////////////////////////////////////////////////////////
  // Num
  //////////////////////////////////////////////////////////////////////////

    public override Int ToInt()
    {
      // TODO
      if (val == Double.PositiveInfinity) return Int.Make(Int64.MaxValue);
      if (Double.IsNaN(val)) return Int.Zero;
      return Int.Make((long)val);
    }

    public override Float ToFloat() { return this; }

  //////////////////////////////////////////////////////////////////////////
  // Math
  //////////////////////////////////////////////////////////////////////////

    public Float Abs()
    {
      if (val >= 0) return this;
      return Make(-val);
    }

    public Float Min(Float that)
    {
      if (val <= that.val) return this;
      return that;
    }

    public Float Max(Float that)
    {
      if (val >= that.val) return this;
      return that;
    }

    public Float Ceil()
    {
      return Make(Math.Ceiling(val));
    }

    public Float Floor()
    {
      return Make(Math.Floor(val));
    }

    public Float Round()
    {
      return Make(Math.Round(val));
    }

    public Float Exp()
    {
      return Make(Math.Exp(val));
    }

    public Float Log()
    {
      return Make(Math.Log(val));
    }

    public Float Log10()
    {
      return Make(Math.Log10(val));
    }

    public Float Pow(Float pow)
    {
      return Make(Math.Pow(val, pow.val));
    }

    public Float Sqrt()
    {
      return Make(Math.Sqrt(val));
    }

  //////////////////////////////////////////////////////////////////////////
  // Trig
  //////////////////////////////////////////////////////////////////////////

    public Float Acos()
    {
      return Make(Math.Acos(val));
    }

    public Float Asin()
    {
      return Make(Math.Asin(val));
    }

    public Float Atan()
    {
      return Make(Math.Atan(val));
    }

    public static Float Atan2(Float y, Float x)
    {
      return Make(Math.Atan2(y.val, x.val));
    }

    public Float Cos()
    {
      return Make(Math.Cos(val));
    }

    public Float Cosh()
    {
      return Make(Math.Cosh(val));
    }

    public Float Sin()
    {
      return Make(Math.Sin(val));
    }

    public Float Sinh()
    {
      return Make(Math.Sinh(val));
    }

    public Float Tan()
    {
      return Make(Math.Tan(val));
    }

    public Float Tanh()
    {
      return Make(Math.Tanh(val));
    }

    public Float ToDegrees()
    {
      return Make((180 / Math.PI) * val);
    }

    public Float ToRadians()
    {
      return Make((val * Math.PI) / 180);
    }

  //////////////////////////////////////////////////////////////////////////
  // Conversion
  //////////////////////////////////////////////////////////////////////////

    public override Str ToStr()
    {
      if (Double.IsNaN(val)) return NaNStr;
      if (val == Double.PositiveInfinity) return PosInfStr;
      if (val == Double.NegativeInfinity) return NegInfStr;
      string s = "" + val;
      if (s.IndexOf('.') == -1) s += ".0";  // to match java behavior
      return Str.Make(s);
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public static readonly Float zero   = new Float(0);
    public static readonly Float posInf = new Float(Double.PositiveInfinity);
    public static readonly Float negInf = new Float(Double.NegativeInfinity);
    public static readonly Float nan    = new Float(Double.NaN);
    public static readonly Float e      = new Float(Math.E);
    public static readonly Float pi     = new Float(Math.PI);
    public static readonly Str PosInfStr = Str.Make("+Inf");
    public static readonly Str NegInfStr = Str.Make("-Inf");
    public static readonly Str NaNStr    = Str.Make("NaN");

    public readonly double val;
  }
}