//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Dec 06  Andy Frank  Creation
//

using System.Collections;
using Fanx.Fcode;

namespace Fan.Sys
{
  /// <summary>
  /// GenericType is the base class for ListType, MapType, and MethodType
  /// which all support parameterization of the generic parameter types (such
  /// as A-H, V, K).  Instances of GenericType are used to represent generic
  /// instances (for example an instance of ListType is used to represent Str[]).
  /// </summary>
  public abstract class GenericType : Type
  {

  //////////////////////////////////////////////////////////////////////////
  // Variegate
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// This method is done during fcode reflect to variegate a variant type (List,
    /// Map, Method).  At this point we parameterize the variance types from the
    /// specified tuple binary fcode representation.
    /// </summary>
    static Type Variegate(Pod pod, Type type, FTuple ftuple)
    {
      Variances var = new Variances().Load(pod, ftuple.val);
      if (type == Sys.ListType) return var.v.ToListOf();
      throw Err.Make("Not variant type: " + type).val;
    }

    class Variances
    {
     public Variances Load(Pod pod, int[] val)
     {
       int varCount = (val.Length-2)/2;
       for (int i=0; i<varCount; i++)
         Load(pod, val[i*2+2], val[i*2+3]);
       return this;
     }

     public void Load(Pod pod, int id, int typeRef)
     {
       Type type = pod.FindType(typeRef);
       switch (id)
       {
         case 'A': a = type; break;
         case 'B': b = type; break;
         case 'C': c = type; break;
         case 'D': d = type; break;
         case 'E': e = type; break;
         case 'F': f = type; break;
         case 'G': g = type; break;
         case 'H': h = type; break;
         case 'K': k = type; break;
         case 'R': r = type; break;
         case 'V': v = type; break;
         default: throw Err.Make("Unknown variance " + (char)id + " [" + id + "]").val;
       }
     }

     internal Type a, b, c, d, e, f, g, h, k, r, v;
    }

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    internal GenericType(Type baseType) :
      base(baseType.pod, baseType.name.val, baseType.flags, baseType.facets)
    {
    }

  //////////////////////////////////////////////////////////////////////////
  // Type
  //////////////////////////////////////////////////////////////////////////

    public override abstract Type Base();

    public override List Mixins()
    {
      return Base().Mixins();
    }

    public override abstract Str Signature();

    public override bool IsGenericInstance() { return true; }

    public override bool Is(Type type)
    {
      if (type == this || type == Base()) return true;
      return Base().Is(type);
    }

    public Map Params()
    {
      if (pars == null) pars = MakeParams();
      return pars;
    }

    internal abstract Map MakeParams();

  //////////////////////////////////////////////////////////////////////////
  // Reflect
  //////////////////////////////////////////////////////////////////////////

    /**
     * On reflection, we parameterize the master's methods.
     */
    protected override void DoReflect()
    {
      // ensure master type is reflected
      Type master = Base();
      master.Reflect();
      List masterSlots = master.slots;

      // allocate slot data structures
      fields = new List(Sys.FieldType, master.fields.Sz());
      methods = new List(Sys.MethodType, master.methods.Sz());
      slots = new List(Sys.SlotType, masterSlots.Sz());
      slotsByName = new Hashtable(masterSlots.Sz()*3);

      // parameterize master's slots
      for (int i=0; i<masterSlots.Sz(); i++)
      {
        Slot slot = (Slot)masterSlots.Get(i);
        if (slot is Method)
        {
          slot = Parameterize((Method)slot);
          methods.Add(slot);
        }
        else
        {
          fields.Add(slot);
        }
        slots.Add(slot);
        slotsByName[slot.name.val] = slot;
      }
    }

    /**
     * Parameterize the specified method (if reuse if generic
     * parameterization isn't necessary).
     */
    internal Method Parameterize(Method m)
    {
      // if not generic, short circuit and reuse original
      if (!m.IsGenericMethod()) return m;

      // new signature
      Type ret;
      List pars = new List(Sys.MethodParamType, m.pars.Sz());

      // parameterize return type
      if (m.returnType.IsGenericParameter())
        ret = Parameterize(m.returnType);
      else
        ret = m.returnType;

      // narrow pars (or just reuse if not parameterized)
      for (int i=0; i<m.pars.Sz(); i++)
      {
        MethodParam p = (MethodParam)m.pars.Get(i);
        if (p.paramType.IsGenericParameter())
        {
          pars.Add(new MethodParam(p.name, Parameterize(p.paramType), p.mask));
        }
        else
        {
          pars.Add(p);
        }
      }

      return new Method(this, m.name, m.flags, m.facets, ret, m.inheritedReturnType, pars, m);
    }

    /**
     * Parameterize t, where t is a generic parameter type such as V.
     */
    internal Type Parameterize(Type t)
    {
      if (t is ListType)
        return ParameterizeListType((ListType)t);
      else if (t is MethodType)
        return ParameterizeMethodType((MethodType)t);
      else
        return DoParameterize(t);
    }

    /**
     * Recursively parameterize the a generic list type.
     */
    internal Type ParameterizeListType(ListType t)
    {
      return DoParameterize(t.v).ToListOf();
    }

    /**
     * Recursively parameterize the pars of a method type.
     */
    internal MethodType ParameterizeMethodType(MethodType t)
    {
      Type[] pars = new Type[t.pars.Length];
      for (int i=0; i<pars.Length; i++)
      {
        Type param = t.pars[i];
        if (param.IsGenericParameter()) param = DoParameterize(param);
        pars[i] = param;
      }

      Type ret = t.ret;
      if (ret.IsGenericParameter()) ret = DoParameterize(ret);

      return new MethodType(pars, ret);
    }

    /**
     * Parameterize t, where t is a generic parameter type such as V.
     */
    protected abstract Type DoParameterize(Type t);

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    private Map pars;
  }

}