//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   13 Sep 06  Andy Frank  Creation
//

using System;

namespace Fan.Sys
{
  /// <summary>
  /// Int is a 64-bit integer value.
  /// </summary>
  public sealed class Int : Num
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    public static Int Parse(Str s) { return Parse(s, Ten, Bool.True); }
    public static Int Parse(Str s, Int radix) { return Parse(s, radix, Bool.True); }
    public static Int Parse(Str s, Int radix, Bool check)
    {
      try
      {
        return Make(Convert.ToInt64(s.val, (int)radix.val));
      }
      catch (FormatException)
      {
        if (!check.val) return null;
        throw ArgErr.Make("Invalid Int: '" + s + "'").val;
      }
    }

    public static Int Make(long val)
    {
      if (val < POS)
      {
        if (val >= 0) return pos[(int)val];
        if (val > NEG) return neg[-(int)val];
      }
      return new Int(val);
    }

    public static Int Pos(long val)
    {
      if (val < POS)
      {
        return pos[(int)val];
      }
      return new Int(val);
    }

    private Int(long val) { this.val = val; }

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override bool Equals(Object obj)
    {
      return val == (obj as Int).val;
    }

    public override Bool Equals(Obj obj)
    {
      if (obj is Int)
        return val == ((Int)obj).val ? Bool.True : Bool.False;
      else
        return Bool.False;
    }

    public override Int Compare(Obj obj)
    {
      long that = ((Int)obj).val;
      if (val < that) return LT; return val == that ? EQ : GT;
    }

    public override int GetHashCode()
    {
      //return (int)(val ^ (int)(((ulong)val) >> 32));
      return (int)(val ^ (int)(((ulong)val) >> 32));
    }

    public override Int Hash()
    {
      return this;
    }

    public override Type Type()
    {
      return Sys.IntType;
    }

  //////////////////////////////////////////////////////////////////////////
  // Operators
  //////////////////////////////////////////////////////////////////////////

    public Int Negate()  { return Make(-val); }
    public Int Tilde()   { return Make(~val); }
    public Int Star      (Int x) { return Make(val * x.val); }
    public Int Slash     (Int x) { return Make(val / x.val); }
    public Int Percent   (Int x) { return Make(val % x.val); }
    public Int Plus      (Int x) { return Make(val + x.val); }
    public Int Minus     (Int x) { return Make(val - x.val); }
    public Int Amp       (Int x) { return Make(val & x.val); }
    public Int Pipe      (Int x) { return Make(val | x.val); }
    public Int Caret     (Int x) { return Make(val ^ x.val); }
    public Int Lshift    (Int x) { return Make(val << (int)x.val); }
    public Int Rshift    (Int x) { return Make(val >> (int)x.val); }
    public Int Increment ()      { return Make(val+1); }
    public Int Decrement ()      { return Make(val-1); }

  //////////////////////////////////////////////////////////////////////////
  // Num
  //////////////////////////////////////////////////////////////////////////

    public override Int ToInt() { return this; }

    public override Float ToFloat() { return Float.Make(val); }

  //////////////////////////////////////////////////////////////////////////
  // Math
  //////////////////////////////////////////////////////////////////////////

    public Int Abs()
    {
      if (val >= 0) return this;
      return Int.Pos(-val);
    }

    public Int Min(Int that)
    {
      if (val <= that.val) return this;
      return that;
    }

    public Int Max(Int that)
    {
      if (val >= that.val) return this;
      return that;
    }

    public Bool IsEven()
    {
      return (val % 2) == 0 ? Bool.True : Bool.False;
    }

    public Bool IsOdd()
    {
      return (val % 2) != 0 ? Bool.True : Bool.False;
    }

  /////////////////////////////////////////////////////////////////////////
  // Char
  //////////////////////////////////////////////////////////////////////////

    public Bool IsSpace()
    {
      try
      {
        if (val < 128 && (charMap[(int)val] & SPACE) != 0)
          return Bool.True;
        else
          return Bool.False;
      }
      catch (IndexOutOfRangeException)
      {
        // should be very rare to use this method with negative
        // numbers, so don't take the hit every call
        return Bool.False;
      }
    }

    public Bool IsAlpha()
    {
      try
      {
        if (val < 128 && (charMap[(int)val] & (UPPER|LOWER)) != 0)
          return Bool.True;
        else
          return Bool.False;
      }
      catch (IndexOutOfRangeException)
      {
        // should be very rare to use this method with negative
        // numbers, so don't take the hit every call
        return Bool.False;
      }
    }

    public Bool IsAlphaNum()
    {
      try
      {
        if (val < 128 && (charMap[(int)val] & (UPPER|LOWER|DIGIT)) != 0)
          return Bool.True;
        else
          return Bool.False;
      }
      catch (IndexOutOfRangeException)
      {
        // should be very rare to use this method with negative
        // numbers, so don't take the hit every call
        return Bool.False;
      }
    }

    public Bool IsUpper()
    {
      return 'A' <= val && val <= 'Z' ? Bool.True : Bool.False;
    }

    public Bool IsLower()
    {
      return 'a' <= val && val <= 'z' ? Bool.True : Bool.False;
    }

    public Int Upper()
    {
      if ('a' <= val && val <= 'z')
        return pos[((int)val) & ~0x20];
      else
        return this;
    }

    public Int Lower()
    {
      if ('A' <= val && val <= 'Z')
        return pos[((int)val) | 0x20];
      else
        return this;
    }

    public Bool IsDigit()
    {
      return '0' <= val && val <= '9' ? Bool.True : Bool.False;
    }

    public Bool IsDigit(Int radix)
    {
      if (val < 0 || val >= 128) return Bool.False;
      int v = (int)val;
      int r = (int)radix.val;

      if (r == 10)
      {
        return ((charMap[v] & DIGIT) != 0) ? Bool.True : Bool.False;
      }

      if (r == 16)
      {
        return ((charMap[v] & HEX) != 0) ? Bool.True : Bool.False;
      }

      if (r <= 10)
      {
        return '0' <= v && v <= ('0'+r) ?  Bool.True : Bool.False;
      }
      else
      {
        if ((charMap[v] & DIGIT) != 0) return Bool.True;
        int x = v-10;
        if ('a' <= v && v <= 'a'+x) return Bool.True;
        if ('A' <= v && v <= 'A'+x) return Bool.True;
        return Bool.False;
      }
    }

    public Int ToDigit()
    {
      if (0 <= val && val <= 9) return pos[(int)val + '0'];
      return null;
    }

    public Int ToDigit(Int radix)
    {
      int val = (int)this.val;
      int r   = (int)radix.val;
      if (val < 0 || val >= r) return null;

      if (val < 10) return pos[val + '0'];
      return pos[val - 10 + 'a'];
    }

    public Int FromDigit()
    {
      if ('0' <= val && val <= '9') return pos[(int)val - '0'];
      return null;
    }

    public Int FromDigit(Int radix)
    {
      if (val < 0 || val >= 128) return null;
      int v = (int)val;
      int r = (int)radix.val;

      int ten = r < 10 ? r : 10;
      if ('0' <= v && v < '0'+ten) return pos[v - '0'];
      if (r > 10)
      {
        int alpha = r-10;
        if ('a' <= v && v < 'a'+alpha) return pos[v + 10 - 'a'];
        if ('A' <= v && v < 'A'+alpha) return pos[v + 10 - 'A'];
      }
      return null;
    }

    private static readonly byte[] charMap = new byte[128];
    private static readonly byte SPACE = 0x01;
    private static readonly byte UPPER = 0x02;
    private static readonly byte LOWER = 0x04;
    private static readonly byte DIGIT = 0x08;
    private static readonly byte HEX   = 0x10;
    // static initializer below

  //////////////////////////////////////////////////////////////////////////
  // Conversion
  //////////////////////////////////////////////////////////////////////////

    /*
    public Str ToChar()
    {
      if (val < 0 || val > 0xFFFF) throw Err.make("Invalid unicode char: " + val).val;
      if (val < Str.ascii.length) return Str.ascii[(int)val];
      return Str.make(String.valueOf((char)val));
    }
    */

    public Str ToHex()
    {
      return Str.Make(val.ToString("X").ToLower());
    }

    public override Str ToStr()
    {
      if (str == null) str = Str.Make("" + val);
      return str;
    }

  //////////////////////////////////////////////////////////////////////////
  // Closures
  //////////////////////////////////////////////////////////////////////////

    public void Times(Method closure)
    {
      for (long i=0; i<val; i++)
        closure.Call1(Int.Make(i));
    }

  //////////////////////////////////////////////////////////////////////////
  // Intern
  //////////////////////////////////////////////////////////////////////////

    public static readonly long NEG = -1025;
    public static readonly long POS = 65537;
    public static readonly Int[] neg = new Int[-(int)NEG];
    public static readonly Int[] pos = new Int[(int)POS];
    public static readonly Int Zero;    // 0
    public static readonly Int One;     // 1
    public static readonly Int NegOne;  // -1
    public static readonly Int Ten;     // 10
    public static readonly Int LT;      // compare() -> -1
    public static readonly Int EQ;      // compare() ->  0
    public static readonly Int GT;      // compare() ->  1

    static Int()
    {
      charMap[' ']  |= SPACE;
      charMap['\n'] |= SPACE;
      charMap['\r'] |= SPACE;
      charMap['\t'] |= SPACE;
      charMap['\f'] |= SPACE;

      // alpha characters
      for (int i='a'; i<='z'; i++) charMap[i] |= LOWER;
      for (int i='A'; i<='Z'; i++) charMap[i] |= UPPER;

      // digit characters
      for (int i='0'; i<='9'; i++) charMap[i] |= DIGIT;

      // hex characters
      for (int i='0'; i<='9'; i++) charMap[i] |= HEX;
      for (int i='a'; i<='f'; i++) charMap[i] |= HEX;
      for (int i='A'; i<='F'; i++) charMap[i] |= HEX;

      // Intern
      for (int i=1; i<neg.Length; i++) neg[i] = new Int(-i);
      for (int i=0; i<pos.Length; i++) pos[i] = new Int(i);
      Zero   = Make(0);
      One    = Make(1);
      NegOne = Make(-1);
      Ten    = Make(10);
      LT     = Make(-1);
      EQ     = Make(0);
      GT     = Make(1);
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public readonly long val;
    private Str str;
  }
}