//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Sep 06  Andy Frank  Creation
//

using System;
using System.Reflection;
using System.Text;
using Fanx.Fcode;
using Fanx.Util;

namespace Fan.Sys
{
  /// <summary>
  /// Method is an invocable operation on a Type.
  /// </summary>
  public class Method : Slot
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Constructor used by Type.reflect.
    /// </summary>
    public Method(Type parent, Str name, int flags, Facets facets, Type returnType, Type inheritedReturnType, List pars)
      : base(parent, name, flags, facets)
    {
      this.returnType = returnType;
      this.inheritedReturnType = inheritedReturnType;
      this.pars       = pars;
      this.mask       = ToMask(parent, returnType, pars);
      this.generic    = null;
    }

    /// <summary>
    /// Constructor used by GenericType and we are given the generic
    /// method that is being parameterized.
    /// </summary>
    public Method(Type parent, Str name, int flags, Facets facets, Type returnType, Type inheritedReturnType, List pars, Method generic)
      : base(parent, name, flags, facets)
    {
      this.returnType = returnType;
      this.inheritedReturnType = inheritedReturnType;
      this.pars       = pars;
      this.mask       = 0;
      this.generic    = generic;
    }

    /// <summary>
    /// Compute if the method signature contains generic parameter types.
    /// </summary>
    private static int ToMask(Type parent, Type returnType, List pars)
    {
      // we only use generics in Sys
      if (parent.Pod() != Sys.SysPod) return 0;

      int p = returnType.IsGenericParameter() ? 1 : 0;
      for (int i=0; i<pars.Sz(); i++)
        p |= ((MethodParam)pars.Get(i)).paramType.IsGenericParameter() ? 1 : 0;

      int mask = 0;
      if (p != 0) mask |= GENERIC;
      return mask;
    }

  //////////////////////////////////////////////////////////////////////////
  // Methods
  //////////////////////////////////////////////////////////////////////////

    public override Type Type() { return Sys.MethodType; }

    public Type ReturnType() { return returnType; }

    public List Params() { return pars.RO(); }

    public override Str Signature()
    {
      StringBuilder s = new StringBuilder();
      s.Append(ReturnType()).Append(' ').Append(name).Append('(');
      for (int i=0; i<pars.Sz(); i++)
      {
        if (i > 0) s.Append(", ");
        MethodParam p = (MethodParam)pars.Get(i);
        s.Append(p.ParamType()).Append(' ').Append(p.Name());
      }
      s.Append(')');
      return Str.Make(s.ToString());
    }

  //////////////////////////////////////////////////////////////////////////
  // Generics
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Return if this method contains generic parameters in it's signature.
    /// </summary>
    public bool IsGenericMethod()
    {
      return (mask & GENERIC) != 0;
    }

    /// <summary>
    /// Return if this method is the parameterization of a generic method,
    /// with all the generic parameters filled in with real types.
    /// </summary>
    public bool IsGenericInstance()
    {
      return generic != null;
    }

    /// <summary>
    /// If isGenericInstance is true, then return the generic method which
    /// this method instantiates.  The generic method may be used to access
    /// the actual signatures used in the Java code (via getRawType).  If
    /// this method is not a generic instance, return null.
    /// </summary>
    public Method GetGenericMethod()
    {
      return generic;
    }

  //////////////////////////////////////////////////////////////////////////
  // Invoke
  //////////////////////////////////////////////////////////////////////////

    // TODO - support for calling ctor via reflection
    public virtual Obj Call(List args)
    {
      /*
      bool isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
      CheckArgs(args, isStatic);
      if (isStatic)
      {
        Obj[] a = new Obj[pars.Sz()];
        if (args != null && a.Length > 0) args.ToArray(a, 0, a.Length);
        return Invoke(null, a);
      }
      else
      {
        Obj i = null;
        Obj[] a = new Obj[pars.Sz()];
        if (args != null)
        {
          i = args.Get(0);
          if (a.Length > 0) args.ToArray(a, 1, a.Length);
        }
        return Invoke(i, a);
      }
      */

      int argsSize = args == null ? 0 : args.Sz();

      bool isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
      int p = CheckArgs(argsSize, isStatic);
      Obj[] a = new Obj[p];

      if (isStatic)
      {
        if (args != null && a.Length > 0) args.ToArray(a, 0, a.Length);
        return Invoke(null, a);
      }
      else
      {
        Obj i = args.Get(0);
        if (a.Length > 0) args.ToArray(a, 1, a.Length);
        return Invoke(i, a);
      }

    }

    public Obj CallOn(Obj target, List args)
    {
      int argsSize = args == null ? 0 : args.Sz();

      // from checkArgs perspective we pass true for isStatic
      // because we don't include target as part of arguments
      int p = CheckArgs(argsSize, true);

      Obj[] a = new Obj[p];
      if (args != null && a.Length > 0) args.ToArray(a, 0, a.Length);
      return Invoke(target, a);
    }

    public virtual Obj Call0()
    {
      bool isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
      CheckArgs(0, isStatic);
      return Invoke(null, noArgs);
    }

    public virtual Obj Call1(Obj a)
    {
      bool isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
      CheckArgs(1, isStatic);
      int p = pars.Sz();
      Obj[] args = new Obj[p];
      if (isStatic)
      {
        switch (p)
        {
          case 1: args[0] = a; break;
        }
        return Invoke(null, args);
      }
      else
      {
        return Invoke(a, args);
      }
    }

    public virtual Obj Call2(Obj a, Obj b)
    {
      bool isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
      CheckArgs(2, isStatic);
      int p = pars.Sz();
      Obj[] args = new Obj[p];
      if (isStatic)
      {
        switch (p)
        {
          case 2: args[1] = b; goto case 1;
          case 1: args[0] = a; break;
        }
        return Invoke(null, args);
      }
      else
      {
        switch (p)
        {
          case 1: args[0] = b; break;
        }
        return Invoke(a, args);
      }
    }

    public virtual Obj Call3(Obj a, Obj b, Obj c)
    {
      bool isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
      CheckArgs(3, isStatic);
      int p = pars.Sz();
      Obj[] args = new Obj[p];
      if (isStatic)
      {
        switch (p)
        {
          case 3: args[2] = c; goto case 2;
          case 2: args[1] = b; goto case 1;
          case 1: args[0] = a; break;
        }
        return Invoke(null, args);
      }
      else
      {
        switch (p)
        {
          case 2: args[1] = c; goto case 1;
          case 1: args[0] = b; break;
        }
        return Invoke(a, args);
      }
    }

    public virtual Obj Call4(Obj a, Obj b, Obj c, Obj d)
    {
      bool isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
      CheckArgs(4, isStatic);
      int p = pars.Sz();
      Obj[] args = new Obj[p];
      if (isStatic)
      {
        switch (p)
        {
          case 4: args[3] = d; goto case 3;
          case 3: args[2] = c; goto case 2;
          case 2: args[1] = b; goto case 1;
          case 1: args[0] = a; break;
        }
        return Invoke(null, args);
      }
      else
      {
        switch (p)
        {
          case 3: args[2] = d; goto case 2;
          case 2: args[1] = c; goto case 1;
          case 1: args[0] = b; break;
        }
        return Invoke(a, args);
      }
    }

    public virtual Obj Call5(Obj a, Obj b, Obj c, Obj d, Obj e)
    {
      bool isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
      CheckArgs(5, isStatic);
      int p = pars.Sz();
      Obj[] args = new Obj[p];
      if (isStatic)
      {
        switch (p)
        {
          case 5: args[4] = e; goto case 4;
          case 4: args[3] = d; goto case 3;
          case 3: args[2] = c; goto case 2;
          case 2: args[1] = b; goto case 1;
          case 1: args[0] = a; break;
        }
        return Invoke(null, args);
      }
      else
      {
        switch (p)
        {
          case 4: args[3] = e; goto case 3;
          case 3: args[2] = d; goto case 2;
          case 2: args[1] = c; goto case 1;
          case 1: args[0] = b; break;
        }
        return Invoke(a, args);
      }
    }

    public virtual Obj Call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f)
    {
      bool isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
      CheckArgs(6, isStatic);
      int p = pars.Sz();
      Obj[] args = new Obj[p];
      if (isStatic)
      {
        switch (p)
        {
          case 6: args[5] = f; goto case 5;
          case 5: args[4] = e; goto case 4;
          case 4: args[3] = d; goto case 3;
          case 3: args[2] = c; goto case 2;
          case 2: args[1] = b; goto case 1;
          case 1: args[0] = a; break;
        }
        return Invoke(null, args);
      }
      else
      {
        switch (p)
        {
          case 5: args[4] = f; goto case 4;
          case 4: args[3] = e; goto case 3;
          case 3: args[2] = d; goto case 2;
          case 2: args[1] = c; goto case 1;
          case 1: args[0] = b; break;
        }
        return Invoke(a, args);
      }
    }

    public virtual Obj Call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g)
    {
      bool isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
      CheckArgs(7, isStatic);
      int p = pars.Sz();
      Obj[] args = new Obj[p];
      if (isStatic)
      {
        switch (p)
        {
          case 7: args[6] = g; goto case 6;
          case 6: args[5] = f; goto case 5;
          case 5: args[4] = e; goto case 4;
          case 4: args[3] = d; goto case 3;
          case 3: args[2] = c; goto case 2;
          case 2: args[1] = b; goto case 1;
          case 1: args[0] = a; break;
        }
        return Invoke(null, args);
      }
      else
      {
        switch (p)
        {
          case 6: args[5] = g; goto case 5;
          case 5: args[4] = f; goto case 4;
          case 4: args[3] = e; goto case 3;
          case 3: args[2] = d; goto case 2;
          case 2: args[1] = c; goto case 1;
          case 1: args[0] = b; break;
        }
        return Invoke(a, args);
      }
    }

    public virtual Obj Call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h)
    {
      bool isStatic = (flags & (FConst.Static|FConst.Ctor)) != 0;
      CheckArgs(8, isStatic);
      int p = pars.Sz();
      Obj[] args = new Obj[p];
      if (isStatic)
      {
        switch (p)
        {
          case 8: args[7] = h; goto case 7;
          case 7: args[6] = g; goto case 6;
          case 6: args[5] = f; goto case 5;
          case 5: args[4] = e; goto case 4;
          case 4: args[3] = d; goto case 3;
          case 3: args[2] = c; goto case 2;
          case 2: args[1] = b; goto case 1;
          case 1: args[0] = a; break;
        }
        return Invoke(null, args);
      }
      else
      {
        switch (p)
        {
          case 7: args[6] = h; goto case 6;
          case 6: args[5] = g; goto case 5;
          case 5: args[4] = f; goto case 4;
          case 4: args[3] = e; goto case 3;
          case 3: args[2] = d; goto case 2;
          case 2: args[1] = c; goto case 1;
          case 1: args[0] = b; break;
        }
        return Invoke(a, args);
      }
    }

//    private void CheckArgs(List args)
//    {
//      CheckArgs(args == null ? 0 : args.Sz());
//    }

    private int CheckArgs(int args, bool isStatic)
    {
      // ensure parent has finished emitting so that reflect is populated
      parent.Finish();

      // compuate min/max parameters - reflect contains all the method versions
      // with full params at index zero, and full defaults at reflect.length-1
      int max = pars.Sz();
      int min = max; //max-reflect.Length+1;

      // do checking
      if (isStatic)
      {
        if (args < min) throw ArgErr.Make("Too few arguments: " + args + " < " + min+".."+max).val;
      }
      else
      {
        if (args < min+1) throw ArgErr.Make("Too few arguments: " + args + " < instance+" + min+".."+max).val;
      }

      // return size of arguments to pass to java method
      if (!isStatic) args--;
      if (args <= max) return args;
      return max;

//      if ((flags & (FConst.Static|FConst.Ctor)) != 0)
//      {
//        if (args < pars.Sz()) throw Err.Make("Too few arguments: " + args + " < " + pars.Sz()).val;
//      }
//      else
//      {
//        if (args < pars.Sz()+1) throw Err.Make("Too few arguments: " + args + " < instance+" + pars.Sz()).val;
//      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Reflection
  //////////////////////////////////////////////////////////////////////////

//    internal Obj Invoke(object instance, List args)
//    {
//      return Invoke(instance, args.ToArray());
//    }

    internal Obj Invoke(object instance, object[] args)
    {
      try
      {
        //parent.Finish();
        //return (Obj)reflect.Invoke(instance, (Object[])args);

        // zero index is full signature up to using max defaults
        //int index = params.sz()-args.length;
        //if (index < 0) index = 0;
        //MethodInfo m = reflect[index];

        string methName = NameUtil.Upper(name.val);

        BindingFlags flags = BindingFlags.Public | BindingFlags.InvokeMethod;
        if ((this.flags & (FConst.Static|FConst.Ctor)) != 0)
          flags |= BindingFlags.Static;
        else
          flags |= BindingFlags.Instance;

        System.Type t = (instance == null) ? parent.Emit() : instance.GetType();
        return (Obj)t.InvokeMember(methName, flags, null, instance, args);
      }
      catch (TargetInvocationException e)
      {
        if (e.GetBaseException() is Err.Val)
          throw (Err.Val)e.GetBaseException();
        else
          throw Err.Make(e.GetBaseException()).val;
      }
      catch (Exception e)
      {
        //if (reflect == null)
        //  throw Err.Make("Method not mapped to java.lang.reflect correctly").val;

        //if (args.Length != pars.Sz())
        //  throw ArgErr.Make("Invalid argument count for '" + this + "'").val;

        System.Console.WriteLine(e.StackTrace);
        throw Err.Make("Cannot call '" + this + "': " + e).val;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Indirect
  //////////////////////////////////////////////////////////////////////////

    public static readonly int MaxIndirectParams = 8;  // max callX()

    /// <summary>
    /// Constructor used by Indirect.
    /// </summary>
    protected Method(MethodType methodType)
      : base(null, null, FConst.Internal, null)
    {
      this.returnType = methodType.ret;
      this.pars       = methodType.ToMethodParams();
      this.mask       = 0;
      this.generic    = null;
    }

    /// <summary>
    /// Indirect is the base class for the IndirectX classes, which are
    /// used as the common base classes for closures and curried methods.
    /// An Indirect method takes a MethodType for it's type, and
    /// also extends Method with the call() implementation.
    /// </summary>
    public abstract class Indirect : Method
    {
      protected Indirect(MethodType type) : base(type) { this.type = type; }

      public override Str  Name()  { return Str.Make(GetType().Name); }
      public override Type Type()  { return type; }
      public override Str  ToStr() { return type.Signature(); }
      public Err TooFewArgs(int given) { throw Err.Make("Too few arguments: " + given + " < " + type.pars.Length).val; }

      MethodType type;
    }

    public abstract class Indirect0 : Indirect
    {
      protected Indirect0(MethodType type) : base(type) {}
      public override Obj Call(List args) { return Call0(); }
      public override abstract Obj Call0();
      public override Obj Call1(Obj a) { return Call0(); }
      public override Obj Call2(Obj a, Obj b) { return Call0(); }
      public override Obj Call3(Obj a, Obj b, Obj c) { return Call0(); }
      public override Obj Call4(Obj a, Obj b, Obj c, Obj d) { return Call0(); }
      public override Obj Call5(Obj a, Obj b, Obj c, Obj d, Obj e) { return Call0(); }
      public override Obj Call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { return Call0(); }
      public override Obj Call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return Call0(); }
      public override Obj Call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return Call0(); }
    }

    public abstract class Indirect1 : Indirect
    {
      protected Indirect1(MethodType type) : base(type) {}
      public override Obj Call(List args) { return Call1(args.Get(0)); }
      public override Obj Call0() { throw Err.Make("Too few arguments: 0 < 1").val; }
      public override abstract Obj Call1(Obj a);
      public override Obj Call2(Obj a, Obj b) { return Call1(a); }
      public override Obj Call3(Obj a, Obj b, Obj c) { return Call1(a); }
      public override Obj Call4(Obj a, Obj b, Obj c, Obj d) { return Call1(a); }
      public override Obj Call5(Obj a, Obj b, Obj c, Obj d, Obj e) { return Call1(a); }
      public override Obj Call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { return Call1(a); }
      public override Obj Call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return Call1(a); }
      public override Obj Call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return Call1(a); }
    }

    public abstract class Indirect2 : Indirect
    {
      protected Indirect2(MethodType type) : base(type) {}
      public override Obj Call(List args) { return Call2(args.Get(0), args.Get(1)); }
      public override Obj Call0() { throw Err.Make("Too few arguments: 0 < 2").val; }
      public override Obj Call1(Obj a) { throw Err.Make("Too few arguments: 1 < 2").val; }
      public override abstract Obj Call2(Obj a, Obj b);
      public override Obj Call3(Obj a, Obj b, Obj c) { return Call2(a, b); }
      public override Obj Call4(Obj a, Obj b, Obj c, Obj d) { return Call2(a, b); }
      public override Obj Call5(Obj a, Obj b, Obj c, Obj d, Obj e) { return Call2(a, b); }
      public override Obj Call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { return Call2(a, b); }
      public override Obj Call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return Call2(a, b); }
      public override Obj Call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return Call2(a, b); }
    }

    public abstract class Indirect3 : Indirect
    {
      protected Indirect3(MethodType type) : base(type) {}
      public override Obj Call(List args) { return Call3(args.Get(0), args.Get(1), args.Get(2)); }
      public override Obj Call0() { throw Err.Make("Too few arguments: 0 < 3").val; }
      public override Obj Call1(Obj a) { throw Err.Make("Too few arguments: 1 < 3").val; }
      public override Obj Call2(Obj a, Obj b)  { throw Err.Make("Too few arguments: 2 < 3").val; }
      public override abstract Obj Call3(Obj a, Obj b, Obj c);
      public override Obj Call4(Obj a, Obj b, Obj c, Obj d) { return Call3(a, b, c); }
      public override Obj Call5(Obj a, Obj b, Obj c, Obj d, Obj e) { return Call3(a, b, c); }
      public override Obj Call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { return Call3(a, b, c); }
      public override Obj Call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return Call3(a, b, c); }
      public override Obj Call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return Call3(a, b, c); }
    }

    public abstract class Indirect4 : Indirect
    {
      protected Indirect4(MethodType type) : base(type) {}
      public override Obj Call(List args) { return Call4(args.Get(0), args.Get(1), args.Get(2), args.Get(3)); }
      public override Obj Call0() { throw Err.Make("Too few arguments: 0 < 4").val; }
      public override Obj Call1(Obj a) { throw Err.Make("Too few arguments: 1 < 4").val; }
      public override Obj Call2(Obj a, Obj b)  { throw Err.Make("Too few arguments: 2 < 4").val; }
      public override Obj Call3(Obj a, Obj b, Obj c) { throw Err.Make("Too few arguments: 3 < 4").val; }
      public override abstract Obj Call4(Obj a, Obj b, Obj c, Obj d);
      public override Obj Call5(Obj a, Obj b, Obj c, Obj d, Obj e) { return Call4(a, b, c, d); }
      public override Obj Call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { return Call4(a, b, c, d); }
      public override Obj Call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return Call4(a, b, c, d); }
      public override Obj Call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return Call4(a, b, c, d); }
    }

    public abstract class Indirect5 : Indirect
    {
      protected Indirect5(MethodType type) : base(type) {}
      public override Obj Call(List args) { return Call5(args.Get(0), args.Get(1), args.Get(2), args.Get(3), args.Get(4)); }
      public override Obj Call0() { throw Err.Make("Too few arguments: 0 < 5").val; }
      public override Obj Call1(Obj a) { throw Err.Make("Too few arguments: 1 < 5").val; }
      public override Obj Call2(Obj a, Obj b)  { throw Err.Make("Too few arguments: 2 < 5").val; }
      public override Obj Call3(Obj a, Obj b, Obj c) { throw Err.Make("Too few arguments: 3 < 5").val; }
      public override Obj Call4(Obj a, Obj b, Obj c, Obj d) { throw Err.Make("Too few arguments: 4 < 5").val; }
      public override abstract Obj Call5(Obj a, Obj b, Obj c, Obj d, Obj e);
      public override Obj Call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { return Call5(a, b, c, d, e); }
      public override Obj Call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return Call5(a, b, c, d, e); }
      public override Obj Call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return Call5(a, b, c, d, e); }
    }

    public abstract class Indirect6 : Indirect
    {
      protected Indirect6(MethodType type) : base(type) {}
      public override Obj Call(List args) { return Call6(args.Get(0), args.Get(1), args.Get(2), args.Get(3), args.Get(4), args.Get(5)); }
      public override Obj Call0() { throw Err.Make("Too few arguments: 0 < 6").val; }
      public override Obj Call1(Obj a) { throw Err.Make("Too few arguments: 1 < 6").val; }
      public override Obj Call2(Obj a, Obj b)  { throw Err.Make("Too few arguments: 2 < 6").val; }
      public override Obj Call3(Obj a, Obj b, Obj c) { throw Err.Make("Too few arguments: 3 < 6").val; }
      public override Obj Call4(Obj a, Obj b, Obj c, Obj d) { throw Err.Make("Too few arguments: 4 < 6").val; }
      public override Obj Call5(Obj a, Obj b, Obj c, Obj d, Obj e) { throw Err.Make("Too few arguments: 5 < 6").val; }
      public override abstract Obj Call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f);
      public override Obj Call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return Call6(a, b, c, d, e, f); }
      public override Obj Call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return Call6(a, b, c, d, e, f); }
    }

    public abstract class Indirect7 : Indirect
    {
      protected Indirect7(MethodType type) : base(type) {}
      public override Obj Call(List args) { return Call7(args.Get(0), args.Get(1), args.Get(2), args.Get(3), args.Get(4), args.Get(5), args.Get(6)); }
      public override Obj Call0() { throw Err.Make("Too few arguments: 0 < 7").val; }
      public override Obj Call1(Obj a) { throw Err.Make("Too few arguments: 1 < 7").val; }
      public override Obj Call2(Obj a, Obj b)  { throw Err.Make("Too few arguments: 2 < 7").val; }
      public override Obj Call3(Obj a, Obj b, Obj c) { throw Err.Make("Too few arguments: 3 < 7").val; }
      public override Obj Call4(Obj a, Obj b, Obj c, Obj d) { throw Err.Make("Too few arguments: 4 < 7").val; }
      public override Obj Call5(Obj a, Obj b, Obj c, Obj d, Obj e) { throw Err.Make("Too few arguments: 5 < 7").val; }
      public override Obj Call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { throw Err.Make("Too few arguments: 6 < 7").val; }
      public override abstract Obj Call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g);
      public override Obj Call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return Call7(a, b, c, d, e, f, g); }
    }

    public abstract class Indirect8 : Indirect
    {
      protected Indirect8(MethodType type) : base(type) {}
      public override Obj Call(List args) { return Call8(args.Get(0), args.Get(1), args.Get(2), args.Get(3), args.Get(4), args.Get(5), args.Get(6), args.Get(7)); }
      public override Obj Call0() { throw Err.Make("Too few arguments: 0 < 8").val; }
      public override Obj Call1(Obj a) { throw Err.Make("Too few arguments: 1 < 8").val; }
      public override Obj Call2(Obj a, Obj b)  { throw Err.Make("Too few arguments: 2 < 8").val; }
      public override Obj Call3(Obj a, Obj b, Obj c) { throw Err.Make("Too few arguments: 3 < 8").val; }
      public override Obj Call4(Obj a, Obj b, Obj c, Obj d) { throw Err.Make("Too few arguments: 4 < 8").val; }
      public override Obj Call5(Obj a, Obj b, Obj c, Obj d, Obj e) { throw Err.Make("Too few arguments: 5 < 8").val; }
      public override Obj Call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { throw Err.Make("Too few arguments: 6 < 8").val; }
      public override Obj Call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { throw Err.Make("Too few arguments: 7 < 8").val; }
      public override abstract Obj Call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h);
    }

    public abstract class IndirectX : Indirect
    {
      protected IndirectX(MethodType type) : base(type) {}
      public override abstract Obj Call(List args);
      public override Obj Call0() { throw TooFewArgs(0).val; }
      public override Obj Call1(Obj a) { throw TooFewArgs(1).val; }
      public override Obj Call2(Obj a, Obj b)  { throw TooFewArgs(2).val; }
      public override Obj Call3(Obj a, Obj b, Obj c) { throw TooFewArgs(3).val; }
      public override Obj Call4(Obj a, Obj b, Obj c, Obj d) { throw TooFewArgs(4).val; }
      public override Obj Call5(Obj a, Obj b, Obj c, Obj d, Obj e) { throw TooFewArgs(5).val; }
      public override Obj Call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { throw TooFewArgs(6).val; }
      public override Obj Call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { throw TooFewArgs(7).val; }
      public override Obj Call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { throw TooFewArgs(8).val; }
    }

  //////////////////////////////////////////////////////////////////////////
  // Compiler Support
  //////////////////////////////////////////////////////////////////////////

    //public MethodParam[] paramArr()
    //{
    //  return (MethodParam[])pars.toArray(new MethodParam[pars.Sz()]);
    //}

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    static readonly int GENERIC = 0x01;  // is this a generic method
    static readonly object[] noArgs = new object[0];

    internal readonly Type returnType;
    internal readonly Type inheritedReturnType; // for covariance
    internal readonly List pars;
    internal readonly int mask;
    internal readonly Method generic;
    //internal MethodInfo[] reflect;

  }
}