//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   24 Sep 06  Andy Frank  Creation
//

using System;
using System.Collections;
using Fanx.Emit;
using Fanx.Fcode;
using Fanx.Util;

namespace Fan.Sys
{
  /// <summary>
  /// Pod is a module containing Types.  A Pod is always backed by a FPod
  /// instance which defines all the definition tables.  Usually the FPod
  /// is in turn backed by a FStore for the pod's zip file.  However in the
  /// case of memory-only pods defined by the compiler, the fpod.store field
  /// will be null.
  ///
  /// Pods is loaded as soon as it is constructed:
  ///  1) All the types defined by the fpod are mapped into hollow Types.
  ///  2) It is emitted as a Java class called "fan.{podName}.$Pod".  The
  ///     emitted class is basically a manifestation of the literal tables,
  ///     after which we can clear the fpod data structures.
  /// </summary>
  public class Pod : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Management
  //////////////////////////////////////////////////////////////////////////

    public static Pod Find(Str name) { return Find(name.val, true, null); }
    public static Pod Find(Str name, Bool check) { return Find(name.val, check.val, null); }
    public static Pod Find(string name, bool check, FPod fpod)
    {
      // TODO
      return Sys.FindPod(name, check, fpod);
    }

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    internal Pod(Str name, FPod fpod)
    {
      this.name = name;
      Load(fpod);
    }

    // used by ShimPod and Sys.stubSysPod
    public Pod(String name)
    {
      this.name = Str.Make(name);
    }

  //////////////////////////////////////////////////////////////////////////
  // Methods
  //////////////////////////////////////////////////////////////////////////

    public override Type Type() { return Sys.PodType; }

    public Str Name()  { return name; }

    //public Version Version()  { return Version.parse(Str.Make("0.1")); }

  //////////////////////////////////////////////////////////////////////////
  // Types
  //////////////////////////////////////////////////////////////////////////

    public List Types() { return new List(Sys.TypeType, types); }

    public Type FindType(Str name) { return FindType(name.val, true); }
    public Type FindType(Str name, Bool check) { return FindType(name.val, check.val); }
    public Type FindType(string name, bool check)
    {
      Type type = (Type)typesByName[name];
      if (type != null) return type;
      if (check) throw UnknownTypeErr.Make(this.name + "::" + name).val;
      return null;
    }

  //////////////////////////////////////////////////////////////////////////
  // Files
  //////////////////////////////////////////////////////////////////////////

    //public Map files()
    //{
    //  if (files == null)
    //    files = fpod.store.podFiles();
    //  return files;
    //}

  //////////////////////////////////////////////////////////////////////////
  // Documentation
  //////////////////////////////////////////////////////////////////////////

    public Str Doc()
    {
      // TODO
      if (name.val == "sys") return Str.Make("Fan core runtime system");
      if (name.val == "sysTest") return Str.Make("Test suite for the sys pod");
      if (name.val == "compiler") return Str.Make("Fan compiler");
      if (name.val == "web") return Str.Make("Standard Weblet APIs for processing HTTP requests");
      if (name.val == "fandoc") return Str.Make("Weblet for publishing fan documentation");
      if (name.val == "fantext") return Str.Make("Parser and translators for working with fantext");
      if (name.val == "webServlet") return Str.Make("WebEnv for a Java servlet engine");
      return null;
    }

  //////////////////////////////////////////////////////////////////////////
  // Conversion
  //////////////////////////////////////////////////////////////////////////

    public override Str ToStr() { return name; }

  //////////////////////////////////////////////////////////////////////////
  // Load
  //////////////////////////////////////////////////////////////////////////

    internal void Load(FPod fpod)
    {
      this.fpod = fpod;
      this.typesByName = new Hashtable();

      // create a hollow Type for each FType (this requires two steps,
      // because we don't necessary have all the Types created for
      // superclasses until this loop completes)
      types = new Type[fpod.types.Length];
      for (int i=0; i<fpod.types.Length; i++)
      {
        // create type instance
        Type type = new Type(this, fpod.types[i]);

        // add to my data structures
        types[i] = type;
        if (typesByName[type.name.val] != null)
          throw Err.Make("Invalid pod: " + name + " type already defined: " + type.name).val;
        typesByName[type.name.val] = type;
      }

      // get TypeType to use for mixin List (we need to handle case
      // when loading sys itself - and lookup within my own pod)
      /*
      Type typeType = Sys.TypeType;
      if (typeType == null)
        typeType = (Type)typesByName.get("Type");
      */

      // now that everthing is mapped, we can fill in the super
      // class fields (unless something is wacked, this will only
      // use Types in my pod or in pods already loaded)
      for (int i=0; i<fpod.types.Length; i++)
      {
        FType ftype   = fpod.types[i];
        Type type     = types[i];
        type.baseType = FindType(ftype.baset);

        //List mixins = new List(typeType, ftype.mixins.length);
        //for (int j=0; j<ftype.mixins.length; j++)
        //  mixins.add(findType(ftype.mixins[j]));
        //type.mixins = mixins.ro();
      }

      // emit the pod class itself (which declares all constants)
      // TODO - have to do this in FTypeEmit?
      //if (name.val != "sys")
      //  this.type = FPodEmit.EmitAndLoad(fpod);

      // now it is safe to clear the literal tables
      //fpod.freeLiteralTables();
    }

    internal Type FindType(int qname)
    {
      if (qname == 0xFFFF || qname == -1) return null;

      // lookup type with typeRef index
      FTypeRef reference = fpod.TypeRef(qname);

      // if generic instance, then this type must be used in a method
      // signature, not type meta-data (b/c I can't subclass generic types),
      // so it's safe that my pod has been loaded and is now registered (in
      // case the generic type is parameterized via types in my pod)
      if (reference.IsGenericInstance())
        return TypeParser.Load(reference.sig, true, this);

      // otherwise I need to handle if I am loading my own pod, because
      // I might not yet be added to the system namespace if I'm just
      // loading my own hollow types
      String podName  = fpod.Name(reference.podName);
      String typeName = fpod.Name(reference.typeName);
      Pod pod = podName == name.val ? this : Sys.FindPod(podName, true, null);
      Type type = pod.FindType(typeName, false);
      if (type != null) return type;

      // handle variance types (for sys pod only)
      if (this.name.val == "sys")
      {
        type = Sys.GenericParameterType(typeName);
        if (type != null) return type;
      }

      // lost cause
      throw UnknownTypeErr.Make(podName + "::" + typeName).val;
    }

    /// <summary>
    /// Close the pod which should release any locks on the pod
    /// file.  This method exists only for testing, and should
    /// not otherwise be used.
    /// </summary>
    public void Close()
    {
      fpod.store.Close();
    }

  //////////////////////////////////////////////////////////////////////////
  // Compiler Support
  //////////////////////////////////////////////////////////////////////////

    // stub is used when sys.pod is not found, so that we can stub
    // enough to do self tests and compile sys itself
    internal Type Stub(Type type)
    {
      if (typesByName == null) typesByName = new Hashtable();
      typesByName[type.Name().val] = type;
      return type;
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    internal readonly Str name;
    internal FPod fpod;
    internal Type[] types;
    internal Hashtable typesByName;
    //internal System.Type type; // TODO - exists?
    //internal Map files;

  }
}