//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   29 Jan 07  Andy Frank  Creation
//

namespace Fan.Sys
{
  /// <summary>
  /// Range.
  /// </summary>
  public sealed class Range : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////

    public static Range MakeInclusive(Int start, Int end)
    {
      return new Range(start, end, false);
    }

    public static Range MakeExclusive(Int start, Int end)
    {
      return new Range(start, end, true);
    }

    public static Range Make(Int start, Int end, Bool exclusive)
    {
      return new Range(start, end, exclusive.val);
    }

    private Range(Int start, Int end, bool exclusive)
    {
      if (start == null || end == null) throw NullErr.Make().val;
      this.start = start;
      this.end = end;
      this.exclusive = exclusive;
    }

  //////////////////////////////////////////////////////////////////////////
  // Range
  //////////////////////////////////////////////////////////////////////////

    public Int Start()
    {
      return start;
    }

    public Int End()
    {
      return end;
    }

    public Bool Inclusive()
    {
      return exclusive ? Bool.False : Bool.True;
    }

    public Bool Exclusive()
    {
      return exclusive ? Bool.True : Bool.False;
    }

    public override Bool Equals(Obj obj)
    {
      if (obj is Range)
      {
        Range that = (Range)obj;
        return Bool.Make(this.start.val == that.start.val &&
                         this.end.val == that.end.val &&
                         this.exclusive == that.exclusive);
      }
      return Bool.False;
    }

    public override int GetHashCode()
    {
      return start.GetHashCode() ^ end.GetHashCode();
    }

    public override Int Hash()
    {
      return Int.Make(start.val ^ end.val);
    }

    public override Str ToStr()
    {
      if (exclusive)
        return Str.Make(start.ToStr().val + "..." + end.ToStr().val);
      else
        return Str.Make(start.ToStr().val + ".." + end.ToStr().val);
    }

    public override Type Type() { return Sys.RangeType; }

  //////////////////////////////////////////////////////////////////////////
  // Slice Utils
  //////////////////////////////////////////////////////////////////////////

    internal int Start(int size)
    {
      int x = (int)start.val;
      if (x < 0) x = size + x;
      if (x > size) throw IndexErr.Make(this).val;
      return x;
    }

    internal int End(int size)
    {
      int x = (int)end.val;
      if (x < 0) x = size + x;
      if (exclusive) x--;
      if (x >= size) throw IndexErr.Make(this).val;
      return x;
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    private Int start, end;
    private bool exclusive;
  }
}