//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   13 Sep 06  Andy Frank  Creation
//

using System;
using System.Collections;
using System.Text;
using System.Text.RegularExpressions;

namespace Fan.Sys
{
  ///
  /// Str is a string of Unicode characters.
  ///
  public sealed class Str : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    public static Str Make(string val)
    {
      if (val == null) return null;
      if (val == "")   return Empty;
      return new Str(val);
    }

    private Str(string val)
    {
      this.val = val;
    }

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override bool Equals(Object obj)
    {
      return val == (obj as Str).val;
    }

    public override Bool Equals(Obj obj)
    {
      if (obj is Str)
        return val.Equals(((Str)obj).val) ? Bool.True : Bool.False;
      else
        return Bool.False;
    }

    public override Int Compare(Obj obj)
    {
      int cmp = String.CompareOrdinal(val, ((Str)obj).val);
      if (cmp < 0) return Int.LT;
      return cmp == 0 ? Int.EQ : Int.GT;
    }

    public override int GetHashCode()
    {
      return val.GetHashCode();
    }

    public override Int Hash()
    {
      return Int.Make(val.GetHashCode());
    }

    public override Str ToStr()
    {
      return this;
    }

    public override Type Type()
    {
      return Sys.StrType;
    }

  //////////////////////////////////////////////////////////////////////////
  // Operators
  //////////////////////////////////////////////////////////////////////////

    public Int Get(Int index)
    {
      int i = (int)index.val;
      if (i < 0) i = val.Length+i;
      return Int.Pos(val[i]);
    }

    public Str Slice(Range r)
    {
      int size = val.Length;

      int s = r.Start(size);
      int e = r.End(size);
      if (e+1 < s) throw IndexErr.Make(r).val;

      return Make(val.Substring(s, (e-s)+1));
    }

    public Str Plus(Obj obj)
    {
      if (obj == null) return Make(String.Concat(val, "null"));
      Str x = obj.ToStr();
      if (x.val == "") return this;
      return Make(String.Concat(val, x.val));
    }

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public Str Intern()
    {
      Str s = (Str)interns[val];
      if (s == null) interns[val] = s = this;
      return s;
    }

    public Bool IsEmpty()
    {
      return val.Length == 0 ? Bool.True : Bool.False;
    }

    public Int Size()
    {
      return Int.Pos(val.Length);
    }

    public Bool StartsWith(Str s)
    {
      return Bool.Make(val.StartsWith(s.val));
    }

    public Bool EndsWith(Str s)
    {
      return Bool.Make(val.EndsWith(s.val));
    }

    public Bool Contains(Str s)
    {
      return Index(s, Int.Zero) != null ? Bool.True : Bool.False;
    }

    public Int Index(Str s) { return Index(s, Int.Zero); }
    public Int Index(Str s, Int off)
    {
      int i = (int)off.val;
      if (i < 0) i = val.Length+i;

      int r;
      string sval = s.val;
      if (sval.Length == 1)
        r = val.IndexOf(sval[0], i);
      else
        r = val.IndexOf(sval, i);

      if (r < 0) return null;
      return Int.Make(r);
    }

    public Int Indexr(Str s) { return Indexr(s, Int.NegOne); }
    public Int Indexr(Str s, Int off)
    {
      int i = (int)off.val;
      if (i < 0) i = val.Length+i;

      int r;
      String sval = s.val;
      if (sval.Length == 1)
        r = val.LastIndexOf(sval[0], i);
      else
        r = val.LastIndexOf(sval, i);

      if (r < 0) return null;
      return Int.Make(r);
    }

  //////////////////////////////////////////////////////////////////////////
  // Iterators
  //////////////////////////////////////////////////////////////////////////

    public void Each(Method m)
    {
      for (int i=0; i<val.Length; i++)
        m.Call2(Int.Pos(val[i]), Int.Pos(i));
    }

    public void Eachr(Method m)
    {
      for (int i=val.Length-1; i>=0; i--)
        m.Call2(Int.Pos(val[i]), Int.Pos(i));
    }

  //////////////////////////////////////////////////////////////////////////
  // Utils
  //////////////////////////////////////////////////////////////////////////

    public static Str Spaces(Int n)
    {
      // do an array lookup for reasonable length
      // strings since that is the common case
      int count = (int)n.val;
      try { return spaces[count]; } catch (IndexOutOfRangeException) {}

      // otherwise we build a new one
      StringBuilder s = new StringBuilder(spaces[spaces.Length-1].val);
      for (int i=spaces.Length-1; i<count; i++)
        s.Append(' ');
      return Str.Make(s.ToString());
    }

    public Str Lower() { return Str.Make(Lower(val)); }
    public static string Lower(String val)
    {
      StringBuilder s = new StringBuilder(val.Length);
      for (int i=0; i<val.Length; i++)
      {
        int ch = val[i];
        if ('A' <= ch && ch <= 'Z') ch |= 0x20;
        s.Append((char)ch);
      }
      return s.ToString();
    }

    public Str Upper()
    {
      string val = this.val;
      StringBuilder s = new StringBuilder(val.Length);
      for (int i=0; i<val.Length; i++)
      {
        int ch = val[i];
        if ('a' <= ch && ch <= 'z') ch &= ~0x20;
        s.Append((char)ch);
      }
      return Make(s.ToString());
    }

    public Str Justl(Int width)
    {
      string val = this.val;
      int w = (int)width.val;
      if (val.Length >= w) return this;
      StringBuilder s = new StringBuilder(w);
      s.Append(val);
      for (int i=val.Length; i<w; i++)
        s.Append(' ');
      return Make(s.ToString());
    }

    public Str Justr(Int width)
    {
      string val = this.val;
      int w = (int)width.val;
      if (val.Length >= w) return this;
      StringBuilder s = new StringBuilder(w);
      for (int i=val.Length; i<w; i++)
        s.Append(' ');
      s.Append(val);
      return Make(s.ToString());
    }

    public Str Reverse()
    {
      string val = this.val;
      if (val.Length < 2) return this;
      StringBuilder s = new StringBuilder(val.Length);
      for (int i=val.Length-1; i>=0; i--)
        s.Append(val[i]);
      return Make(s.ToString());
    }

    public Str Trim()
    {
      string val = this.val;
      int len = val.Length;
      if (len == 0) return this;
      if (val[0] > ' ' && val[len-1] > ' ') return this;
      return Make(val.Trim());
    }

    public List Split(Str seps)
    {
      // TODO?
      string[] split = Regex.Split(val, seps.val);
      Str[] str = new Str[split.Length];
      for (int i=0; i<str.Length; i++) str[i] = Str.Make(split[i]);
      return new List(Sys.StrType, str);
    }

    /*
    public Str Replace(Str from, Str to)
    {
      return Make(StrUtil.Replace(val, from.val, to.val));
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Conversion
  //////////////////////////////////////////////////////////////////////////

    public Bool ToBool() { return Bool.Parse(this, Bool.True); }
    public Bool ToBool(Bool check) { return Bool.Parse(this, check); }

    public Int ToInt() { return Int.Parse(this, Int.Ten, Bool.True); }
    public Int ToInt(Int radix) { return Int.Parse(this, radix, Bool.True); }
    public Int ToInt(Int radix, Bool check) { return Int.Parse(this, radix, check); }

    public Float ToFloat() { return Float.Parse(this, Bool.True); }
    public Float ToFloat(Bool check) { return Float.Parse(this, check); }

    public Uri ToUri() { return Uri.Parse(this); }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public static readonly Str Empty = new Str("");
    public static readonly Str Null = new Str("null");

    private static readonly Hashtable interns = new Hashtable();

    internal static readonly Str[] ascii = new Str[128];
    internal static Str[] spaces = new Str[20];

    static Str()
    {
      // ascii
      for (int i=0; i<ascii.Length; i++)
        ascii[i] = Make("" + (char)i).Intern();

      // spaces
      StringBuilder s = new StringBuilder();
      for (int i=0; i<spaces.Length; i++)
      {
        spaces[i] = Str.Make(s.ToString());
        s.Append(' ');
      }
    }

    public readonly string val;
  }
}