//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   19 Oct 06  Andy Frank  Creation
//

using System;
using System.Text;
using URI = System.Uri;

namespace Fan.Sys
{
  /// <summary>
  /// Uri is used to immutably represent a Universal Resource Identifier.
  /// </summary>
  public sealed class Uri : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    public static Uri Parse(string s) { return Parse(Str.Make(s)); }
    public static Uri Parse(Str s)
    {
      try
      {
        return new Uri(s);
      }
      catch (Exception)
      {
        throw ArgErr.Make("Invalid Uri: `" + s + "`").val;
      }
    }

    public Uri(Str s) : this(s, new URI(s.val, UriKind.RelativeOrAbsolute)) {}

    public Uri(URI uri) : this(Str.Make(uri.ToString()), uri) {}

    private Uri(Str s, URI u)
    {
      str = s;
      uri = u;

      // these properties throw if URI is relative
      try
      {
        scheme   = Lower(u.Scheme);
        host     = Str.Make(u.Host);
        pathStr  = Str.Make(u.AbsolutePath);
        queryStr = Str.Make(u.Query);
      }
      catch (System.InvalidOperationException)
      {
        scheme   = null;
        host     = null;
        pathStr  = s;
        queryStr = null;
      }

      path  = ParsePath(pathStr);
      query = ParseQuery(queryStr);
    }

    internal static Str Lower(String s)
    {
      if (s == null) return null;
      return Str.Make(Str.Lower(s));
    }

    internal static List ParsePath(Str pathStr)
    {
      if (pathStr == null) return null;
      List list = new List(Sys.StrType);

      try
      {
        string p = pathStr.val;
        int start = 0, len = p.Length;
        for (int i=0; i<len; i++)
        {
          int ch = p[i];
          if (ch != '/') continue;
          if (start < i)
            list.Add(Str.Make(p.Substring(start, (i-start))));
          start = i+1;
        }
        if (start < len)
          list.Add(Str.Make(p.Substring(start, (len-start))));
      }
      catch (Exception e)
      {
        // don't let internal error bring down whole uri
        System.Console.WriteLine(e);
      }

      return list.Trim().RO();
    }

    internal static Map ParseQuery(Str queryStr)
    {
      if (queryStr == null) return null;
      Map map = new Map(Sys.StrType, Sys.StrType);

      try
      {
        string q = queryStr.val;
        int start = 0, eq = 0, len = q.Length;
        for (int i=0; i<len; i++)
        {
          int ch = q[i];
          if (ch == '=') eq = i;
          if (ch != '&' && ch != ';') continue;

          if (start < i)
          {
            if (start == eq)
              map.Set(Str.Make(q.Substring(start, (i-start))), null);
            else
              map.Set(Str.Make(q.Substring(start, (eq-start))), Str.Make(q.Substring(eq+1, (i-(eq+1)))));
          }

          start = eq = i+1;
        }

        if (start < len)
        {
          if (start == eq)
            map.Set(Str.Make(q.Substring(start, (len-start))), null);
          else
            map.Set(Str.Make(q.Substring(start, (eq-start))), Str.Make(q.Substring(eq+1, (len-(eq+1)))));
        }
      }
      catch (Exception e)
      {
        // don't let internal error bring down whole uri
        System.Console.WriteLine(e);
      }

      return map.RO();
    }

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override bool Equals(Object obj)
    {
      return str.val == (obj as Uri).str.val;
    }

    public override Bool Equals(Obj obj)
    {
      if (obj is Uri)
      {
        return str.Equals((obj as Uri).str);
      }
      return Bool.False;
    }

    public override int GetHashCode()
    {
      return str.GetHashCode();
    }

    public override Int Hash()
    {
      return str.Hash();
    }

    public override Str ToStr()
    {
      return str;
    }

    public override Type Type()
    {
      return Sys.UriType;
    }

  //////////////////////////////////////////////////////////////////////////
  // Components
  //////////////////////////////////////////////////////////////////////////

    public Bool IsAbs()
    {
      return scheme != null ? Bool.True : Bool.False;
    }

    public Bool IsRel()
    {
      return scheme == null ? Bool.True : Bool.False;
    }

    public Bool IsDir()
    {
      if (pathStr != null)
      {
        string p = pathStr.val;
        int len = p.Length;
        if (len > 0 && p[len-1] == '/')
          return Bool.True;
      }
      return Bool.False;
    }

    public Str Scheme()
    {
      return scheme;
    }

    public Str Body()
    {
      if (scheme == null) return str;
      return Str.Make(str.val.Substring(scheme.val.Length+1));
    }

    /*
    public Str auth()
    {
      if (host == null) return null;
      int port = uri.getPort();
      String userInfo = uri.getUserInfo();
      if (port < 0)
      {
        if (userInfo == null) return host;
        else return Str.make(userInfo + '@' + host.val);
      }
      else
      {
        if (userInfo == null) return Str.make(host.val + ':' + port);
        else return Str.make(userInfo + '@' + host.val + ':' + port);
      }
    }
    */

    public Str Host()
    {
      return host;
    }

    public Str UserInfo()
    {
      return null; //Str.make(uri.getUserInfo());
    }

    public Int Port()
    {
      int port = 80; //uri.getPort();
      return port < 0 ? null : Int.Make(port);
    }

    public List Path()
    {
      return path;
    }

    public Str PathStr()
    {
      return pathStr;
    }

    public Str Name()
    {
      if (path == null || path.Sz() == 0) return null;
      return (Str)path.Last();
    }

    public Str Basename()
    {
      Str name = Name();
      if (name == null) return null;
      string n = name.val;
      int dot = n.LastIndexOf('.');
      if (dot < 2)
      {
        if (dot < 0) return name;
        if (n == ".") return name;
        if (n == "..") return name;
      }
      return Str.Make(n.Substring(0, dot));
    }

    public Str Ext()
    {
      Str name = Name();
      if (name == null) return null;
      string n = name.val;
      int dot = n.LastIndexOf('.');
      if (dot < 2)
      {
        if (dot < 0) return null;
        if (n == ".") return null;
        if (n == "..") return null;
      }
      return Str.Make(n.Substring(dot+1));
    }

    public Map Query()
    {
      return query;
    }

    public Str QueryStr()
    {
      return queryStr;
    }

    public Str Frag()
    {
      return Str.Make(uri.Fragment);
    }

  //////////////////////////////////////////////////////////////////////////
  // Manipulation
  //////////////////////////////////////////////////////////////////////////

    public Uri Normalize()
    {
      //URI norm = uri.Normalize();
      //if (norm == uri) return this;
      //return new Uri(norm);
      return this;
    }

    public Uri Plus(Uri join)
    {
      //return new Uri(uri.resolve(join.uri));
      return new Uri(new URI(uri, join.uri));
    }

    public Uri Minus(Uri bse)
    {
      //return new Uri(base.uri.relativize(uri));
      return new Uri(bse.uri.MakeRelativeUri(uri));
    }

    public Uri Tail() { return Tail(Int.One); }
    public Uri Tail(Int levels)
    {
      int lvl = (int)levels.val;
      if (path == null || path.Sz() < lvl)
        throw ArgErr.Make("Cannot call tail(" + lvl + ") on `" + str + "`").val;
      StringBuilder s = new StringBuilder();
      for (int i=lvl; i<path.Sz(); i++)
      {
        if (i > lvl) s.Append('/');
        s.Append(path.Get(i));
      }
      return Parse(Str.Make(s.ToString()));
    }

  //////////////////////////////////////////////////////////////////////////
  // Resolution
  //////////////////////////////////////////////////////////////////////////

    public Obj Get() { return Get(null); }
    public Obj Get(Obj baseUri)
    {
      if (scheme != null)
      {
        throw UnsupportedErr.Make("Absolute Uris not supported yet: " + str).val;
      }

      if (baseUri == null)
        throw ArgErr.Make("Must specify base for `" + str + "`").val;

      //return baseUri.trapUri(this);
      return null;
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    internal URI uri;
    internal Str str;
    internal Str scheme;
    internal Str host;
    internal List path;
    internal Str pathStr;
    internal Map query;
    internal Str queryStr;

  }
}