//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   14 Sep 06  Andy Frank  Creation
//

using System;
using System.Collections;
using System.IO;
using PERWAPI;
using Fanx.Util;

namespace Fanx.Emit
{
  /// <summary>
  /// Emitter.
  /// </summary>
  public class Emitter
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Convenience for init().
    /// </summary>
    public Emitter(string assemblyName)
    {
      Init(assemblyName);
    }

    /// <summary>
    /// Uninitialized Emitter - must call init().
    /// </summary>
    public Emitter() {}

    /// <summary>
    /// Initialize an Emitter to generate types for an assembly
    /// with the given name.
    /// </summary>
    public void Init(string assemblyName)
    {
      this.assemblyName = assemblyName;
      buf = new MemoryStream(1000);
      peFile = new PEFile(assemblyName + ".dll", assemblyName);
      peFile.SetOutputStream(buf);
    }

    /// <summary>
    /// Commit the current definition to the PE file.
    /// </summary>
    public byte[] Commit()
    {
      peFile.WritePEFile(true); // TODO - debug flag
      //return buf.ToArray();
      byte[] b = buf.ToArray();

      if (true) // dump to file
      {
        string path = @"c:\dev\fan\tmp\" + assemblyName + ".dll";
        BinaryWriter w = new BinaryWriter(new FileInfo(path).OpenWrite());
        w.Write(b, 0, b.Length);
        w.Flush();
        w.Close();
      }

      return b;
    }

  //////////////////////////////////////////////////////////////////////////
  // Methods
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Define a new class to emit for this assembly. This class
    /// becomes the 'current' class, where subsequent EmitMethod
    /// and EmitField calls will popuate this class.
    /// </summary>
    public void EmitClass(string baseQName, string qname, TypeAttr attr)
    {
      string[] s = NameUtil.SplitQName(qname);
      className = qname;

      // first check if this type was already stubbed out
      classDef = (ClassDef)types[qname];
      if (classDef == null)
      {
        // if not, we need to create it
        classDef = peFile.AddClass(attr, s[0], s[1]);
      }

      if (baseQName == null)
        classDef.SuperType = null;
      else
        classDef.SuperType = FindType(baseQName) as PERWAPI.Class;

      types[qname] = classDef;
    }

    /// <summary>
    /// Define a new field to emit for the current class.
    /// </summary>
    public void EmitField(string name, string type, FieldAttr attr)
    {
      string key = className + "/F" + name;
      if (fields[key] == null)
      {
        // only add if not already stubbed out
        Field field = classDef.AddField(attr, name, FindType(type));
        fields[key] = field;
      }
    }

    /// <summary>
    /// Define a new method to emit for the current class.
    /// </summary>
    public CILInstructions EmitMethod(
      string name, string retType, string[] paramNames, string[] paramTypes,
      MethAttr attr, string[] localNames, string[] localTypes)
    {
      Param[] pars = new Param[paramNames.Length];
      for (int i=0; i<pars.Length; i++)
        pars[i] = new Param(ParamAttr.Default, paramNames[i], FindType(paramTypes[i]));

      // first check if this method was already stubbed out
      methodDef = (MethodDef)methods[GetMethodKey(className, name, paramTypes)];
      if (methodDef == null)
      {
        // if not, we need to create it
        methodDef = classDef.AddMethod(attr, ImplAttr.IL, name, FindType(retType), pars);
      }
      else
      {
        // if stubbed out, make sure we define the method correctly
        methodDef.SetMethAttributes(attr);
      }

      if ((attr & MethAttr.Static) != MethAttr.Static)
        methodDef.AddCallConv(CallConv.Instance);

      if (localNames.Length > 0)
      {
        Local[] locals = new Local[localNames.Length];
        for (int i=0; i<locals.Length; i++)
          locals[i] = new Local(localNames[i], FindType(localTypes[i]));
        methodDef.AddLocals(locals, true);
      }

      // TODO - what the fuck should this be?
      methodDef.SetMaxStack(10 + pars.Length + localNames.Length);

      // add to lookup table
      AddToMethodMap(className, name, paramTypes, methodDef);

      // don't create code buffer if abstract
      if ((attr & MethAttr.Abstract) == MethAttr.Abstract)
        return null;

      return methodDef.CreateCodeBuffer();
    }

    /// <summary>
    /// Add a MethodDef into the lookup table.
    /// </summary>
    public void AddToMethodMap(string className, string name, string[] paramTypes, MethodDef def)
    {
      string key = GetMethodKey(className, name, paramTypes);
      methods[key] = def;
    }

  //////////////////////////////////////////////////////////////////////////
  // Util
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Find the Type instance for this fully qualified type name.
    /// </summary>
    internal PERWAPI.Type FindType(string qname)
    {
      // Always convert voids to native
      if (qname == "Fan.Sys.Void") qname = "System.Void";

      PERWAPI.Type type = (PERWAPI.Type)types[qname];
      if (type == null)
      {
        string aname = NameUtil.GetPodName(qname);
        if (aname == null) aname = "mscorlib";
        if (qname.StartsWith("Fanx.")) aname = "sys";  // hack for support classes

        // first check if this is a type in this pod that
        // hasn't been defined yet
        if (aname == assemblyName)
        {
          // stub out type - fill get filled in later (we hope)
          // TODO - do we need to handle nested types here?
          // TODO - hack the attr here - how do we tell for real?
          string[] sn = NameUtil.SplitQName(qname);
          ClassDef stub = peFile.AddClass(PERWAPI.TypeAttr.Public, sn[0], sn[1]);
          types[qname] = stub;
          return stub;
        }

        AssemblyRef aref = (AssemblyRef)assemblies[aname];
        if (aref == null)
        {
          aref = peFile.MakeExternAssembly(aname);
          assemblies[aname] = aref;
        }

        string[] s = NameUtil.SplitQName(qname);
        if (qname.IndexOf("/") != -1)
        {
          // Nested class
          PERWAPI.ClassRef cref = (PERWAPI.ClassRef)FindType(s[0]);
          type = cref.AddNestedClass(s[1]);
        }
        else
        {
          // Normal class
          // TODO - this broke when I moved to PERWAPI - not sure why
          // so now I have to check that it doesn't exist b/f adding
          type = aref.GetClass(s[0], s[1]);
          if (type == null) type = aref.AddClass(s[0], s[1]);
        }
        types[qname] = type;
      }
      return type;
    }

    /// <summary>
    /// Find the Field for this field on the given type.
    /// </summary>
    internal Field FindField(string qname, string fieldName, string fieldType)
    {
      string key = qname + "/F" + fieldName;
      Field field = (Field)fields[key];
      if (field == null)
      {
        // should be same for both cases
        PERWAPI.Type ftype = FindType(fieldType);

        // check for stubs
        object obj = FindType(qname);
        if (obj is ClassDef)
        {
          // class is an internal stub, so we need to stub this field
          // out too, which is actually complete and not a stub
          ClassDef cdef = obj as ClassDef;
          // TODO - fix attr
          field = cdef.AddField(FieldAttr.Public, fieldName, ftype);
        }
        else if (obj is ClassRef)
        {
          // class is external, just add ref
          ClassRef cref = obj as ClassRef;
          field = cref.AddField(fieldName, ftype);
        }
        else throw new System.Exception("Don't know how to handle: " + obj.GetType());

        // remember to add field to lookup table
        fields[key] = field;
      }
      return field;
    }

    /// <summary>
    /// Find the Method for this method on the given type.
    /// </summary>
    internal Method FindMethod(string qname, string methodName, string[] paramTypes, string returnType)
    {
      string key = GetMethodKey(qname, methodName, paramTypes);
      Method method = (Method)methods[key];

      if (method == null)
      {
        // this stuff should be the same for both cases
        PERWAPI.Type rtype = FindType(returnType);
        PERWAPI.Type[] pars = new PERWAPI.Type[paramTypes.Length];
        for (int i=0; i<paramTypes.Length; i++)
          pars[i] = FindType(paramTypes[i]);

        // check for stubs
        object obj = FindType(qname);
        if (obj is ClassDef)
        {
          // class is an internal stub, so we need to stub this
          // method out too, which should get flushed out later
          ClassDef cdef = obj as ClassDef;

          // TODO - need to fix attrs
          Param[] mpars = new Param[pars.Length];
          for (int i=0; i<mpars.Length; i++)
            mpars[i] = new Param(ParamAttr.Default, "v"+i, pars[i]);

          // TODO - fix param names
          // Use Public here for stub - real MethAttr will be set
          // when method is flushed out
          method = cdef.AddMethod(MethAttr.Public, ImplAttr.IL, methodName, rtype, mpars);
        }
        else if (obj is ClassRef)
        {
          // class is external, just add ref
          ClassRef cref = obj as ClassRef;
          method = cref.AddMethod(methodName, rtype, pars);
        }
        else throw new System.Exception("Don't know how to handle: " + obj.GetType());

        // make sure we add method to hashtable
        methods[key] = method;
      }

      return method;
    }

    /// <summary>
    /// Return the key for this method in the method lookup table.
    /// </summary>
    internal string GetMethodKey(string qname, string methodName, string[] paramTypes)
    {
      string key = qname + "/M" + methodName;
       for (int i=0; i<paramTypes.Length; i++)
        key += "/" + paramTypes[i];
      return key;
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    private string assemblyName;  // name of this assbemly
    private MemoryStream buf;     // memory buffer for pe file
    private PEFile peFile;        // model for pe file

    private string className;     // the current class name being emitted
    internal ClassDef classDef;   // the current class being emitted
    internal MethodDef methodDef; // the current or last method emitted

    private Hashtable assemblies = new Hashtable();  // Assembly lookup table
    private Hashtable types = new Hashtable();       // Type lookup table
    private Hashtable fields = new Hashtable();      // Field lookup table
    private Hashtable methods = new Hashtable();     // Method lookup table
  }
}