//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   5 Oct 06  Andy Frank  Creation
//

using System.Collections;
using System.Reflection;
using Fan.Sys;
using Fanx.Fcode;
using Fanx.Util;

namespace Fanx.Emit
{
  /// <summary>
  /// FMethodEmit is used to emit IL methods from fcode methods. It
  /// encapsulates lot of nitty details like when to include an implicit
  /// self paramater, etc.
  /// </summary>
  public class FMethodEmit
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Constructor.
    /// </summary>
    public FMethodEmit(FTypeEmit emit, FMethod method)
    {
      this.emitter    = emit.emitter;
      this.emit       = emit;
      this.method     = method;
      this.code       = method.code;
      this.name       = NameUtil.Upper(emit.Name(method.name));
      this.paramLen   = method.paramCount;
      this.isStatic   = (method.flags & FConst.Static) != 0;
      this.isInternal = (method.flags & FConst.Internal) != 0;
      this.isPrivate  = (method.flags & FConst.Private) != 0;
      this.isAbstract = (method.flags & FConst.Abstract) != 0;
      this.isVirtual  = (method.flags & FConst.Virtual) != 0;
      this.isOverride = (method.flags & FConst.Override) != 0;
      this.isCtor     = (method.flags & FConst.Ctor) != 0;
      //this.isNative = (method.flags & FConst.Native) != 0;
      this.ret        = emit.Nname(method.ret);
      this.isVoid     = ret == "Fan.Sys.Void";
      this.selfName   = emit.selfName;
    }

    /// <summary>
    /// Constructor.
    /// </summary>
    public FMethodEmit(FTypeEmit emit)
    {
      this.emitter = emit.emitter;
      this.emit = emit;
    }

  //////////////////////////////////////////////////////////////////////////
  // Emit
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Emit a standard instance/static class method.
    /// </summary>
    public void EmitStandard()
    {
      // emit method
      DoEmit();

      // emit param default wrappers
      EmitWrappers();
    }

    /// <summary>
    /// Emit a constructor - constructors get created as a static
    /// factory methods, so that that CallNew can just push args
    /// and invoke them
    ///
    ///   fan:
    ///     class Foo { new make(Int a) { ... } }
    ///   .net:
    ///     static Foo Make(Int a) { return Make$(new Foo(), a) }
    ///     static Foo Make$(Foo self, Int a) { ... return self }
    ///
    /// We call the first method "make" the "factory" and the
    /// second method "make$" the "body".  CallNew opcodes are
    /// routed to the ctor factory, and CallCtor opcodes are routed
    /// to the ctor body.
    /// </summary>
    public void EmitCtor()
    {
      string ctorName = this.name;

      // both factory and body are static from CLR's perspective
      this.isStatic = true;

      // first emit the body with implicit self
      this.name   = ctorName + "$";
      this.self   = true;
      this.isVoid = true;
      DoEmit();
      PERWAPI.MethodDef make = emitter.methodDef;

      // emit body default parameter wrappers
      EmitWrappers();

      // then emit the factory
      this.name   = ctorName;
      this.self   = false;
      this.isVoid = false;
      this.ret    = selfName;
      this.code   = null;

      PERWAPI.CILInstructions code = DoEmit();
      PERWAPI.Method ctor = emitter.FindMethod(selfName, ".ctor", new string[0], "System.Void");
      code.MethInst(PERWAPI.MethodOp.newobj, ctor);
      code.Inst(PERWAPI.Op.dup);
      PushArgs(code, method.paramCount);
      code.MethInst(PERWAPI.MethodOp.call, make);
      code.Inst(PERWAPI.Op.ret);

      // emit factory default parameter wrappers
      EmitWrappers();
    }

    /// <summary>
    /// Emit a native method
    /// <summary>
    public void EmitNative()
    {
      /*
      // emit an empty method
      this.code = null;
      MethodEmit main = doEmit();

      // emit code which calls the peer
      CodeEmit code = main.emitCode();
      if (isStatic)
      {
        int peerMethod = emit.method(selfName + "Peer." + name + sig);
        pushArgs(code, paramLen);
        code.op2(INVOKESTATIC, peerMethod);
      }
      else
      {
        // generate peer's signature with self
        this.self = true;
        String sig = signature();
        this.self = false;

        int peerMethod = emit.method(selfName + "Peer." + name + sig);
        code.op(ALOAD_0);
        code.op2(GETFIELD, emit.peerField.ref());
        pushArgs(code, paramLen+1);
        code.op2(INVOKEVIRTUAL, peerMethod);
      }
      code.op(isVoid ? RETURN : ARETURN);

      // emit default parameter wrappers
      EmitWrappers(main);
      */
    }

    /// <summary>
    /// Emit the method as a mixin interface
    /// </summary>
    public void EmitMixinInterface()
    {
      // we only emit instance methods in the interface
      if (isStatic || isCtor) return;

      // set abstract flag and clear code
      code = null;

      // force public
      isInternal = false;
      isPrivate  = false;

      // force abstract virtual
      isAbstract = true;
      isVirtual = true;

      // emit main
      DoEmit();

      // emit a signature for each overload based on param defaults
      for (int i=0; i<method.paramCount; i++)
      {
        if (method.vars[i].def != null)
        {
          paramLen = i;
          DoEmit();
        }
      }
    }

    /// <summary>
    /// Emit the method as a mixin body class which ends with $.
    /// <summary>
    public void EmitMixinBody()
    {
      // skip abstract methods without code
      if (method.code == null) return;

      // instance methods have an implicit self
      if (!isStatic) this.self = true;

      // bodies are always static
      isStatic = true;

      // emit main body
      //MethodEmit main = DoEmit();
      DoEmit();

      // emit param default wrappers
      //EmitWrappers(main);
    }

    /// <summary>
    /// Emit a mixin router from a class to the mixin body methods.
    /// </summary>
    public void EmitMixinRouter(Method m)
    {
      /*
      String parent  = "fan/" + m.parent().pod().name().val + "/" + m.parent().name().val;
      String name    = m.name().val;
      int jflags     = emit.jflags(m.flags());
      List params    = m.params();
      int paramCount = params.sz();

      // find first param with default value
      int firstDefault = paramCount;
      for (int i=0; i<paramCount; ++i)
        if (((MethodParam)params.get(i)).hasDefault().val)
          { firstDefault = i; break; }

      // generate routers
      for (int i=firstDefault; i<=paramCount; ++i)
      {
        String mySig = signature(m, null, i);
        String implSig = signature(m, parent, i);
        boolean isVoid = mySig.endsWith("V");

        MethodEmit me = emit.emitMethod(name, mySig, jflags);
        CodeEmit code = me.emitCode();
        pushArgs(code, 1+i); // push this + params
        code.op2(INVOKESTATIC, emit.method(parent + "$." + name + implSig));
        code.op(isVoid ? RETURN : ARETURN);
      }
      */
    }

  //////////////////////////////////////////////////////////////////////////
  // Param Default Wrappers
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Emit wrappers.
    /// </summary>
    private void EmitWrappers() //MethodEmit main)
    {
      // change flags so that defaults aren't abstract
// TODO
//      int oldFlags = this.jflags;
//      this.jflags = jflags & ~ABSTRACT;

      PERWAPI.MethodDef main = emitter.methodDef;

      // handle generating default param wrappers
      for (int i=0; i<method.paramCount; i++)
        if (method.vars[i].def != null)
          EmitWrapper(main, i);
      this.paramLen = method.paramCount;

// TODO
//      this.jflags = oldFlags;
    }

    /// <summary>
    /// Emit wrapper.
    /// </summary>
    private void EmitWrapper(PERWAPI.MethodDef main, int paramLen)
    {
      // use explicit param count, and clear code
      this.paramLen = paramLen;
      this.code     = null;

      // emit code
      PERWAPI.CILInstructions code = DoEmit(); //EmitCode();

      // push arguments passed thru
      PushArgs(code, isStatic && !self ? paramLen : paramLen+1);

      // emit default arguments
      for (int i=paramLen; i<method.paramCount; i++)
        new FCodeEmit(emit, method.vars[i].def, code).Emit();

      // call master implementation
      if (isStatic)
        code.MethInst(PERWAPI.MethodOp.call, main);
      else
        code.MethInst(PERWAPI.MethodOp.callvirt, main);

      // return
      code.Inst(PERWAPI.Op.ret);
    }

  //////////////////////////////////////////////////////////////////////////
  // Emit
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// This is the method that all the public emitX methods
    /// route to once everything is setup correctly.
    /// </summary>
    protected PERWAPI.CILInstructions DoEmit()
    {
      int paramCount = paramLen;
      if (self) paramCount++;

      string[] parNames = new string[paramCount];
      string[] parTypes = new string[paramCount];
      for (int i=0; i<paramCount; i++)
      {
        int z = i;
        if (self) z--;

        if (self && i == 0)
        {
          parNames[0] = "self";
          parTypes[0] = selfName;
        }
        else
        {
          parNames[i] = emit.Name(method.vars[z].name);
          parTypes[i] = emit.Nname(method.vars[z].type);
        }
      }

      string[] localNames = new string[method.localCount];
      string[] localTypes = new string[method.localCount];
      for (int i=0; i<method.localCount; i++)
      {
        int z = i + paramCount;
        localNames[i] = emit.Name(method.vars[z].name);
        localTypes[i] = emit.Nname(method.vars[z].type);
      }

      PERWAPI.MethAttr attr;
      if (isPrivate) attr = PERWAPI.MethAttr.Private;
      else if (isInternal) attr = PERWAPI.MethAttr.Assembly;
      else attr = PERWAPI.MethAttr.Public;

      if (isStatic)   attr |= PERWAPI.MethAttr.Static;
      if (isAbstract) attr |= PERWAPI.MethAttr.Abstract;
      if (isVirtual)  attr |= PERWAPI.MethAttr.Virtual;
      if (isOverride) attr |= PERWAPI.MethAttr.Virtual;

      PERWAPI.CILInstructions code = emitter.EmitMethod(name, ret, parNames, parTypes, attr,
        localNames, localTypes);
      if (this.code != null) new FCodeEmit(emit, method, code).Emit();
      return code;
    }

  //////////////////////////////////////////////////////////////////////////
  // Signature Utils
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Generate the java method signature base on our current setup.
    /// </summary>
    //private string Signature()
    //{
      /*
      StringBuilder sig = new StringBuilder();

      // params (with optional implicit self)
      sig.append('(');
      if (self) sig.append('L').append(selfName).append(';');
      for (int i=0; i<paramLen; ++i)
      {
        FMethodVar param = method.vars[i];
        sig.append('L').append(emit.jname(param.type)).append(';');
      }
      sig.append(')');

      // return
      if (isVoid)
        sig.append('V');
      else
        sig.append('L').append(ret).append(';');

      return sig.toString();
      */
    //  return "";
    //}

    /// <summary>
    /// Generate a method signature from a reflection sys::Method.
    /// </summary>
    //private string Signature(Method m, string self, int paramLen)
    //{
      /*
      StringBuilder sig = new StringBuilder();

      // params
      sig.append('(');
      if (self != null) sig.append('L').append(self).append(';');
      for (int i=0; i<paramLen; ++i)
      {
        MethodParam param = (MethodParam)m.params().get(i);
        toSig(sig, param.paramType());
      }
      sig.append(')');

      // return
      if (m.returnType() == Sys.VoidType)
        sig.append('V');
      else
        toSig(sig, m.returnType());

      return sig.toString();
      */
    // return "";
    //}

    /// <summary>
    /// Append specified type to the signature string.
    /// </summary>
    //private static void ToSig(StringBuilder s, Type t)
    //{
    //  //s.append("Lfan/").append(t.pod().name().val).append('/').append(t.name().val).append(';');
    //}

  //////////////////////////////////////////////////////////////////////////
  // Code Utils
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Push the specified number of arguments onto the stack.
    /// </summary>
    private static void PushArgs(PERWAPI.CILInstructions code, int count)
    {
      switch (count)
      {
        case 0:
          return;
        case 1:
          code.Inst(PERWAPI.Op.ldarg_0);
          return;
        case 2:
          code.Inst(PERWAPI.Op.ldarg_0);
          code.Inst(PERWAPI.Op.ldarg_1);
          return;
        case 3:
          code.Inst(PERWAPI.Op.ldarg_0);
          code.Inst(PERWAPI.Op.ldarg_1);
          code.Inst(PERWAPI.Op.ldarg_2);
          return;
        case 4:
          code.Inst(PERWAPI.Op.ldarg_0);
          code.Inst(PERWAPI.Op.ldarg_1);
          code.Inst(PERWAPI.Op.ldarg_2);
          code.Inst(PERWAPI.Op.ldarg_3);
          return;
        default:
          code.Inst(PERWAPI.Op.ldarg_0);
          code.Inst(PERWAPI.Op.ldarg_1);
          code.Inst(PERWAPI.Op.ldarg_2);
          code.Inst(PERWAPI.Op.ldarg_3);
          for (int i=4; i<count; i++)
            code.IntInst(PERWAPI.IntOp.ldarg, i);
          return;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    internal Emitter emitter;  // the emitter for the current assembly
    internal FTypeEmit emit;   // parent type class emitter
    internal FMethod method;   // fan method info
    internal FBuf code;        // code to emit
    internal string name;      // method name
    internal bool isStatic;    // are we emitting a static method
    internal bool isInternal;  // are we emitting a internal method
    internal bool isPrivate;   // are we emitting a private method
    internal bool isAbstract;  // are we emitting an abstract method
    internal bool isVirtual;   // are we emitting a virtual method
    internal bool isOverride;  // are we emitting an overridden method
    internal bool isCtor;      // are we emitting a constructor
    //internal bool isNative;    // are we emitting a native method
    internal bool isVoid;      // is return void
    internal string ret;       // java return sig
    internal bool self;        // add implicit self as first parameter
    internal string selfName;  // class name for self if self is true
    internal int paramLen;     // number of parameters to use
    //internal string sig;       // last java signature emitted
    //internal MethodEmit me;    // last java method emitted

  }
}
