//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   12 Sep 06  Andy Frank  Creation
//

using System;
using System.IO;
using System.Text;
using Fan.Sys;
using Uri = Fan.Sys.Uri;
using Fanx.Util;

namespace Fanx.Fcode
{
  ///
  /// FTable is a 16-bit indexed lookup table for pod constants.
  ///
  public abstract class FTable
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    protected FTable(FPod pod)
    {
      this.pod = pod;
    }

  //////////////////////////////////////////////////////////////////////////
  // Access
  //////////////////////////////////////////////////////////////////////////

    ///
    /// Get the size of the table.
    ///
    public int Size() { return size; }

    ///
    /// Get the object identified by the specified 16-bit index.
    ///
    public Object Get(int index)
    {
      return table[index];
    }

    ///
    /// Dump to the specified print writer.
    ///
    public void Dump(FPod pod, TextWriter writer)
    {
      for (int i=0; i<size; i++)
      {
         writer.Write(StrUtil.Padr("  [" + i + "] ", 8));
         writer.WriteLine(ToString(i));
      }
      writer.Flush();
    }

    ///
    /// Get the value at specified index formated as a string.
    ///
    public virtual string ToString(int index)
    {
      return table[index].ToString();
    }

    ///
    /// Serialize.
    ///
    public abstract FTable Read(FStore.Input input);

  //////////////////////////////////////////////////////////////////////////
  // Names
  //////////////////////////////////////////////////////////////////////////

    internal class Names : FTable
    {
      internal Names(FPod pod) : base(pod) {}

      public override FTable Read(FStore.Input input)
      {
        if (input == null) { size = 0; return this; }
        size = input.U2();
        table = new Object[size];
        for (int i=0; i<size; i++)
          table[i] = String.Intern(input.Utf());
        return this;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // TypeRefs
  //////////////////////////////////////////////////////////////////////////

    internal class TypeRefs : FTable
    {
      internal TypeRefs(FPod pod) : base(pod) {}

      public override string ToString(int index)
      {
        if (index == -1) return "null";
        return ((FTypeRef)table[index]).Sig(pod);
      }

      public override FTable Read(FStore.Input input)
      {
        if (input == null) { size = 0; return this; }
        size = input.U2();
        table = new Object[size];
        for (int i=0; i<size; i++)
          table[i] = FTypeRef.Read(input);
        return this;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // FieldRefs
  //////////////////////////////////////////////////////////////////////////

    internal class FieldRefs : FTable
    {
      internal FieldRefs(FPod pod) : base(pod) {}

      public override string ToString(int index)
      {
        int[] val = ((FTuple)table[index]).val;
        return pod.typeRefs.ToString(val[0]) + "." + pod.Name(val[1]) +
          " -> " + pod.typeRefs.ToString(val[2]);
      }

      public override FTable Read(FStore.Input input)
      {
        if (input == null) { size = 0; return this; }
        size = input.U2();
        table = new Object[size];
        for (int i=0; i<size; i++)
        {
          int[] x = { input.U2(), input.U2(), input.U2() };
          table[i] = new FTuple(x);
        }
        return this;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // MethodRefs
  //////////////////////////////////////////////////////////////////////////

    internal  class MethodRefs : FTable
    {
      internal MethodRefs(FPod pod) : base(pod) {}

      public override string ToString(int index)
      {
        int[] val = ((FTuple)table[index]).val;
        StringBuilder s = new StringBuilder();
        s.Append(pod.typeRefs.ToString(val[0])).Append(".").Append(pod.Name(val[1]));
        s.Append('(');
        for (int i=3; i<val.Length; i++)
        {
          if (i > 3) s.Append(", ");
          s.Append(pod.typeRefs.ToString(val[i]));
        }
        s.Append(") -> ").Append(pod.typeRefs.ToString(val[2]));
        return s.ToString();
      }

      public override FTable Read(FStore.Input input)
      {
        if (input == null) { size = 0; return this; }
        size = input.U2();
        table = new Object[size];
        for (int i=0; i<size; i++)
        {
          int parent = input.U2();
          int name   = input.U2();
          int ret    = input.U2();
          int paramn = input.U1();
          int[] x = new int[3+paramn];
          x[0] = parent;
          x[1] = name;
          x[2] = ret;
          for (int j=0; j<paramn; ++j)
            x[j+3] = input.U2();
          table[i] = new FTuple(x);
        }
        return this;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Ints
  //////////////////////////////////////////////////////////////////////////

    internal class Ints : FTable
    {
      internal Ints(FPod pod) : base(pod) {}

      public override FTable Read(FStore.Input input)
      {
         if (input == null) { size = 0; return this; }
         size = input.U2();
         table = new Object[size];
         for (int i=0; i<size; i++)
           table[i] = Int.Make(input.U8());
         return this;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Floats
  //////////////////////////////////////////////////////////////////////////

    internal class Floats : FTable
    {
      internal Floats(FPod pod) : base(pod) {}

      public override FTable Read(FStore.Input input)
      {
         if (input == null) { size = 0; return this; }
         size = input.U2();
         table = new Object[size];
         for (int i=0; i<size; i++)
           table[i] = Float.Make(input.F8());
         return this;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Strs
  //////////////////////////////////////////////////////////////////////////

    internal class Strs : FTable
    {
      internal Strs(FPod pod) : base(pod) {}

      public override FTable Read(FStore.Input input)
      {
         if (input == null) { size = 0; return this; }
         size = input.U2();
         table = new Object[size];
         for (int i=0; i<size; i++)
           table[i] = Str.Make(String.Intern(input.Utf()));
         return this;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Durations
  //////////////////////////////////////////////////////////////////////////

    internal class Durations : FTable
    {
      internal Durations(FPod pod) : base(pod) {}

      public override FTable Read(FStore.Input input)
      {
         if (input == null) { size = 0; return this; }
         size = input.U2();
         table = new Object[size];
         for (int i=0; i<size; i++)
           table[i] = Duration.Make( input.U8() );
         return this;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Uris
  //////////////////////////////////////////////////////////////////////////

    internal class Uris : FTable
    {
      internal Uris(FPod pod) : base(pod) {}

      public override FTable Read(FStore.Input input)
      {
         if (input == null) { size = 0; return this; }
         size = input.U2();
         table = new Object[size];
         for (int i=0; i<size; i++)
           table[i] = Uri.Parse( Str.Make(input.Utf()) );
         return this;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    internal FPod pod;
    internal int size;
    internal Object[] table;
  }
}