//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   26 Dec 06  Andy Frank  Creation
//

using System.Reflection;
using Fan.Sys;

namespace Fanx.Test
{
  /// <summary>
  /// ListTest.
  /// </summary>
  public class ListTest: CompileTest
  {

  //////////////////////////////////////////////////////////////////////////
  // Main
  //////////////////////////////////////////////////////////////////////////

    public override void Run()
    {
      VerifyKernel();
      VerifyReflect();
      VerifyLang();
      VerifyUnparameterized();
    }

  //////////////////////////////////////////////////////////////////////////
  // Kernel
  //////////////////////////////////////////////////////////////////////////

    public void VerifyKernel()
    {
      Str a = Str.Make("a"), b = Str.Make("b"), c = Str.Make("c"),
          d = Str.Make("d"), e = Str.Make("e"), f = Str.Make("f");
      Int i0 = Int.Make(0);
      Int i1 = Int.Make(1);
      Int i2 = Int.Make(2);

      List x = new List(Sys.StrType);
      Verify(x);

      x.Add(a); Verify(x, a);
      x.Add(b); Verify(x, a, b);
      x.Add(c); Verify(x, a, b, c);
      x.Insert(i0, d); Verify(x, d, a, b, c);
      x.Insert(i1, e); Verify(x, d, e, a, b, c);
      x.RemoveAt(i0); Verify(x, e, a, b, c);
      x.RemoveAt(i2); Verify(x, e, a, c);
      x.RemoveAt(i2); Verify(x, e, a);
      x.RemoveAt(i1); Verify(x, e);
      x.RemoveAt(i0); Verify(x);

      for (int i=0; i<100; ++i)
      {
        List list = new List(Sys.ObjType);
        Obj[] match = new Obj[i];

        for (int j=0; j<i; ++j)
          list.Add(match[j] = Int.Make(j));
        Verify(list, match);

        for (int j=0; j<i; ++j)
          list.RemoveAt(i0);
        Verify(list);
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Reflect
  //////////////////////////////////////////////////////////////////////////

    public void VerifyReflect()
    {
      System.Type nt = CompileToType("class Foo { static Type f() { return Str[,].type() } }");
      Type t = (Type)InvokeStatic(nt, "F");
      //t.Dump();

      Method get = t.Method("get", true);
      Verify(get.ReturnType() == Sys.StrType);

      Method add = t.Method("add", true);
      Verify(((MethodParam)add.Params().Get(0)).ParamType() == Sys.StrType);
      Verify(add.ReturnType() == t);
    }

  //////////////////////////////////////////////////////////////////////////
  // Lang
  //////////////////////////////////////////////////////////////////////////

    public void VerifyLang()
    {
      Verify("Obj f() { return [,] }",    new List(Sys.ObjType));
      Verify("Obj f() { return Str[,] }", new List(Sys.StrType));

      // adds
      Verify("Obj f(Str[] x) { x.add(\"a\"); return x }", new Obj[] {new List(Sys.StrType)}, Make("a"));
      Verify("Obj f(Str[] x) { x.add(\"a\"); x.add(\"b\"); return x }", new Obj[] {new List(Sys.StrType)}, Make("a", "b"));

      // literals
      Verify("Obj f() { return [,] }", new List(Sys.ObjType));
      Verify("Obj f() { return [\"a\"] }", Make("a"));
      Verify("Obj f() { return [\"a\", \"b\"]; }", Make("a", "b"));
      Verify("Obj f() { return [ \"a\" , \"b\" , \"c\" ] }", Make("a", "b", "c"));
      Verify("Obj f() { return [ \"a\" , \"b\" , \"c\", ] }", Make("a", "b", "c"));  // extra comma

      // explicit typing: empty literal
      Verify("Obj f() { return Str[,] }", new List(Sys.StrType));
      // explicit typing: single item - not figured out until resolve time
      Verify("Obj f() { return Str[\"a\"] }", Make("a"));
      // explicit typing: single item, trailing comma
      Verify("Obj f() { return Str[\"a\",] }", Make("a"));
      // explicit typing: multiple items
      Verify("Obj f() { return Str[\"a\", \"b\"] }", Make("a", "b"));

      // slicing
      object[] args = { Make("0", "1", "2", "3") };
      Verify("Obj f(Str[] x) { return x[0] }", args, Str.Make("0"));
      Verify("Str f(Str[] x) { return x[1] }", args, Str.Make("1"));
      Verify("Obj f(Str[] x) { return x[2] }", args, Str.Make("2"));
      Verify("Str f(Str[] x) { return x[3] }", args, Str.Make("3"));
      Verify("Obj f(Str[] x) { return x[-1] }", args, Str.Make("3"));
      Verify("Str f(Str[] x) { return x[-2] }", args, Str.Make("2"));
      Verify("Obj f(Str[] x) { return x[-3] }", args, Str.Make("1"));
      Verify("Str f(Str[] x) { return x[-4] }", args, Str.Make("0"));
      VerifyThrows("Obj f(Str[] x) { return x[4] }",  args, System.Type.GetType("Fan.Sys.IndexErr+Val"));
      VerifyThrows("Str f(Str[] x) { return x[5] }",  args, System.Type.GetType("Fan.Sys.IndexErr+Val"));
      VerifyThrows("Obj f(Str[] x) { return x[-5] }", args, System.Type.GetType("Fan.Sys.IndexErr+Val"));
      VerifyThrows("Str f(Str[] x) { return x[-6] }", args, System.Type.GetType("Fan.Sys.IndexErr+Val"));

      // parameterized typing
//      VerifyErr("Int f(Str[] x) { return x[0] }", "Cannot return 'sys::Str' as 'sys::Int'");
//      VerifyErr("Void f(Str[] x) { x.add(3) }", "Invalid args add(sys::Int) for add(sys::Str)");
//      VerifyErr("Void f(Str[] x, Obj a) { x.add(a) }", "Invalid args add(sys::Obj) for add(sys::Str)");
      //VerifyErr("Int[] f(Str[] x) { return x }", "Cannot return 'sys::Str[]' as 'sys::Int[]'");

      // TODO - just List; should we allow List/Map to be used or force use of Obj[]?
      // Verify("Obj f(List x) { return x[0] }", args, Str.Make("0"));

      // errors
//      VerifyErr("Obj f([] x) {}",             "Expected type identifier, not ']'");
//      VerifyErr("Obj f() { return [] }",      "Invalid list literal; use '[,]' for empty Obj[] list");
//      VerifyErr("Obj f() { return Str[] }",   "Invalid use of type name as expression");
//      VerifyErr("Obj f() { return this[] }",  "Expected expression, not ']'");
//      VerifyErr("Obj f() { return X[] }",     "Unresolved type 'X'");
//      VerifyErr("Obj f() { return X[,] }",    "Unresolved type 'X'");
//      VerifyErr("Obj f() { return x[3] }",    "Unknown variable 'x'");
//      VerifyErr("Obj f() { return [,3] }",    "Expected ']', not '3'");
//      VerifyErr("Obj f() { return [,foo] }",  "Expected ']', not 'foo'");
    }

  //////////////////////////////////////////////////////////////////////////
  // Unparameterized
  //////////////////////////////////////////////////////////////////////////

    public void VerifyUnparameterized()
    {
      object[] args = { Make("0", "1", "2", "3") };
      Verify("Obj f(List x) { return x[0] }", args, Str.Make("0"));
      Verify("Obj f(List x) { return x.get(2) }", args, Str.Make("2"));
    }

  //////////////////////////////////////////////////////////////////////////
  // Convenience
  //////////////////////////////////////////////////////////////////////////

    public List Make(string a) { return new List(Sys.StrType, MakeStrs(a)); }
    public List Make(string a, string b) { return new List(Sys.StrType, MakeStrs(a, b)); }
    public List Make(string a, string b, string c) { return new List(Sys.StrType, MakeStrs(a, b, c)); }
    public List Make(string a, string b, string c, string d) { return new List(Sys.StrType, MakeStrs(a, b, c, d)); }

    public void Verify(List list) { Verify(list, new Obj[0]); }
    public void Verify(List list, Str a) { Verify(list, new Obj[]{a}); }
    public void Verify(List list, Str a, Str b) { Verify(list, new Obj[]{a, b}); }
    public void Verify(List list, Str a, Str b, Str c) { Verify(list, new Obj[]{a, b, c}); }
    public void Verify(List list, Str a, Str b, Str c, Str d) { Verify(list, new Obj[]{a, b, c, d}); }
    public void Verify(List list, Str a, Str b, Str c, Str d, Str e) { Verify(list, new Obj[] {a, b, c, d, e}); }

    public void Verify(List list, Obj[] v)
    {
  //System.Console.WriteLine(list);
      Verify(list.IsEmpty().val == (v.Length == 0));
      Verify(list.Size().val == v.Length);
      for (int i=0; i<list.Size().val; ++i)
        Verify(list.Get(Int.Make(i)) == v[i]);
    }

  }
}