//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   4 Jan 06  Brian Frank  Creation
//

**
** Slot represents a member field or method on a Type.
**
abstract class Slot
{

//////////////////////////////////////////////////////////////////////////
// Management
//////////////////////////////////////////////////////////////////////////

  **
  ** Find a Slot by it's qualified name "pod::Type.slot".  If the slot
  ** doesn't exist and checked is false then return null, otherwise
  ** throw UnknownSlotErr.
  **
  static Slot find(Str qname, Bool checked := true)

  **
  ** Convenience for (Method)find(qname, checked)
  **
  static Method findMethod(Str qname, Bool checked := true)

  **
  ** Convenience for (Field)find(qname, checked)
  **
  static Field findField(Str qname, Bool checked := true)

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  **
  ** Private constructor.
  **
  private new make()

//////////////////////////////////////////////////////////////////////////
// Naming
//////////////////////////////////////////////////////////////////////////

  **
  ** Parent type which defines this slot.
  **
  Type parent()

  **
  ** Simple name of the slot such as "size".
  **
  Str name()

  **
  ** Qualified name such as "sys:Str.size".
  **
  Str qname()

  **
  ** Return true if this is an instance of Field.
  **
  Bool isField()

  **
  ** Return true if this is an instance of Method.
  **
  Bool isMethod()

//////////////////////////////////////////////////////////////////////////
// Flags
//////////////////////////////////////////////////////////////////////////

  **
  ** Return if slot is abstract (no implementation provided).
  ** Applies to Methods only.
  **
  Bool isAbstract()

  **
  ** Return if slot is constant and thread safe.  A constant field
  ** is explicitly marked with the const modifier and guaranteed
  ** to always reference the same immutable object for the life of
  ** the VM.  A const method is guaranteed to not capture any
  ** state from its thread, and is safe to execute on other threads.
  ** The compiler marks methods as const based on the following
  ** analysis:
  **   - static methods are always automatically const
  **   - instance methods are never const
  **   - closures which don't capture any variables from their
  **     scope are automatically const
  **   - curried methods which only capture const variables from
  **     their scope are automatically const
  **
  Bool isConst()

  **
  ** Return if slot is constructor.
  ** Applies to Methods only.
  **
  Bool isCtor()

  **
  ** Return if slot has internal protection scope.
  ** Applies to both Fields and Methods.
  **
  Bool isInternal()

  **
  ** Return if slot is native.
  ** Applies to Methods only.
  **
  Bool isNative()

  **
  ** Return if slot is an override (of parent's virtual method).
  ** Applies to Methods only.
  **
  Bool isOverride()

  **
  ** Return if slot has private protection scope.
  ** Applies to both Fields and Methods.
  **
  Bool isPrivate()

  **
  ** Return if slot has protected protection scope.
  ** Applies to both Fields and Methods.
  **
  Bool isProtected()

  **
  ** Return if slot has public protection scope.
  ** Applies to both Fields and Methods.
  **
  Bool isPublic()

  **
  ** Return if slot is static (class based, rather than instance).
  ** Applies to both Fields and Methods.
  **
  Bool isStatic()

  **
  ** Return if this slot was generated by the compiler.
  **
  Bool isSynthetic()

  **
  ** Return if slot is virtual (may be overridden in subclasses).
  ** Applies to Methods only.
  **
  Bool isVirtual()

//////////////////////////////////////////////////////////////////////////
// Facets
//////////////////////////////////////////////////////////////////////////

  **
  ** Return the facets meta-data for this slot.
  **
  Facets facets()

  **
  ** Convenience for facets.getInt("LineNumber")
  **
  Int lineNumber()

//////////////////////////////////////////////////////////////////////////
// Documentation
//////////////////////////////////////////////////////////////////////////

  **
  ** Return the raw fandoc for this slot or null if not available.
  **
  Str doc()

//////////////////////////////////////////////////////////////////////////
// Conversion
//////////////////////////////////////////////////////////////////////////

  **
  ** Always return qname().
  **
  virtual override Str toStr()

  **
  ** Return a string representation of the Fan code signature.
  **
  virtual Str signature()

}