//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   4 Jan 06  Brian Frank  Creation
//

**
** Type represents a static definition for an Obj class.
**
final class Type
{

//////////////////////////////////////////////////////////////////////////
// Management
//////////////////////////////////////////////////////////////////////////

  **
  ** Find a Type by it's qualified name "pod::Type".  If the type
  ** doesn't exist and checked is false then return null, otherwise
  ** throw UnknownTypeErr.
  **
  static Type find(Str qname, Bool checked := true)

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  **
  ** Private constructor.
  **
  private new privateMake()

//////////////////////////////////////////////////////////////////////////
// Naming
//////////////////////////////////////////////////////////////////////////

  **
  ** Parent pod which defines this type.  For parameterized types derived
  ** from List, Map, or Method, this method always returns the sys pod.
  **
  ** Examples:
  **   Str.type.pod         -> sys
  **   acme::Foo.type.pod   -> acme
  **   acme::Foo[].type.pod -> sys
  **
  Pod pod()

  **
  ** Simple name of the type such as "Str".  For parameterized types derived
  ** from List, Map, or Method, this method always returns "List", "Map",
  ** or "Method" respectively.
  **
  ** Examples:
  **   Str.type.name         -> "Str"
  **   acme::Foo.type.name   -> "Foo"
  **   acme::Foo[].type.name -> "List"
  **
  Str name()

  **
  ** Qualified name formatted as "pod::name".  For parameterized types derived
  ** from List, Map, or Method, this method always returns "sys::List",
  ** "sys::Map", or "sys::Method" respectively.
  **
  ** Examples:
  **   Str.type.qname         -> "sys::Str"
  **   acme::Foo.type.qname   -> "acme::Foo"
  **   acme::Foo[].type.qname -> "sys::List"
  **
  Str qname()

  **
  ** Return the formal signature of this type.  In the case of non-parameterized
  ** types the signature is the same as qname.  However, for parameterized types
  ** derived from List, Map, or Method the signature uses the following special
  ** syntax:
  **   List   -> V[]
  **   Map    -> [K:V]
  **   Method -> |A,B...->R|
  **
  ** Examples:
  **   Str.type.signature -> "sys::Str"
  **   Int[].type.signature -> "sys::Int[]"
  **   Int:Str.type.signature -> "[sys::Int:sys::Str]"
  **   Str:Buf[].type.signature -> [sys::Str:sys::Buf[]]
  **   |Float x->Bool|.type.signature -> "|sys::Float->sys::Bool|"
  **   |Float x, Int y|.type.signature -> |sys::Float,sys::Int->sys::Void|
  **
  Str signature()

//////////////////////////////////////////////////////////////////////////
// Inheritance
//////////////////////////////////////////////////////////////////////////

  **
  ** The direct super class of this type (null for Obj).
  ** Return sys::Obj for all mixin types.
  **
  ** Examples:
  **   Obj.type.base -> null
  **   Int.type.base -> sys::Num
  **   OutStream.type.base -> sys::Obj
  **
  Type base()

  **
  ** Return the mixins directly implemented by this type.
  **
  ** Examples:
  **   Obj.type.mixins -> [,]
  **   Buf.type.mixins -> [sys::InStream, sys::OutStream]
  **   OutStream.type.mixins -> [,]
  **
  Type[] mixins()

  **
  ** Return a recursive flattened list of all the types this type
  ** inherits from.  The result list always includes this type itself.
  ** The result of this method represents the complete list of types
  ** implemented by this type - instances of this type are assignable
  ** to any type in this list.  All types (including mixins) will
  ** include sys::Obj in this list.
  **
  ** Examples:
  **   Obj.type.inheritance -> [sys::Obj]
  **   Int.type.inheritance -> [sys::Obj, sys::Int, sys::Num]
  **   OutStream.type.inheritance -> [sys::Obj, sys::OutStream]
  **
  Type[] inheritance()

  **
  ** Does this type implement the specified type.  If true, then
  ** this type is assignable to the specified type (although the
  ** converse is not necessarily true).  This method provides the
  ** same semantics as the "is" operator, but between two types
  ** rather than an instance and a type.  All types (including
  ** mixin types) fit sys::Obj.
  **
  ** Example:
  **   Float.type.fits(Float.type) -> true
  **   Float.type.fits(Num.type)  -> true
  **   Float.type.fits(Obj.type)  -> true
  **   Float.type.fits(Str.type)  -> false
  **   Obj.type.fits(Float.type)  -> false
  **
  Bool fits(Type t)

//////////////////////////////////////////////////////////////////////////
// Generics
//////////////////////////////////////////////////////////////////////////

  **
  ** A generic type contains slot signatures which may be parameterized - for
  ** example Map leaves the key and value types generic as K and V.  Fan supports
  ** three built-in generic types: List, Map, and Method.   A parameterized
  ** type such as Str[] is not a generic type (all of its generic parameters
  ** have been filled in).  User defined generic types are not supported in Fan.
  **
  ** Examples:
  **   Str.type.isGeneric   -> false
  **   List.type.isGeneric  -> true
  **   Str[].type.isGeneric -> false
  **
  Bool isGeneric()

  **
  ** If this is a parameterized type, then return the map of names to
  ** types.  If this is not a parameterized type return an empty map.
  **
  ** Examples:
  **   Str.type.params -> [:]
  **   Str[].type.params -> ["V":Str, "L":Str[]]
  **   Int:Slot.type.params -> ["K":Int, "V":Slot, "M":Int:Slot]
  **   |Int x, Float y->Bool|.type.params -> ["A":Int, "B":Float, "R":Bool]
  **
  Str:Type params()

  **
  ** If this is a generic type, then dynamically create a new parameterized
  ** type with the specified name to type map.  If this type is not generic
  ** then throw UnsupportedErr.  Throw ArgErr if params fails to specify
  ** the required parameters:
  **    List   -> V required
  **    Map    -> K, V required
  **    Method -> R required, A-H optional
  **
  ** Examples:
  **   List.type.parameterize(["V":Bool.type]) -> Bool[]
  **   Map.type.parameterize(["K":Str.type, "V":Obj.type]) -> Str:Obj
  **
  Type parameterize(Str:Type params)

  **
  ** Convenience for List.type.parameterize(["V":this])
  **
  ** Examples:
  **   Int.type.toListOf -> Int[]
  **   Str[].type.toListOf -> Str[][]
  **
  Type toListOf()

//////////////////////////////////////////////////////////////////////////
// Flags
//////////////////////////////////////////////////////////////////////////

  **
  ** Return if this Type is abstract and cannot be instantiated.  This
  ** method will always return true if the type is a mixin.
  **
  Bool isAbstract()

  **
  ** Return if this Type is a class (as opposed to enum or mixin)
  **
  Bool isClass()

  **
  ** Return if this is a const class which means instances of this
  ** class are immutable.
  **
  Bool isConst()

  **
  ** Return if this Type is an Enum type.
  **
  Bool isEnum()

  **
  ** Return if this Type is marked final which means it may not be subclassed.
  **
  Bool isFinal()

  **
  ** Return if this Type has internal protection scope.
  **
  Bool isInternal()

  **
  ** Return if this Type is a mixin type and cannot be instantiated.
  **
  Bool isMixin()

  **
  ** Return if this Type has public protection scope.
  **
  Bool isPublic()

  **
  ** Return if this Type was generated by the compiler.
  **
  Bool isSynthetic()

//////////////////////////////////////////////////////////////////////////
// Reflection
//////////////////////////////////////////////////////////////////////////

  **
  ** List of the all defined fields (including inherited fields).
  **
  Field[] fields()

  **
  ** List of the all defined methods (including inherited methods).
  **
  Method[] methods()

  **
  ** List of the all defined slots, both fields and methods (including
  ** inherited slots).
  **
  Slot[] slots()

  **
  ** Convenience for (Field)slot(name, checked)
  **
  Field field(Str name, Bool checked := true)

  **
  ** Convenience for (Method)slot(name, checked)
  **
  Method method(Str name, Bool checked := true)

  **
  ** Lookup a slot by name.  If the slot doesn't exist and checked
  ** is false then return null, otherwise throw UnknownSlotErr.
  **
  Slot slot(Str name, Bool checked := true)

  **
  ** Create a new instance of this Type using a public constructor
  ** called "make".  Throw Err if public constructor "make" is
  ** not available.
  **
  Obj make(Obj[] args := null)

//////////////////////////////////////////////////////////////////////////
// Facets
//////////////////////////////////////////////////////////////////////////

  **
  ** Return the facets meta-data for this type.
  **
  Facets facets()

  **
  ** Convenience for facets.get("SourceFile")
  **
  Str sourceFile()

  **
  ** Convenience for facets.getInt("LineNumber")
  **
  Int lineNumber()

//////////////////////////////////////////////////////////////////////////
// Documentation
//////////////////////////////////////////////////////////////////////////

  **
  ** Return the raw fandoc for this type or null if not available.
  **
  Str doc()

//////////////////////////////////////////////////////////////////////////
// Conversion
//////////////////////////////////////////////////////////////////////////

  **
  ** Always return signature().
  **
  override Str toStr()

}