//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   2 Apr 06  Brian Frank  Creation
//

**
** BufTest
**
class BufTest : Test
{

//////////////////////////////////////////////////////////////////////////
// Equality
//////////////////////////////////////////////////////////////////////////

  Void testEquality()
  {
    a := Buf.make
    b := Buf.make

    verifyEq(a, b)

    a.write(0); b.write(0)
    verifyEq(a, b)

    a.write(0xff); b.write(0xff)
    verifyEq(a, b)

    a.flip
    verifyEq(a, b)

    a.write(0xab); b.write(0xab)
    verifyNotEq(a, b)
  }

//////////////////////////////////////////////////////////////////////////
// Basic IO
//////////////////////////////////////////////////////////////////////////

  Void testBasicIO()
  {
    buf := Buf.make
    verifyEq(buf.pos,  0)
    verifyEq(buf.size, 0)
    verify(buf.empty)

    // write size=1
    buf.write('a')
    verifyEq(buf.pos,  1)
    verifyEq(buf.size, 1)
    verifyEq(buf[0], 'a')
    verifyFalse(buf.empty)

    // write size=2
    buf.write('b')
    verifyEq(buf.pos,  2)
    verifyEq(buf.size, 2)
    verifyEq(buf[0], 'a')
    verifyEq(buf[1], 'b')

    // write size=3
    buf.write('c')
    verifyEq(buf.pos,  3)
    verifyEq(buf.size, 3)
    verifyEq(buf[0], 'a')
    verifyEq(buf[1], 'b')
    verifyEq(buf[2], 'c')

    // rewrite pos 1
    buf.seek(1).write('x')
    verifyEq(buf.pos,  2)
    verifyEq(buf.size, 3)
    verifyEq(buf[0], 'a')
    verifyEq(buf[1], 'x')
    verifyEq(buf[2], 'c')
    buf.seek(1).write('b')

    // seek pos=0
    buf.seek(0)
    verifyEq(buf.pos,  0)
    verifyEq(buf.size, 3)

    // read pos=1
    verifyEq(buf.read, 'a')
    verifyEq(buf.pos,  1)
    verifyEq(buf.size, 3)
    verifyEq(buf.remaining, 2)

    // read pos=2
    verifyEq(buf.read, 'b')
    verifyEq(buf.pos,  2)
    verifyEq(buf.size, 3)
    verifyEq(buf.remaining, 1)

    // read pos=3
    verifyEq(buf.read, 'c')
    verifyEq(buf.pos,  3)
    verifyEq(buf.size, 3)
    verifyEq(buf.remaining, 0)

    // read pos=end
    verifyEq(buf.read, null)
    verifyEq(buf.read, null)
    verifyEq(buf.pos,  3)
    verifyEq(buf.size, 3)
    verifyEq(buf.remaining, 0)

    // gets
    verifyEq(buf[0],  'a')
    verifyEq(buf[1],  'b')
    verifyEq(buf[2],  'c')
    verifyEq(buf[-1], 'c')
    verifyEq(buf[-2], 'b')
    verifyEq(buf[-3], 'a')

    // get errors
    verifyErr(IndexErr.type) |->Int| { return Buf.make[0] }
    verifyErr(IndexErr.type) |->Int| { return buf[3] }
    verifyErr(IndexErr.type) |->Int| { return buf[-4] }

    // sets
    buf[0] = 'A';
    buf[1] = 'B';
    buf[2] = 'C';
    verifyEq(buf[0], 'A')
    verifyEq(buf[1], 'B')
    verifyEq(buf[2], 'C')
    buf[-1] = 0xff;
    buf[-2] = 0xdd;
    buf[-3] = 0xcc;
    verifyEq(buf[0], 0xcc)
    verifyEq(buf[1], 0xdd)
    verifyEq(buf[2], 0xff)

    // set errors
    verifyErr(IndexErr.type) |,| { Buf.make[0] = 99 }
    verifyErr(IndexErr.type) |,| { buf[3] = 99 }
    verifyErr(IndexErr.type) |,| { buf[-4] = 99 }

    // clear
    buf.clear
    verifyEq(buf.pos,  0)
    verifyEq(buf.size, 0)
    verifyEq(buf.read, null)
  }

//////////////////////////////////////////////////////////////////////////
// Size and Capacity
//////////////////////////////////////////////////////////////////////////

  Void testSizeCapacity()
  {
    verifyEq(Buf.make.capacity, 1024)

    b := Buf.make(2)
    verifyEq(b.capacity, 2)
    verifyEq(b.size, 0)

    b.write('a').write('b')
    verifyEq(b.capacity, 2)
    verifyEq(b.size, 2)

    b.write('c')
    verifyEq(b.capacity, 4)
    verifyEq(b.size, 3)

    b.write('d').write('e')
    verifyEq(b.capacity, 8)
    verifyEq(b.size, 5)

    b.capacity = 6
    verifyEq(b.capacity, 6)
    verifyEq(b.size, 5)

    b.write('f')
    verifyEq(b.capacity, 6)
    verifyEq(b.size, 6)

    b.write('g')
    verifyEq(b.capacity, 12)
    verifyEq(b.size, 7)

    b.capacity = 7
    verifyEq(b.capacity, 7)
    verifyEq(b.size, 7)

    verifyErr(ArgErr.type) |,| { b.capacity = 6 }
    verifyErr(ArgErr.type) |,| { b.capacity = -9 }

    b.size = 8
    verifyEq(b.capacity, 8)
    verifyEq(b.size, 8)
    verifyEq(b[0], 'a')
    verifyEq(b[6], 'g')
    verifyEq(b[7], 0)

    b.size = 4
    verifyEq(b.capacity, 8)
    verifyEq(b.size, 4)
    verifyEq(b[0], 'a')
    verifyEq(b[3], 'd')
    verifyErr(IndexErr.type) |->Int| { return b[4] }
    verifyErr(IndexErr.type) |->Int| { return b[7] }

    b.size = 7
    verifyEq(b.capacity, 8)
    verifyEq(b.size, 7)
    verifyEq(b[0], 'a')
    verifyEq(b[6], 'g')
    verifyErr(IndexErr.type) |->Int| { return b[7] }
  }

//////////////////////////////////////////////////////////////////////////
// Slice
//////////////////////////////////////////////////////////////////////////

  Void testSlice()
  {
     buf := Buf.make
     buf.charset = Charset.utf16BE
     buf.write(0xaa).write(0xbb).write(0xcc).write(0xdd)
     verifyEq(buf[0..0].toStr, "0x[aa]")
     verifyEq(buf[0..1].toStr, "0x[aabb]")
     verifyEq(buf[1..2].toStr, "0x[bbcc]")
     verifyEq(buf[0..2].toStr, "0x[aabbcc]")
     verifyEq(buf[3..3].toStr, "0x[dd]")
     verifyEq(buf[0..-1].toStr, "0x[aabbccdd]")
     verifyEq(buf[-3..-2].toStr, "0x[bbcc]")
     verifyEq(buf[1...2].toStr, "0x[bb]")
     verifyEq(buf[1...3].toStr, "0x[bbcc]")
     verifyEq(buf[-2..-1].toStr, "0x[ccdd]")
     verifyEq(buf[0...2].toStr, "0x[aabb]")
     verifyEq(buf[1..-2].toStr, "0x[bbcc]")
     verifyEq(buf[4..-1].toStr, "0x[]")

     verifyEq(buf[1..-2].size, 2)
     verifyEq(buf[1..-2].charset, Charset.utf16BE)

     verifyErr(IndexErr.type) |->Buf| { return buf[0..4] }
     verifyErr(IndexErr.type) |->Buf| { return buf[3..4] }
     verifyErr(IndexErr.type) |->Buf| { return buf[0...5] }
     verifyErr(IndexErr.type) |->Buf| { return buf[3..1] }
     verifyErr(IndexErr.type) |->Buf| { return buf[3...2] }
  }

//////////////////////////////////////////////////////////////////////////
// Dummies
//////////////////////////////////////////////////////////////////////////

  Void testBufDummies()
  {
    b := Buf.make
    b.flush
    verifyEq(b.close, true)
  }

//////////////////////////////////////////////////////////////////////////
// Write
//////////////////////////////////////////////////////////////////////////

  Void testWriteBuf()
  {
    a := Buf.make
    b := Buf.make

    a.write('a').write('b').write('c')
    verifyEq(a.pos, 3)
    verifyBufEq(a, "abc")

    // write entirity
    a.flip
    b.writeBuf(a)
    verifyEq(a.pos,  3)
    verifyEq(a.size, 3)
    verifyEq(a.remaining, 0)
    verifyEq(b.pos,  3)
    verifyEq(b.size, 3)
    verifyEq(b.remaining, 0)
    verifyBufEq(b, "abc")

    // write one byte at a time
    a.flip
    3.times |Int i|
    {
      b.writeBuf(a, 1)
      verifyEq(a.pos,  i+1)
      verifyEq(a.size, 3)
      verifyEq(a.remaining, 3-i-1)
      verifyEq(b.pos,  4+i)
      verifyEq(b.size, 4+i)
      verifyEq(b.remaining, 0)
    }
    verifyBufEq(b, "abcabc")

    a.flip
    verifyEq(b.size, 6)
    b.seek(1).writeBuf(a).write('|')
    verifyBufEq(b, "aabc|c")
    verifyEq(b.pos, 5)
    verifyEq(b.size, 6)

    b.seek(b.size).write('x')
    verifyBufEq(b, "aabc|cx")
  }

//////////////////////////////////////////////////////////////////////////
// Read
//////////////////////////////////////////////////////////////////////////

  Void testReadBuf()
  {
    a := Buf.make
    b := Buf.make

    a.write('a').write('b').write('c')
    verifyEq(a.pos,  3)
    verifyBufEq(a, "abc")

    // read the entirity exactly
    a.seek(0)
    verifyEq(a.readBuf(b, 3), 3)
    verifyEq(a.pos,  3)
    verifyEq(a.remaining, 0)
    verifyEq(b.pos,  3)
    verifyBufEq(b, "abc")
    verifyEq(b.readBuf(b, 1), null)

    // reset and read the more than entirity
    a.seek(0)
    b.clear
    verifyEq(a.readBuf(b, 100), 3)
    verifyEq(a.pos,  3)
    verifyEq(a.remaining, 0)
    verifyEq(b.pos,  3)
    verifyBufEq(b, "abc")
    verifyEq(b.readBuf(b, 1), null)

    // reset and read 1 byte at a time
    a.seek(0)
    b.clear
    verifyEq(a.readBuf(b, 1), 1)
    verifyEq(a.pos,  1)
    verifyEq(a.remaining, 2)
    verifyEq(b.pos,  1)
    verifyBufEq(b, "a")
    verifyEq(a.readBuf(b, 1), 1)
    verifyEq(a.pos,  2)
    verifyEq(a.remaining, 1)
    verifyEq(b.pos,  2)
    verifyBufEq(b, "ab")
    verifyEq(a.readBuf(b, 1), 1)
    verifyEq(a.pos,  3)
    verifyEq(a.remaining, 0)
    verifyEq(b.pos,  3)
    verifyBufEq(b, "abc")
    verifyEq(a.readBuf(b, 1), null)
    verifyBufEq(b, "abc")

    // reset and read 2 bytes at a time
    a.seek(0)
    b.clear
    verifyEq(a.readBuf(b, 2), 2)
    verifyEq(a.pos,  2)
    verifyEq(a.remaining, 1)
    verifyEq(b.pos,  2)
    verifyBufEq(b, "ab")
    verifyEq(a.readBuf(b, 2), 1)
    verifyEq(a.pos,  3)
    verifyEq(a.remaining, 0)
    verifyEq(b.pos,  3)
    verifyBufEq(b, "abc")
  }

//////////////////////////////////////////////////////////////////////////
// Binary IO
//////////////////////////////////////////////////////////////////////////

  Void testBinary()
  {
    // reuse StreamTests
    buf := Buf.make(0)
    StreamTest.writeBinary(buf)

    buf.flip()
    StreamTest.readBinary(this, buf)

    // sanity check sizes
    buf.charset = Charset.utf8
    buf.clear; buf.writeChar('a'); verifyEq(buf.size, 1)
    buf.clear; buf.writeChar('\u00fe'); verifyEq(buf.size, 2)
    buf.clear; buf.writeChar('\uabcd'); verifyEq(buf.size, 3)

    buf.charset = Charset.utf16BE
    buf.clear; buf.writeChar('a'); verifyEq(buf.size, 2)
    buf.clear; buf.writeChar('\u00fe'); verifyEq(buf.size, 2)
    buf.clear; buf.writeChar('\uabcd'); verifyEq(buf.size, 2)
  }

//////////////////////////////////////////////////////////////////////////
// Char IO
//////////////////////////////////////////////////////////////////////////

  Void testChar()
  {
    buf := Buf.make
    buf.clear
    StreamTest.writeChar(buf)

    buf.flip
    StreamTest.readChar(this, buf)
  }

//////////////////////////////////////////////////////////////////////////
// Unread
//////////////////////////////////////////////////////////////////////////

  Void testUnread()
  {
    buf := Buf.make
    buf.write(0xaa).write(0xbb)
    buf.flip
    buf.unread(0x11)
    buf.unread(0x22)
    verifyEq(buf.read, 0x22)
    verifyEq(buf.read, 0x11)
    verifyEq(buf.read, 0xaa)
    buf.unread(0xaa)
    verifyEq(buf.read, 0xaa)
    verifyEq(buf.read, 0xbb)
    verifyEq(buf.read, null)

    buf.clear
    StreamTest.writeUnread(buf)

    buf.flip
    StreamTest.readUnread(this, buf)
  }

//////////////////////////////////////////////////////////////////////////
// Read Lines
//////////////////////////////////////////////////////////////////////////

  Void testReadLine()
  {
    buf := Buf.make
    buf.clear
    StreamTest.writeReadLine(buf)

    buf.flip
    StreamTest.readReadLine(this, buf)
  }

  Void testReadAllLines()
  {
    buf := Buf.make
    buf.clear
    StreamTest.writeReadAllLines(buf)

    buf.flip
    StreamTest.readReadAllLines(this, buf)
  }

  Void testReadAllStr()
  {
    buf := Buf.make
    buf.clear
    StreamTest.writeReadAllStr(buf)

    buf.flip
    StreamTest.readReadAllStr0(this, buf)

    buf.seek(0)
    StreamTest.readReadAllStr1(this, buf)
  }

//////////////////////////////////////////////////////////////////////////
// Pipe
//////////////////////////////////////////////////////////////////////////

  Void testPipe()
  {
    src  := Buf.make
    sink := Buf.make

    100.times |Int i| { src.write(i) }

    src.flip
    src.pipe(sink)
    verifyEq(sink.size, 100)
    100.times |Int i| { verifyEq(sink[i], i) }

    src.seek(0)
    sink.clear
    src.pipe(sink, 5)
    verifyEq(sink.size, 5)
    5.times |Int i| { verifyEq(sink[i], i) }
  }

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  Void verifyBufEq(Buf buf, Str ascii)
  {
    verifyEq(buf.size, ascii.size)
    for (Int i := 0; i<buf.size; ++i)
      verifyEq(buf[i], ascii[i])
  }

  // TODO
  //  - test Buf binary IO
  //  - test Buf text IO
  //  - test Buf subclass

}