//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   26 Mar 06  Brian Frank  Creation
//

**
** FileTest
**
class FileTest : Test
{

  Void testHomeDir()
  {
    verify(Sys.homeDir.exists)
    verify(Sys.homeDir.isDir)
  }

  Void testTestDir()
  {
    // verify clean empty directory
    verify(tempDir.exists)
    verify(tempDir.isDir)
    verify(tempDir.list.isEmpty)
    verify(tempDir.list.size == 0)
    verifyEq(tempDir.parent.uri, tempDir.uri.parent)
  }

  Void testNormalize()
  {
    // this test really doesn't do anything other
    // than verify the method is there and does something
    f := File.make(`./`)
    verify(f.path.size < f.normalize.path.size)
    verifyEq(f.uri, f.uri.toFile.uri)
  }

  Void testPlus()
  {
    f := File.make(`a/b/c/`);
    verifyEq(f + `d`,   File.make(`a/b/c/d`))
    verifyEq(f + `d/e`,  File.make(`a/b/c/d/e`))
    verifyEq(f + `../d`, File.make(`a/b/d`))

    f = File.make(`a/b/c`);
    verifyEq(f + `d`,   File.make(`a/b/d`))
    verifyEq(f + `d/e`,  File.make(`a/b/d/e`))
    verifyEq(f + `../d`, File.make(`a/d`))
  }

  Void testCreateTemp()
  {
    file := File.createTemp.deleteOnExit
    verifyEq(file.size, 0)
    verify(file.exists)
    verify(file.name.startsWith("fan"))
    verify(file.name.endsWith(".tmp"))

    file = File.createTemp("f").deleteOnExit
    verifyEq(file.size, 0)
    verify(file.exists)
    verify(file.name.startsWith("f"))
    verify(file.name.endsWith(".tmp"))

    file = File.createTemp("fantest", ".fantest").deleteOnExit
    verifyEq(file.size, 0)
    verify(file.exists)
    verify(file.name.startsWith("fantest"))
    verify(file.name.endsWith(".fantest"))

    file = File.createTemp("fantest", ".fantest", Sys.homeDir).deleteOnExit
    verifyEq(file.size, 0)
    verify(file.exists)
    verify(file.name.startsWith("fantest"))
    verify(file.name.endsWith(".fantest"))
    verify(file.toStr.startsWith(Sys.homeDir.toStr))

    verifyErr(IOErr.type) |,| { File.createTemp("xyz", ".tmp", file) }
    verifyErr(IOErr.type) |,| { File.createTemp("xyz", ".tmp", FileTest.type.pod.files[`/res/test.txt`]) }
  }

  Void testCreateAndDelete()
  {
    // create file - no extension
    f := (tempDir + `file`).create
    verify(!f.isDir)
    verify(f.list.isEmpty)
    verifyEq(f.size, 0)
    verifyEq(f.name, "file")
    verifyEq(f.basename, "file")
    verifyEq(f.ext, null)
    verifyEq(f.uri.path.last, "file")
    verifyEq(f.path.last, "file")
    verifyEq(f.pathStr, f.toStr)
    verifyEq(tempDir.list, [f])

    // create file - with extension
    e := (tempDir + `file.txt`).create
    verify(!e.isDir)
    verify(e.list.isEmpty)
    verifyEq(e.size, 0)
    verifyEq(e.name, "file.txt")
    verifyEq(e.basename, "file")
    verifyEq(e.ext, "txt")
    verifyEq(e.uri.path.last, "file.txt")
    verifyEq(e.path.last, "file.txt")
    verifyEq(e.pathStr, e.toStr)
    verifyEq(tempDir.list.sort, [f, e])

    // create dir
    d := (tempDir + `dir/`).create
    verify(d.isDir)
    verify(d.list.isEmpty)
    verifyEq(d.name, "dir")
    verifyEq(d.basename, "dir")
    verifyEq(d.ext, null)
    verifyEq(d.uri.path.last, "dir")
    verifyEq(d.path.last, "dir")
    verifyEq(d.pathStr, d.toStr)
    verifyEq(d.size, null)
    verifyEq(tempDir.list.sort, [d, f, e])

    // create another representation of "file"
    fx := tempDir + `file`
    verify(f == fx)
    verify(!(f != fx))
    verify(!(f === fx))
    verifyEq(f, fx)
    verifyEq(f.hash, fx.hash)

    // check create for existing
    verify(d.exists); d.create
    verify(f.exists); f.create

    // check errors
    verifyErr(IOErr.type) |,| { File.make((f.pathStr+"/").toUri) }
    verifyErr(IOErr.type) |,| { File.make(d.toStr[0..-2].toUri) }

    // delete
    d.delete; verifyFalse(d.exists)
    f.delete; verifyFalse(f.exists)

    // re-create and call deleteOnExit (can't really test this easily,
    // but you can manually look in test directory after running this
    // test to verify files deleted on exit)
    d.create.deleteOnExit; verify(d.exists)
    f.create.deleteOnExit; verify(f.exists)
  }

  Void testStreamConvenience()
  {
    f := tempDir + `testfile.txt`
    out := f.out
    out.writeChars("alpha\nbeta\rgamma")
    out.close

    buf := f.readAllBuf
    verifyEq(buf.readAllStr, "alpha\nbeta\ngamma")

    lines := f.readAllLines
    verifyEq(lines, ["alpha", "beta", "gamma"])

    all := f.readAllStr
    verifyEq(all, "alpha\nbeta\ngamma")

    allNoNorm := f.readAllStr(false)
    verifyEq(allNoNorm, "alpha\nbeta\rgamma")
  }

  Void testModifyTime()
  {
    start := Time.now
    f := tempDir + `testfile.txt`
    verifyEq(f.modified, null)

    f.create
    verify(start <= f.modified && f.modified<= Time.now)

    yesterday := Time.make(Time.now.ticks - 1days.ticks)
    f.modified = yesterday
    verifyEq(f.modified , yesterday)
  }

  Void testOsPath()
  {
    // abs
    f := tempDir
    g := File.os(f.osPath)
    verifyEq(f.uri, g.uri)
    verifyEq(f.osPath, g.osPath)

    // relative
    f = File.make(`foo/file.txt`)
    g = File.os(f.osPath)
    verifyEq(f.uri, g.uri)
    verifyEq(f.osPath, g.osPath)

    // extra Windoze testing
    if (Sys.env.get("os.name", "").contains("Windows"))
    {
      f = File.make(`/a/b/c.txt`)
      verifyEq(f.uri.toStr.replace("/", "\\"), f.osPath)
    }
  }

}